"""
@package: dc
@file CollectURLs.py
@author Scorp <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""
import logging
import lxml.html
import hashlib
import re
import app.Consts as APP_CONSTS
import dc_processor.Constants as PCONSTS
import app.Utils as Utils
import dc.EventObjects
import CrawlerTask
import json
import copy
from app.Filters import Filters
from app.Utils import varDump
try:
  import feedparser
except ImportError:
  feedparser = None


logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)


class CollectURLs(object):

  BINARY_CONTENT_TYPE_PATTERN = re.compile('(text)|(xml)', re.I)
  WITH_URL_XPATH_LIST = ["//a/@href", "//applet/@codebase", "//area/@href", "//base/@href", \
  "//blockquote/@cite", "//body/@background", "//del/@cite", \
  "//frame/@src", "//head/@profile", "//iframe/@src", "//input/@src", "//ins/@cite", \
  "//object/@codebase", "//q/@cite", "//audio/@src", \
  "//button/@formaction", "//command/@icon", "//embed/@src", "//html/@manifest", \
  "//input/@formaction", "//source/@src", "//video/@src", "//video/@poster", "//category/@domain",
  "//link", "//guid"]  # pylint: disable=C0330
  COLLECT_POST_DATA_NAME = "COLLECT_POST_DATA"
  COLLECT_POST_DATA = "1"
  DC_URLS_TABLE_PREFIX = "urls_"
  PATTERN_WITH_PROTOCOL = re.compile('[a-zA-Z]+:(//)?')
  DETECT_MIME_MAIN_CONTENT = "1"
  DETECT_MIME_COLLECTED_URL = "2"
  DETECT_MIME_TIMEOUT = 1


  def __init__(self):
    self.crawledResource = None
    self.url = None
    self.dom = None
    self.realUrl = None
    self.processorName = None
    self.batchItem = None
    self.urlXpathList = None
    self.feedItems = None
    self.feed = None
    self.siteProperties = None
    self.site = None
    self.dbWrapper = None
    self.autoRemoveProps = None
    self.autoDetectMime = None
    self.processContentTypes = None
    self.postForms = None
    self.urlProcess = None
    self.robotsParser = None


  # #checkClassFields method checks all class's mandatory fields
  #
  def checkClassFields(self):
    for field in self.__dict__:
      if field not in ["feedItems", "feed", "processorName", "autoDetectMime", "processContentTypes", "postForms",
                       "robotsParser"]:
        if not hasattr(self, field) or getattr(self, field) is None:
          logger.debug(">>> [CollectURLs] Mandatory field must be initialized, field Name = " + field)
          raise Exception(">>> [CollectURLs] Mandatory field must be initialized, field Name = " + field)


  # #process method
  #
  def process(self, httpCode, readOnly=False, httpApplyHeaders=None):
    self.checkClassFields()
    localSiteId = self.batchItem.siteId if self.batchItem.siteId else "0"
    nextStep = True
    useChains = False
    internalLinks, externalLinks = [], []
    maxURLsFromPage = 0
    params = []
    chainUrls = []
    formUrls = None
    formMethods = None
    formFields = None
    urlSet = set()

    if self.site is not None and self.site.maxURLsFromPage is not None:
      maxURLsFromPage = self.site.maxURLsFromPage

    if self.url is not None and self.url.maxURLsFromPage is not None and self.url.maxURLsFromPage > 0:
      maxURLsFromPage = self.url.maxURLsFromPage

    if nextStep and self.crawledResource is not None and \
    not self.BINARY_CONTENT_TYPE_PATTERN.search(self.crawledResource.content_type):
      logger.debug("Binary content: <<%s>>", self.crawledResource.content_type)
      nextStep = False

    # don't parse url for 4XX or 5XX response
    if nextStep and self.crawledResource is not None:
      code_type = int(self.crawledResource.http_code) / 100
      if code_type == 4 or code_type == 5:
        nextStep = False

    if nextStep and self.crawledResource is not None and not self.crawledResource.html_content:
      nextStep = False

    if nextStep and self.dom is None:
      logger.debug("DOM is None")
      nextStep = False

    if nextStep:
      useChains = True
      self.processProcessor(urlSet, self.dom, self.urlXpathList, self.batchItem.urlObj)

      formUrls, formMethods, formFields = self.extractFormURL(self.dom, self.siteProperties)
      urlSet.update(formUrls)

      if self.url.type == dc.EventObjects.URL.TYPE_SINGLE:
        logger.debug(">>> URL type: single")
        nextStep = False

    if nextStep:
      if self.url.type == dc.EventObjects.URL.TYPE_CHAIN:
        logger.debug(">>> URL type: chain")
        nextStep = False

    if nextStep:
      #(3) END
      urlTable = self.DC_URLS_TABLE_PREFIX + localSiteId

      numUrlsFromPage = 0
      self.urlProcess.urlTable = urlTable
      localFilters = None
      try:
        if self.siteProperties is not None and "RSS_FEED_ZERO_ITEM" in self.siteProperties and \
        int(self.siteProperties["RSS_FEED_ZERO_ITEM"]) == 1:
          if self.processorName == PCONSTS.PROCESSOR_FEED_PARSER or self.processorName == PCONSTS.PROCESSOR_RSS:
            self.feedElementsProcessing(self.url.urlMd5, httpCode, self.url.url, localSiteId, self.url, self.url.url,
                                        params, True)
      except ValueError:
        logger.debug(">>> Wrong \"RSS_FEED_ZERO_ITEM\" property's value")

      logger.debug(">>> urlSet is = " + str(urlSet))

      for elemUrl in urlSet:
        if elemUrl is None:
          logger.debug(">>> Some url from urlSet is None")
          continue
        if maxURLsFromPage > 0 and numUrlsFromPage >= maxURLsFromPage:
          logger.debug("Site maxURLsFromPage=%s limit reached on %s number.",
                       str(maxURLsFromPage), str(numUrlsFromPage))
          break

        localUrl = elemUrl
        self.urlProcess.urlObj = self.url
        self.urlProcess.url = elemUrl
        self.urlProcess.dbWrapper = self.dbWrapper
        self.urlProcess.siteId = localSiteId
        retUrl, retContinue = self.urlProcess.processURL(self.realUrl, internalLinks, externalLinks, self.filtersApply,
                                                         self.site.filters)
        if retUrl is not None:
          elemUrl = retUrl
        else:
          retContinue = True

        if retContinue:
          logger.debug(">>> Candidate URL is not passed general checks, skipped.")
          continue

        urlMd5 = hashlib.md5(elemUrl).hexdigest()
        self.urlProcess.url = elemUrl
        self.urlProcess.siteId = localSiteId
        self.urlProcess.urlTable = urlTable
        if self.urlProcess.isUrlExist(self.site.recrawlPeriod, self.siteProperties, urlMd5):
          logger.debug("Candidate URL %s already exist, skipped.", str(urlMd5))
          continue

        if not self.filtersApply(localFilters, elemUrl, self.url.depth, self.dbWrapper, localSiteId):
          logger.debug("Candidate URL not matched filter, skipped.")
          continue

        if self.site.maxURLs > 0:
          logger.debug("Site MaxURLs limit need to be checked.")
          currentCnt = self.urlProcess.readCurrentCnt(self.site.maxURLs)
          if currentCnt >= self.site.maxURLs:
            logger.debug("Site MaxURLs limit is reached.")
            autoremovedURLs = self.urlProcess.autoRemoveURL(self.autoRemoveProps, self.site.recrawlPeriod, urlTable,
                                                            self.dbWrapper)
            if autoremovedURLs == 0:
              logger.debug("No one URL auto removed, candidate URL skipped!")
              continue
            else:
              logger.debug("%s URLs auto removed.", str(autoremovedURLs))

        #detect collected url mime type and ignore non-match URL
        #(7) Detect collected url mime type and ignore non-match URL
        detectedMime = ''
        if self.autoDetectMime == self.DETECT_MIME_COLLECTED_URL and self.processContentTypes is not None:
          self.urlProcess.url = elemUrl
          detectedMime = self.urlProcess.detectUrlMime()
          if detectedMime not in self.processContentTypes:
            logger.debug("Candidate URL MIME type is not matched, skipped!")
            continue
          #detectedMime = self.urlProcess.detectUrlMime()
        #(7) END

        if self.robotsParser and self.robotsParser.loadRobots(elemUrl, self.batchItem.siteId, httpApplyHeaders):
          isAllowed, retUserAgent = self.robotsParser.checkUrlByRobots(elemUrl, self.batchItem.siteId,
                                                                       httpApplyHeaders)
          if not isAllowed:
            logger.debug(">>> URL " + elemUrl + " is NOT Allowed by user-agent:" + str(retUserAgent))
            continue

        numUrlsFromPage += 1

        self.urlProcess.siteId = localSiteId
        depth = self.urlProcess.getDepthFromUrl(self.batchItem.urlId)

        self.urlProcess.url = elemUrl
        self.urlProcess.siteId = localSiteId
        self.urlProcess.urlObj = self.url
        localUrlObj = self.urlProcess.createUrlObjForCollectURLs(urlMd5, formMethods, self.batchItem.urlId, depth,
                                                                 detectedMime, self.site.maxURLsFromPage)
        logger.debug("!!!self.processorName=\n%s, type(self.processorName)=%s", str(self.processorName),
                     str(type(self.processorName)))

        if self.processorName == PCONSTS.PROCESSOR_FEED_PARSER or self.processorName == PCONSTS.PROCESSOR_RSS:
          self.feedElementsProcessing(urlMd5, httpCode, elemUrl, localSiteId, localUrlObj, localUrl, params)
        else:
          params.append(localUrlObj)

    if useChains and "URL_CHAIN" in self.siteProperties and self.siteProperties["URL_CHAIN"] is not None:
      localChainDict = json.loads(self.siteProperties["URL_CHAIN"])
      depth = self.urlProcess.getDepthFromUrl(self.batchItem.urlId)
      if "url_pattern" in localChainDict:
        for elemUrl in urlSet:
          if elemUrl is None:
            logger.debug(">>> Some url from urlSet is None")
            continue
          logger.debug(">>> Chain eleme URL is = " + str(elemUrl))
          self.urlProcess.url = elemUrl
          retUrl = self.urlProcess.simpleURLCanonize(self.realUrl)
          if retUrl is None or not Utils.UrlNormalizator.isNormalUrl(retUrl):
            logger.debug(">>> Bad url normalization, url = " + retUrl)
            continue
          else:
            elemUrl = retUrl
          detectedMime = ''
          if self.autoDetectMime == self.DETECT_MIME_COLLECTED_URL and self.processContentTypes is not None:
            self.urlProcess.url = elemUrl
            detectedMime = self.urlProcess.detectUrlMime()
          urlMd5 = hashlib.md5(elemUrl).hexdigest()
          self.urlProcess.url = elemUrl
          self.urlProcess.siteId = localSiteId
          self.urlProcess.urlObj = self.url
          try:
            logger.debug(">>> Start chain processing, url = " + self.urlProcess.url)
            localUrlObj = self.urlProcess.createUrlObjForChain(localChainDict["url_pattern"],
                      urlMd5, formMethods, self.batchItem.urlId, depth, detectedMime, self.site.maxURLsFromPage)
            if localUrlObj is not None:
              chainUrls.append(copy.deepcopy(localUrlObj))
          except Exception as excp:
            logger.debug(">>> Some exception in URL_CHAIN deserialize, excp = " + str(excp))
    if len(urlSet) > 0 and len(params) == 0:
      logger.debug(">>> Wrong Collect URLs; 0 collected urls for LEN(urlSet) = " + str(len(urlSet)))
    logger.debug(">>> Chain URLs after collect are = " + str(chainUrls))
    if not readOnly:
      result = self.dbWrapper.urlNew(params)
      logger.debug(">>> Url New result: " + str(result))
      result = self.dbWrapper.urlNew(chainUrls)
      self.urlProcess.updateTypeForURLObjects(chainUrls)
      self.dbWrapper.fieldRecalculating(localSiteId)

    if formFields is not None and self.postForms is not None:
      fieldParams = self.getFieldParams(formFields, self.postForms, localSiteId)
      self.insertNewSiteProperties(fieldParams, self.dbWrapper, localSiteId)

    return nextStep, internalLinks, externalLinks, params, self.feedItems, chainUrls


  # # feedElementsProcessing processed rss element
  #
  # @param urlMd5 - element url's urlMd5
  # @param httpCode - http code of http response
  # @param elemUrl - element url's url
  # @param localSiteId - siteId
  # @param localUrlObj - element's urlObj
  # @param localUrl - root element's url
  # @param params - element container (List type)
  # @param rootFeed - bool param that specifyied use feed element from rss struture of entities element
  def feedElementsProcessing(self, urlMd5, httpCode, elemUrl, localSiteId, localUrlObj, localUrl, params,
                             rootFeed=False):
    if self.feed is not None:
      self.urlProcess.url = elemUrl
      self.urlProcess.siteId = localSiteId
      self.urlProcess.urlObj = localUrlObj
      localRet = self.urlProcess.fillRssFieldInUrlObj(localUrl, self.url.url, self.batchItem, self.processorName,
                                                      self.feed, rootFeed)
      self.urlProcess.urlObj = None
      if localRet is not None:
        localRet["urlMd5"] = urlMd5
        if localRet["urlObj"] is not None:
          localRet["urlObj"].httpCode = httpCode
          localRet["urlObj"].processingDelay = 0
          localRet["urlObj"].parentMd5 = self.url.urlMd5
        logger.debug("self.feedItems[urlMd5]=\n%s", varDump(localRet))
        params.append(localRet["urlObj"])
        self.feedItems.append(localRet)
    else:
      logger.debug(">>> Wrong - None feed object for rss processing")


  # # processProcessor processed element from url's dom
  #
  # @param urlSet - incoming url's list (links from page)
  # @param dom - page's dom model
  # @param urlXpathList - applying xpath
  # @param urlObj - element of urlObject
  def processProcessor(self, urlSet, dom, urlXpathList, urlObj):
    if (self.processorName == PCONSTS.PROCESSOR_FEED_PARSER or self.processorName == PCONSTS.PROCESSOR_RSS) \
    and urlObj.type != dc.EventObjects.URL.TYPE_FETCHED:
      if feedparser is not None:
        try:
          self.feedItems = []
          # Add one more date parsing handler function to fix some wrong datetime format cases; added by bgv
          feedparser.registerDateHandler(self.feedparserParseDateFixes)

          # Remove handlers to process all tags as unknown to save their names unchanged
          if not (self.siteProperties is not None and "RSS_FEEDPARSER_MODE" in self.siteProperties and\
                   self.siteProperties["RSS_FEEDPARSER_MODE"] > 0):
            import inspect
            #, "_start_guid"
            excludes = ["_start_rss", "_start_channel", "_start_feed", "_start_item", "_start_link",
                        "_start_admin_errorreportsto", "_start_admin_generatoragent", "_start_guid", "_start_id",
                        "_start_entry", "_start_enclosure"]
            for methodName, functionObject in inspect.getmembers(feedparser._FeedParserMixin, predicate=inspect.ismethod):  # pylint: disable=W0612,W0212,C0301
              if methodName.startswith("_start_") and methodName not in excludes:
                delattr(feedparser._FeedParserMixin, methodName)  # pylint: disable=W0212
                endMethodName = methodName.replace("_start_", "_end_")
                if hasattr(feedparser._FeedParserMixin, endMethodName):  # pylint: disable=W0212
                  delattr(feedparser._FeedParserMixin, endMethodName)  # pylint: disable=W0212
            setattr(feedparser._FeedParserMixin, "_normalize_attributes", self._normalize_attributes)  # pylint: disable=W0212
            feedparser.FeedParserDict.keymap["guid"] = "guid"
            logger.debug("Feedparser in modified mode")
          else:
            logger.debug("Feedparser in native mode")

          self.feed = feedparser.parse(self.crawledResource.html_content)
          urlSet.update(entry.link for entry in self.feed.entries)
          logger.debug("self.feed=\n%s", varDump(self.feed))
          logger.debug(">>> len(html_content) = " + str(len(self.crawledResource.html_content)))
        except TypeError as err:
          logger.debug("WRONG CONTENT FOR URL <" + str(urlObj.url) + "> not rss feed. " + str(err.message))
        except Exception as err:
          logger.debug("SOME ERROR WITH rss feed parse " + str(err.message))
      else:
        logger.debug("feedparser module not found")
    # won't collect urls from rss feed resources
    elif self.processorName != PCONSTS.PROCESSOR_RSS:
      # Added support of urlXpathList as site's properties
      if len(urlXpathList):
        logger.debug("Site has URL_XPATH_LIST property: %s", str(urlXpathList))
      else:
        urlXpathList = self.WITH_URL_XPATH_LIST
        logger.debug("Site hasn't URL_XPATH_LIST property. Use default xpath's' list.")
      for xpath in urlXpathList:
        elem = dom.xpath(xpath)
        elem_type = type(elem)
        if elem_type == list and len(elem) > 0 and hasattr(elem[0], "tail"):
          urlSet.update([el.tail for el in elem])
        elif elem_type == list and len(elem) > 0 and isinstance(elem[0], lxml.html.HtmlElement):
          urlSet.update([el.text for el in elem])
        else:
          urlSet.update(elem)
    logger.debug("urlSet: %s", str(urlSet))


  # # extrace URL, form action, and fields from html dom
  #
  # @param dom the dom tree
  # @return form_urls    sequence of urls extracted
  # @return form_methods  dict of form methods, {form_action: form_method}
  # @return form_fields  dict of fields {field_name: field_value}
  def extractFormURL(self, dom, siteProperties):
    formUrls, formMethods, formFields = [], {}, {}
    if self.COLLECT_POST_DATA_NAME in siteProperties and \
    siteProperties['COLLECT_POST_DATA'] == self.COLLECT_POST_DATA:
      for form in dom.xpath("//form"):
        formAction = None
        formMethod = 'get'
        for attr in form.keys():
          if attr.lower() == "action":
            formAction = form.get(attr)
            formUrls.append(formAction)
          elif attr.lower() == "method":
            formMethod = form.get(attr)
        if not formAction:
          continue
        formMethods[formAction] = formMethod
        for field in form.getchildren():
          tagName, tagValue = None, ""
          for fieldTag in field.keys():
            if fieldTag.lower() == "name":
              tagName = field.get(fieldTag)
            elif fieldTag.lower() == "value":
              tagValue = field.get(fieldTag)
          if tagName:
            formFields[tagName] = tagValue
      logger.info("extracted form data, formUrls:%s, formMethods:%s, formFields:%s", \
      formUrls, formMethods, formFields)
    return formUrls, formMethods, formFields


  # #filtersApply applies filters and returns bool result
  #
  def filtersApply(self, localFilters, localUrl, depth, wrapper, siteId, stage=Filters.STAGE_COLLECT_URLS):
    ret = False
    fValue = Utils.generateReplacementDict()
    fValue.update({"MAX_DEPTH": str(depth)})
    if localFilters is None:
      if self.site is not None and self.site.filters is not None and len(self.site.filters) > 0:
        localFilters = Filters(self.site.filters, wrapper, siteId)
      else:
        localFilters = Filters(None, wrapper, siteId)

    logger.debug(">>> url before filter = " + localUrl)
    fResult = localFilters.filterAll(stage, fValue, Filters.LOGIC_OR, localUrl)
    logger.debug(">>> filter result - " + str(fResult))

    for elem in fResult:
      if elem > 0:
        ret = True
        break
    return ret


  # #getFieldParams method fill and returns post forms
  #
  #formFields - post fields dict
  #postForms - post form list
  #siteId - site's id
  def getFieldParams(self, formFields, postForms, siteId):
    ret = []
    for fieldName, fieldValue in formFields.iteritems():
      if fieldName in postForms:
        continue
      logger.debug("field_name: %s", fieldName)
      logger.debug("field_value: %s", fieldValue)
      ret.append((siteId, "HTTP_POST_FORM_" + fieldName, fieldValue))
    return ret


  # #insertNewSiteProperties update Site (add new properties in correspond table)
  #
  #params - params list
  #wrapper - db-task wrapper
  #siteId - site's id
  def insertNewSiteProperties(self, params, wrapper, siteId):
    if siteId is not None and hasattr(params, '__iter__') and len(params) > 0:
      localSiteUpdate = dc.EventObjects.SiteUpdate(siteId)
      for attr in localSiteUpdate.__dict__:
        if hasattr(localSiteUpdate, attr):
          setattr(localSiteUpdate, attr, None)
      localSiteUpdate.updateType = dc.EventObjects.SiteUpdate.UPDATE_TYPE_APPEND
      localSiteUpdate.id = siteId
      localSiteUpdate.properties = []
      for param in params:
        newPropElem = {}
        newPropElem["siteId"] = param[0]
        newPropElem["name"] = param[1]
        newPropElem["value"] = param[2]
        localSiteUpdate.properties.append(newPropElem)
      wrapper.siteNewOrUpdate(localSiteUpdate, stype=dc.EventObjects.SiteUpdate)


  # #feedparserParseDateFixes method to fix date parsing for the feedparser
  #
  # @param aDateString the date string to parse
  def feedparserParseDateFixes(self, aDateString):
    ret = None
    ds = aDateString

    # Assumes that date format broken and contains the semicolon ":" in TZ like: "Wed, 19 Aug 2015 08:45:53 +01:00"
    parts = ds.split(' ')
    if ("+" in parts[len(parts) - 1] or "-" in parts[len(parts) - 1]) and ":" in parts[len(parts) - 1]:
      parts[len(parts) - 1] = parts[len(parts) - 1].replace(":", "")
      ds = " ".join(parts)
      #ret = feedparser._parse_date_rfc822(ds)
      ret = feedparser._parse_date(ds)  # pylint: disable=W0212

    return ret


  def _normalize_attributes(self, kv):
    return (kv[0], kv[1])

