"""
HCE project, Python bindings, Distributed Tasks Manager application.
web page fetchers.

@package: dc
@file Fetcher.py
@author madk, bgv <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-201 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

# #The BaseFetcher class, defines the interface of fetchers
# #this class don't implement any fetcher
# concrete fetcher class should extends from this class
#

import logging
import requests
import requests.exceptions
import re
import time
import ctypes
import os
import app.Consts as APP_CONSTS
import dc_crawler.Constants as CONSTS
import base64
import app.Utils as Utils
from app.Utils import varDump
from app.Filters import Filters
from requests.auth import HTTPBasicAuth
from app.Exceptions import SeleniumFetcherException

logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)

#max content size for use chardet to detect charset
MAX_CONTENT_SIZE_FOR_CHARDET = 5000000


class BaseFetcher(object):

  fetchers = None

  TYP_NORMAL = 1
  TYP_DYNAMIC = 2
  TYP_EXTERNAL = 3
  TYP_SPECIAL_DYNAMIC = 4
  TYP_URLLIB = 5
  TYP_CONTENT = 6
  TYP_AUTO = 7

  @staticmethod
  def init():
    #enumerate content_types we don't want to fetch
    BaseFetcher.prohibited_conten_types = ["audio/mpeg", "application/pdf"]

    BaseFetcher.fetchers = {
      BaseFetcher.TYP_NORMAL : RequestsFetcher(),
      #BaseFetcher.TYP_DYNAMIC: GhostpyFetcher(),
      BaseFetcher.TYP_DYNAMIC: SeleniumFetcher(),
      BaseFetcher.TYP_URLLIB: URLLibFetcher(),
      BaseFetcher.TYP_EXTERNAL: ExternalFetcher(),
      BaseFetcher.TYP_CONTENT: ContentFetcher()
    }

  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting, tuple of proxy_type, host, port, username, password
  # @param auth, basic auth setting, tuple of username and password
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
    url,
    external_url=None,
    method='get',
    headers={},
    timeout=100,
    allow_redirects=True,
    proxies=None,
    auth=None,
    data=None,
    logger=None,
    allowed_content_types=None,
    max_resource_size=None,
    max_redirects=CONSTS.MAX_HTTP_REDIRECTS_LIMIT):
    pass


  # #get fetched by fetch type
  #
  # @param typ, the fetch type
  # @return fetcher
  @staticmethod
  def get_fetcher(typ):
    if not BaseFetcher.fetchers:
      BaseFetcher.init()
    if typ in BaseFetcher.fetchers:
      return BaseFetcher.fetchers[typ]
    else:
      raise BaseException("unsupported fetch type:%s" % (typ,))


  # #check whether the fetcher have meta resource
  #
  # @return whether the fetcher have meta resource
  def should_have_meta_res(self):

    return False



# # Fetcher base on the requests module, cann't execute javascript
#
#
class RequestsFetcher(BaseFetcher):

  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting, tuple of name and password
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
    url,
    external_url=None,
    method='get',
    headers=None,
    timeout=100,
    allow_redirects=True,
    proxies=None,
    auth=None,
    data=None,
    logger=None,
    allowed_content_types=None,
    max_resource_size=None,
    max_redirects=CONSTS.MAX_HTTP_REDIRECTS_LIMIT,
    filters=None,
    depth=None):

    if auth:
      auth = HTTPBasicAuth(auth[0], auth[1])
    requestsCallbackWrapper = None
    proxy_setting = None
    if proxies is not None:
      proxy_type, proxy_host, proxy_port, proxy_user, proxy_passwd = proxies
      if proxy_type is None:
        proxy_type = "http"
      if proxy_user is not None:
        proxies = "%s://%s:%s@%s:%s" % (proxy_type, proxy_user, proxy_passwd, proxy_host, proxy_port)
      else:
        proxies = "%s://%s:%s" % (proxy_type, proxy_host, proxy_port)
      proxy_setting = {"http" : proxies}

    rSession = requests.Session()
    if filters is not None:
      requestsCallbackWrapper = RequestsCallbackWrapper(filters, depth)
      #rSession.resolve_redirects = requestsCallbackWrapper.resolveRedirects
      allow_redirects = False
    rSession.max_redirects = int(max_redirects)
    method_func = rSession.__getattribute__(method)
    logger.debug(">>> max_resource_size: %s", str(max_resource_size))
    logger.debug(">>> proxy_setting: " + str(proxy_setting))
    impl_res = None
    res = Response()
    try:
      impl_res = method_func(url,
        timeout=timeout,
        headers=headers,
        allow_redirects=allow_redirects,
        proxies=proxy_setting,
        auth=auth,
        data=data,
        stream=True)
      if filters is not None:
        i = 1
        logger.debug("Enter redirects loop, max_redirects=%s", str(max_redirects))
        for redirect in rSession.resolve_redirects(impl_res, impl_res.request):
          logger.debug("R1 i=%s, redirect:\n%s, impl_res:\n%s", str(i), varDump(redirect), varDump(impl_res))
          i += 1
          if i >= int(max_redirects):
            raise requests.exceptions.TooManyRedirects(">>> Max redirects = " + str(i))
          elif requestsCallbackWrapper is not None:
            requestsCallbackWrapper.resolveRedirects(redirect.url)
          logger.debug(">>> redirect.url = " + redirect.url)
        if i > 1:
          logger.debug("R2 i=%s, redirect:\n%s, impl_res:\n%s", str(i), varDump(redirect), varDump(impl_res))
          impl_res = redirect

      #try to prevent huge content fetching
      if "content-length" in impl_res.headers and \
        max_resource_size != CONSTS.MAX_HTTP_SIZE_UNLIMIT and \
        int(impl_res.headers['content-length']) > max_resource_size:
        logger.debug("Content size overshooted. content-length: %s, max_resource_size: %s" % \
                     (str(impl_res.headers['content-length']), str(max_resource_size)))
        res.content_size = int(impl_res.headers['content-length'])
        res.headers = impl_res.headers
        res.redirects = impl_res.history
        res.status_code = impl_res.status_code
        res.url = impl_res.url
      else:
        #don't detect charset for binary content type or BIG response
        if impl_res.headers.get('content-type', '').startswith('application') or  \
          impl_res.headers.get('content-type', '').startswith('audio') or  \
          len(impl_res.content) >= MAX_CONTENT_SIZE_FOR_CHARDET:
          detected_encoding = impl_res.encoding
          logger.debug("Headers contains 'application' or 'audio' content-type: %s",
                       impl_res.headers.get('content-type', ''))
        else:
          # use chardet to improve encoding detect
          ct = impl_res.headers.get('content-type', '').lower()
          logger.debug("impl_res.encoding1=%s, content-type=%s", impl_res.encoding, ct)
          #Try simple way of charset detection for an html
          encoding = None
          if "html" in ct:
            logger.debug("Using the SimpleCharsetDetector()")
            encoding = SimpleCharsetDetector().detect(impl_res.content)
            logger.debug("encoding=%s", str(encoding))
            if encoding is not None:
              impl_res.encoding = encoding
          if (impl_res.encoding is None) or ((encoding is None) and (impl_res.encoding not in ct and "xml" not in ct)):
            logger.debug("Using the chardet to improve encoding detect")
            detected_encoding = impl_res.apparent_encoding
            if detected_encoding != 'ascii' and detected_encoding != 'ISO-8859-2':
              impl_res.encoding = detected_encoding
          logger.debug("impl_res.encoding2=%s", impl_res.encoding)
        res.encoding = impl_res.encoding
        res.url = impl_res.url
        res.status_code = impl_res.status_code
        res.redirects = impl_res.history
        res.request = impl_res.request
        #Fix for pages that has xml document tag but no html structure inside
        text_buffer = self.fixWrongXMLHeader(impl_res.content)
        if impl_res.headers.get('content-type', '').startswith('application'):
          res.unicode_content = impl_res.content
        else:
          res.unicode_content = text_buffer
        res.str_content = impl_res.content
        if impl_res.headers.get('content-type', '').startswith('application'):
          res.rendered_unicode_content = impl_res.content
        else:
          res.rendered_unicode_content = text_buffer

        #res.content_size = impl_res.raw.tell()
        if res.rendered_unicode_content is None:
          res.content_size = 0
        else:
          res.content_size = len(res.rendered_unicode_content)
        res.headers = impl_res.headers
        res.cookies = requests.utils.dict_from_cookiejar(impl_res.cookies)
    except requests.exceptions.ConnectionError, err:
      res.error_mask = APP_CONSTS.ERROR_FETCH_CONNECTION_ERROR
      logger.debug(">>> Requests fetcher has thrown ConnectionError exception: " + str(err.message))
    except requests.exceptions.HTTPError, err:
      res.error_mask = APP_CONSTS.ERROR_FETCH_HTTP_ERROR
      logger.debug(">>> Requests fetcher has thrown HTTPError exception: " + str(err.message))
    except requests.exceptions.URLRequired, err:
      res.error_mask = APP_CONSTS.ERROR_FETCH_INVALID_URL
      logger.debug(">>> Requests fetcher has thrown URLRequired exception: " + str(err.message))
    except requests.exceptions.TooManyRedirects, err:
      res.error_mask = APP_CONSTS.ERROR_FETCH_TOO_MANY_REDIRECTS
      logger.debug(">>> Requests fetcher has thrown TooManyRedirects exception: " + str(err.message))
    except requests.exceptions.Timeout, err:
      res.error_mask = APP_CONSTS.ERROR_FETCH_TIMEOUT
      logger.debug(">>> Requests fetcher has thrown Timeout exception: " + str(err.message))
    except requests.exceptions.RequestException, err:
      res.error_mask = APP_CONSTS.ERROR_FETCH_AMBIGUOUS_REQUEST
      logger.debug(">>> Requests fetcher has thrown RequestException exception: " + str(err.message))
    except RedirectException, err:
      res.error_mask = APP_CONSTS.ERROR_FETCH_TOO_MANY_REDIRECTS | APP_CONSTS.ERROR_CRAWLER_FILTERS_BREAK
      logger.debug(">>> Redirect filters not allowed: " + str(err.message))
    except Exception, err:
      res.error_mask = APP_CONSTS.ERROR_FETCHER_INTERNAL
      logger.debug(">>> Requests fetcher has thrown exception: " + str(err.message))
    return res



  # #Fix the wrong html document structure in case of XML header is defined first
  #
  # @param content input content
  # @return fixed contentStr
  def fixWrongXMLHeader(self, contentStr):
    #text_file = open("/tmp/fetcher_log.txt", "w")
    #text_file.write("Fetcher: start\n")

    if contentStr.startswith('<?xml ') and '<html' in contentStr and '<head' in contentStr:
      #text_file.write("Fetcher: xml detected!\n")
      p = re.compile(r'<\?xml .*\?>')
      contentStr = p.sub('', contentStr, count=1)
      #text_file.write(contentStr)

    #text_file.close()

    return contentStr



# # Fetcher base on the ghost.py module, which can execute javascript
#
#
class GhostpyFetcher(BaseFetcher):
  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
    url,
    external_url=None,
    method='get',
    headers=None,
    timeout=100,
    allow_redirects=True,
    proxies=None,
    auth=None,
    data=None,
    logger=None,
    allowed_content_types=None,
    max_resource_size=None,
    max_redirects=1,
    filters=None):


    if 'Accept-Encoding' in headers:
      if "gzip" in headers['Accept-Encoding']:
        if logger:
          logger.warn("GhostpyFetcher don't support GZIP, Accept-Encoding request header was ignored")
        del headers['Accept-Encoding']
    from ghost import Ghost  # pylint: disable-all
    agent = headers.get('User-Agent', '')
    browser = Ghost(user_agent=agent)
    if proxies:
      browser.set_proxy(proxies[0], proxies[1], int(proxies[2]), proxies[3], proxies[4])

    try:
      page_res, _ = browser.open(url, method,
        headers=headers,
        auth=auth)
    except ghost.TimeoutError:  # pylint: disable-all
      raise requests.exceptions.Timeout()
    res = Response()
    res.url = page_res.url
    res.status_code = page_res.http_status
    #TODO get all redirect histories
    res.redirects = []
    res.unicode_content = browser.content
    charset, _ = browser.evaluate("document.characterSet")
    if charset:
      charset = str(charset)
    if logger:
      logger.info("got page charset %s by execute javascript", charset)
    if not charset:
      charset = "utf-8"
    res.encoding = charset
    res.str_content = browser.content.encode(res.encoding, res.encoding)

    res.rendered_unicode_content = browser.content
    res.content_size = len(res.str_content)
    res.headers = page_res.headers
    res.cookies = {}
    for cookie in browser.cookies:
      res.cookies[str(cookie.name())] = str(cookie.value())
    try:
      if isinstance(page_res.content, unicode):
        res.meta_res = page_res.content
      else:
        res.meta_res = unicode(page_res.content, charset)
    except BaseException:
      if logger:
        logger.warn("error when get meta resource", exc_info=True)
    if res.meta_res == u'None':
      if logger:
        logger.warn("got None from page response, use rendered content")
      res.meta_res = res.rendered_unicode_content
    res.dynamic_fetcher_type = "ghost"

    return res


  # #check whether the fetcher have meta resource
  #
  # @return whether the fetcher have meta resource
  def should_have_meta_res(self):

    return True



# # Fetcher based on the selenium project web-driver
#
#
class SeleniumFetcher(BaseFetcher):

  DEFAUIL_TIMEOUT = 5

  ERROR_FATAL = 1
  ERROR_GENERAL = 2
  ERROR_CONTENT_OR_COOKIE = 3
  ERROR_NAME_NOT_RESOLVED = 10
  ERROR_TOO_MANY_REDIRECTS = 11

  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting
  # @param data, post data, used only when method is post
  # @param logger
  # @param allowed_content_types
  # @param max_resource_size
  # @param max_redirects
  # @param executable_path path and file name of the driver binary executable
  # @return Response object
  def open(self,
    url,
    external_url=None,
    method='get',
    headers=None,
    timeout=DEFAUIL_TIMEOUT,
    allow_redirects=True,
    proxies=None,
    auth=None,
    data=None,
    logger=None,
    allowed_content_types=None,
    max_resource_size=None,
    max_redirects=1,
    filters=None,
    executable_path=None,
    depth=None):

    if logger is not None:
      logger.debug("Dynamic fetcher call:\nurl:" + str(url) + "\nexternal_url:" + str(external_url) + "\nmethod:" + \
                   str(method) + "\nheaders:" + str(headers) + "\ntimeout:" + str(timeout) + "\nallow_redirects:" + \
                   str(allow_redirects) + "\nproxies:" + str(proxies) + "\nauth:" + str(auth) + "\ndata:" + str(data) + \
                   "\nlogger:" + str(logger) + "\nallowed_content_types:" + str(allowed_content_types) + \
                   "\nmax_resource_size:" + str(max_resource_size) + "\nmax_redirects:" + str(max_redirects) + \
                   "\nexecutable_path:" + str(executable_path) + \
                   "\ncur_dir:" + str(os.getcwd()))

    #Dependent import
    try:
      from selenium import webdriver
      import selenium.webdriver.support.ui
    except Exception, err:
      raise SeleniumFetcherException(err)

    #Initialize defaults
    #exec_path = "../../bin/"
    exec_path = "./"
    driver_name = "chromedriver"
    error_msg = ""
    error_code = 0
    driver = None

    #TODO: experimental for the OpenVZ containers
    disable_setuid_sandbox = "--disable-setuid-sandbox"

    #Check environment
    envVars = {"DISPLAY": "", "LC_ALL":"en_US.UTF-8", "LANG":"en_US.UTF-8", "LANGUAGE":"en_US.UTF-8"}
    for varName in envVars:
      if varName == "DISPLAY":
        if os.getenv(varName, "") == "":
          raise SeleniumFetcherException("Environment variable 'DISPLAY' is not set proper way!")
      else:
        if os.getenv(varName, "") != envVars[varName]:
          raise SeleniumFetcherException("Environment variable '" + varName + "' is not set proper way as '" + \
                                           envVars[varName] + "' !")

    #Define the proxy options
    if proxies is not None:
      chrome_option = webdriver.ChromeOptions()
      proxy_type, proxy_host, proxy_port, proxy_user, proxy_passwd = proxies
      if proxy_user:
        proxies = proxy_type + "://%s:%s@%s:%s" % (proxy_user, proxy_passwd, proxy_host, proxy_port)
      else:
        proxies = proxy_type + "://%s:%s" % (proxy_host, proxy_port)
      chrome_option.add_argument("--proxy-server=" + proxies)
      if disable_setuid_sandbox != "":
        chrome_option.add_argument(disable_setuid_sandbox)
    else:
      if disable_setuid_sandbox != "":
        chrome_option = webdriver.ChromeOptions()
        chrome_option.add_argument(disable_setuid_sandbox)
      else:
        chrome_option = None

    #Create driver instance
    try:
      #The platform-dependent path to the driver executable
      if executable_path is None:
        path = exec_path + driver_name + str(ctypes.sizeof(ctypes.c_voidp) * 8)
      else:
        path = executable_path
      #Get driver
      driver = webdriver.Chrome(executable_path=path, chrome_options=chrome_option)
    except Exception, err:
      error_msg = str(err)
      error_code = self.ERROR_FATAL
    except:
      error_msg = "General driver initialization!"
      error_code = self.ERROR_GENERAL

    if error_code > 0:
      if driver is not None:
        driver.quit()
      if logger is not None:
        logger.error("Fatal error: " + error_msg)
      raise SeleniumFetcherException(error_msg)

    #Make request
    try:
      driver.set_page_load_timeout(timeout)
      driver.get(url)
      #Get logs
      log_types = driver.log_types
      if 'browser' in log_types:
        log_list = driver.get_log('browser')
        if logger is not None:
          logger.debug("Dynamic fetcher logs: " + str(log_list))
        for item_dict in log_list:
          if "message" in item_dict and item_dict["message"] != '' and url in item_dict["message"]:
            error_msg += item_dict["message"] + "\n"
      if error_msg != "":
        entrances = [
                     (r"(.*)net::ERR_NAME_NOT_RESOLVED(.*)", self.ERROR_NAME_NOT_RESOLVED),
                     (r"(.*)net::ERR_TOO_MANY_REDIRECTS(.*)", self.ERROR_TOO_MANY_REDIRECTS),
                     (r"(.*)403 \(Forbidden\)(.*)", 403),
                     (r"(.*)404 \(Not Found\)(.*)", 404),
                     (r"(.*)500 \(Internal Server Error\)(.*)", 500),
                     (r"(.*)net::(.*)", 520)]
        for item in entrances:
          regex = re.compile(item[0])
          r = regex.search(error_msg)
          if r:
            error_code = item[1]
            if logger is not None:
              logger.debug("Page error: " + error_msg)
            break
    except Exception, err:
      error_msg = str(err)
      error_code = self.ERROR_FATAL
    except:
      error_msg = "General driver usage error!"
      error_code = self.ERROR_GENERAL

    fatalErrors = [self.ERROR_FATAL, self.ERROR_GENERAL, self.ERROR_NAME_NOT_RESOLVED, self.ERROR_TOO_MANY_REDIRECTS]
    if error_code == 0:
      #Wait fixed time to load all dynamic parts of the page
      time.sleep(timeout)
    elif error_code in fatalErrors:
      if driver is not None:
        driver.quit()
        if logger is not None:
          logger.debug("Fatal error %s: %s", str(error_code), error_msg)
      if error_code == self.ERROR_NAME_NOT_RESOLVED:
        code = APP_CONSTS.ERROR_FETCH_INVALID_URL
      elif error_code == self.ERROR_TOO_MANY_REDIRECTS:
        code = APP_CONSTS.ERROR_FETCH_TOO_MANY_REDIRECTS
      else:
        code = APP_CONSTS.ERROR_FETCHER_INTERNAL
      raise SeleniumFetcherException(error_msg, code)

    page_source = ""
    cookies = {}
    try:
      page_source = driver.page_source
      cookies = driver.get_cookies()
    except Exception, err:
      error_msg = str(err)
      error_code = self.ERROR_CONTENT_OR_COOKIE
    except:
      error_msg = "Content and cookies get error!"
      error_code = self.ERROR_CONTENT_OR_COOKIE

    content_type = None
    charset = None
    try:
      attr = driver.find_element_by_xpath(".//meta[translate(@http-equiv,'ABCDEFGHIJKLMNOPQRSTUVWXYZ','abcdefghijklmnopqrstuvwxyz')='content-type']").get_attribute("content")
      regex = re.compile(r"(.*); charset=(.*)", re.IGNORECASE)
      items = regex.search(attr)
      if items is not None:
        items = items.groups()
        if len(items) > 1:
          content_type = items[0]
          charset = items[1]
    except Exception, err:
      pass
    if content_type is None:
      try:
        attr = driver.find_element_by_xpath('//html')
        content_type = "text/html"
      except Exception, err:
        pass
    if content_type is not None and charset is None:
      try:
        charset = driver.find_element_by_xpath('//meta[@charset]').get_attribute("charset")
      except Exception, err:
        pass
    if charset is None:
      try:
        charset = driver.execute_script("return document.characterSet;")
      except Exception, err:
        print str(err)
    current_url = driver.current_url

    res = Response()
    res.url = current_url
    if error_code > 100:
      res.status_code = error_code
    else:
      res.status_code = 200
    res.redirects = []
    res.unicode_content = page_source
    res.str_content = res.unicode_content
    res.rendered_unicode_content = res.unicode_content
    res.content_size = len(res.unicode_content)
    res.encoding = charset
    res.headers = {"content-length": res.content_size, "content-type": content_type}
    res.meta_res = res.unicode_content
    res.cookies = cookies
    res.dynamic_fetcher_type = driver_name
    #TODO:possible set some error bits
    #res.error_mask = CONSTS.ERROR_OK

    if driver is not None:
      driver.quit()

    if logger is not None and error_msg != "":
      logger.debug("Dynamic fetcher none fatal error: " + error_msg)

    return res


  # #check whether the fetcher have meta resource
  #
  # @return whether the fetcher have meta resource
  def should_have_meta_res(self):

    return True



# # external Fetcher
#
#
class ExternalFetcher(BaseFetcher):
  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param external_url, the external url template
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
           url,
           external_url=None, **kwargs):
    if external_url:
      url = external_url.replace('%url%', url)

    kwargs['method'] = 'get'

    return BaseFetcher.get_fetcher(BaseFetcher.TYP_NORMAL).open(url, **kwargs)



# # urllib Fetcher
#
#
class URLLibFetcher(BaseFetcher):
  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param external_url, the external url template
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self, url, external_url=None, **kwargs):
    if external_url:
      url = external_url.replace('%url%', url)

    import urllib2
    import os

    logger = kwargs['logger']
    allowed_content_types = kwargs['allowed_content_types']
    max_resource_size = kwargs["max_resource_size"]

    res = Response()
    logger.debug("url: <%s>" % url)
    response = None
    try:
      response = urllib2.urlopen(url)
      headers_info = response.info()
      if headers_info is not None:
        if headers_info.type in allowed_content_types:
          if response is not None:
            #res.encoding = impl_res.encoding
            #res.cookies = requests.utils.dict_from_cookiejar(impl_res.cookies)
            res.url = response.geturl()
            res.status_code = response.getcode()
            content_response = response.read()
            res.unicode_content = content_response
            res.str_content = content_response
            res.rendered_unicode_content = content_response
            res.content_size = len(content_response)
            headers = {}
            headers["content-length"] = res.content_size
            headers["content-type"] = headers_info.type
            res.headers = headers
            history = []
            res.redirects = history
          else:
            logger.debug("URLLib return empty response.")
        else:
          logger.debug("Content-Type not allowed. headers_info.type: %s" % str(headers_info.type))
      else:
        logger.debug("URLLib info is empty.")
    except urllib2.HTTPError, err:
    #except Exception, err:
      logger.debug("Exception <%s>", str(err.code))

    return res



# # external Fetcher
#
#
class ContentFetcher(BaseFetcher):
  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param external_url, the external url template
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self, url, external_url=None, **kwargs):
    try:
      localBuf = base64.b64decode(kwargs["inputContent"])
    except TypeError:
      localBuf = kwargs["inputContent"]
    res = Response()
    res.content_size = len(localBuf)
    res.headers = {}
    res.redirects = []
    res.status_code = 200
    res.url = url
    res.encoding = SimpleCharsetDetector().detect(localBuf)
    if res.encoding is None:
      res.encoding = "utf-8"
    res.unicode_content = localBuf
    res.str_content = localBuf
    res.rendered_unicode_content = localBuf

    return res



# #The Response class
# represents an web page response
class Response(object):
  def __init__(self):
    #final url
    self.url = None
    # http status code
    self.status_code = None
    # redirect lists
    self.redirects = None
    # unicode content
    self.unicode_content = None
    # str content
    self.str_content = None
    # rendered(by command line browser) unicode content
    self.rendered_unicode_content = None
    # http response content size
    self.content_size = None
    # content encoding
    self.encoding = None
    # headers
    self.headers = None
    #meta resource
    self.meta_res = None
    #cookies
    self.cookies = None
    #dynamic fetcher type
    self.dynamic_fetcher_type = None
    #error mask from fetcher
    self.error_mask = APP_CONSTS.ERROR_OK
    #request
    self.request = None



# #The Response class
# represents an web page response
class SimpleCharsetDetector(object):


  def __init__(self, content=None):
    #content
    self.content = content

  def detect(self, content=None):
    ret = None

    try:
      if content is None:
        cnt = self.content
      else:
        cnt = content

      pattern = r'<meta(?!\s*(?:name|value)\s*=)(?:[^>]*?content\s*=[\s"\']*)?([^>]*?)[\s"\';]*charset\s*=[\s"\']*([^\s"\'/>]*)'
      matchObj = re.search(pattern, cnt, re.I | re.M | re.S)
      if matchObj:
        ret = matchObj.group(2)

    except Exception, err:
      del err

    return ret



# #The RedirectException exception class
#class have used in the RequestsCallbackWrapper
class RedirectException(Exception):
  def __init__(self, message):
    Exception.__init__(self, message)



# #The RequestsCallbackWrapper class
# represents callback method for requests library
class RequestsCallbackWrapper(object):


  def __init__(self, filters, depth=None):
    self.filters = filters
    self.depth = depth


  # #resolveRedirects method represent callback for requsts library
  # resolveRedirects must be an generator with resp param in yield
  def resolveRedirects(self, url):
    if self.filters is not None:
      ret = False
      fValue = Utils.generateReplacementDict()
      #TODO: Add support of the current depth value from the batch item
      if self.depth is None:
        fValue.update({"MAX_DEPTH": '1'})
      else:
        fValue.update({"MAX_DEPTH": str(self.depth)})
      logger.debug(">>> Redirect url before filter = " + url + ", filters:" + varDump(self.filters))
      fResult = self.filters.filterAll(Filters.STAGE_COLLECT_URLS, fValue, Filters.LOGIC_OR, url)
      logger.debug(">>> Redirect filter result - " + str(fResult))

      if self.filters.searchFiltersWithStage(Filters.STAGE_COLLECT_URLS) > 0:
        for elem in fResult:
          if elem > 0:
            ret = True
            break
      else:
        ret = True

      if not ret:
        raise RedirectException(">>> Filters not applied for redirect url = " + url)

