"""
@package: dc
@file ProxyResolver.py
@author Scorp <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import logging
import json
import app.Consts as APP_CONSTS
import sys
import time
import copy
import dc.EventObjects
from app.Utils import ExceptionLog
from app.Utils import UrlParser

logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)


## ProxyResolver Class, implements http proxy resolving functionality
#
class ProxyResolver(object):

  SOURCE_PROPERTY = 0
  SOURCE_SQL = 1
  PROXY_SQL_QUERY = ("SELECT * FROM `sites_proxy` WHERE `State` = 1 AND (`Site_Id` = '%s' OR `Site_Id` = '*')" +
                     " ORDER BY `Priority`")
  PROXY_SQL_DB = "dc_sites"

  LIMITS = ["MIN", "HOUR", "DAY", "MONTH", "YEAR"]
  SECONDS_MULTI = [60, 3600, 86400, 2590000, 31100000]
  DEFAULT_PORT = 80

  JSON_SUFF = ".json"


  # #Class constructor
  #
  # @param siteProperties - incomig sites property dict
  # @param dbWrapper - sql db acces wrapper
  # @param siteId - current site id
  # @param url - current resource url
  def __init__(self, siteProperties, dbWrapper, siteId, url=None):
    self.proxyTuple = None
    self.proxyStruct = None
    self.internalIndexes = {}
    self.domain = UrlParser.getDomain(url) if url is not None else None
    self.indexFileName = None
    proxyJson = None
    try:
      if siteProperties is not None:
        if "HTTP_PROXY_HOST" in siteProperties and "HTTP_PROXY_PORT" in siteProperties:
          self.proxyTuple = (siteProperties["HTTP_PROXY_HOST"], siteProperties["HTTP_PROXY_PORT"])
        elif "USER_PROXY" in siteProperties:
          try:
            proxyJson = json.loads(siteProperties["USER_PROXY"])
          except Exception as excp:
            ExceptionLog.handler(logger, excp, ">>> Bad json in USER_PROXY property, ")

      if self.proxyTuple is None and proxyJson is not None:
        if "file_path" in proxyJson:
          self.indexFileName = proxyJson["file_path"]
          if siteId is not None and not (self.indexFileName.find(self.JSON_SUFF) != -1 and \
          self.indexFileName.find(self.JSON_SUFF) == len(self.indexFileName) - len(self.JSON_SUFF)):
            if self.indexFileName[-1] != '/':
              self.indexFileName += '/'
            self.indexFileName += siteId
            self.indexFileName += self.JSON_SUFF

          self.internalIndexes = self.readIndexFile(self.indexFileName)
        if int(proxyJson["source"]) == self.SOURCE_PROPERTY:
          self.proxyStruct = proxyJson["proxies"]
        elif int(proxyJson["source"]) == self.SOURCE_SQL:
          self.proxyStruct = self.readSQLProxy(dbWrapper, siteId)

        if self.proxyStruct is not None:
          self.updateProxyByFreqAndLimits()
    except Exception as excp:
      ExceptionLog.handler(logger, excp, ">>> ProxyResolver exception")


  # #Method readIndexFile reads from file and return param index structure
  #
  # @param fileName - incoming file name
  # @return json structure, just readed from file
  def readIndexFile(self, fileName):
    ret = {}
    fd = None
    try:
      fd = open(fileName, "r")
      if fd is not None:
        ret = json.loads(fd.read())
        fd.close()
    except Exception as excp:
      ExceptionLog.handler(logger, excp, ">>> readIndexFile error, file name = " + str(fileName))
      if fd is not None:
        fd.close()
    return ret


  # #Method saveIndexInFile saves indexes structute into the file
  #
  # @param fileName - incoming file name
  def saveIndexInFile(self, fileName):
    if self.internalIndexes is not None and len(self.internalIndexes) > 0:
      fd = None
      try:
        fd = open(fileName, "w")
        if fd is not None:
          fd.write(json.dumps(self.internalIndexes))
          fd.close()
      except Exception as excp:
        ExceptionLog.handler(logger, excp, ">>> saveIndexInFile error, file name = " + str(fileName))
        if fd is not None:
          fd.close()


  # #Method fieldsToLower return dict, which is copy of elem with lowercase keys
  #
  # @param elem incoming dict
  # @return lowcase dict
  def fieldsToLower(self, elem):
    ret = {}
    for key in elem:
      ret[key.lower()] = elem[key]
    return ret


  # #Method readSQLProxy, reads proxy data from SQL storage
  #
  # @param dbWrapper - sql db acces wrapper
  # @param siteId - current site id
  # @return just fetched proxy data
  def readSQLProxy(self, dbWrapper, siteId):
    ret = {}
    saveDBMode = dbWrapper.affect_db
    dbWrapper.affect_db = True
    result = dbWrapper.customRequest(self.PROXY_SQL_QUERY % siteId, self.PROXY_SQL_DB, 
                                     dc.EventObjects.CustomRequest.SQL_BY_NAME)
    dbWrapper.affect_db = saveDBMode
    if result is not None and len(result) > 0:
      for elem in result:
        elemInLower = self.fieldsToLower(elem)
        ret[elemInLower["host"]] = elemInLower
        #ret[elem["host"]]["freq"] = 0
        if ret[elemInLower["host"]]["limits"] is not None:
          try:
            ret[elemInLower["host"]]["limits"] = json.loads(ret[elemInLower["host"]]["limits"])
            ret[elemInLower["host"]]["domains"] = json.loads(ret[elemInLower["host"]]["domains"])
          except Exception as excp:
            ExceptionLog.handler(logger, excp, ">>> Wrong json in elem, host = " +
                                 (elem["host"] if "host" in elem else "None"))
    return ret


  # #Method updateProxyByFreqAndLimits, updates (if available) freq statistic in common proxy structure
  #
  def updateProxyByFreqAndLimits(self):
    for key in self.proxyStruct:
      if key in self.internalIndexes:
        self.proxyStruct[key]["freq"] = self.internalIndexes[key]["freq"]
        if "limits_stat" in self.internalIndexes[key]:
          self.proxyStruct[key]["limits_stat"] = copy.deepcopy(self.internalIndexes[key]["limits_stat"])
        else:
          self.proxyStruct[key]["limits_stat"] = {}
      else:
        self.proxyStruct[key]["freq"] = sys.maxint
        self.proxyStruct[key]["limits_stat"] = {}


  # #Method checkLimits, checks freq limits for current proxy
  #
  # @param key - proxy key
  # @return bool value - is available proxy by limits or not
  def checkLimits(self, key):
    ret = True
    if self.proxyStruct[key]["limits"] is not None and len(self.proxyStruct[key]["limits_stat"]) > 0:
      index = 0
      curTimeStamp = int(time.time())
      for limit in self.proxyStruct[key]["limits"]:
        if index >= len(self.LIMITS):
          break
        if self.LIMITS[index] + "_START_POINT" in self.proxyStruct[key]["limits_stat"]:
          if curTimeStamp - self.proxyStruct[key]["limits_stat"][self.LIMITS[index] + "_START_POINT"] >= \
          self.SECONDS_MULTI[index]:
            self.proxyStruct[key]["limits_stat"][self.LIMITS[index] + "_START_POINT"] = curTimeStamp
            self.proxyStruct[key]["limits_stat"][self.LIMITS[index] + "_FREQ"] = 0
        if limit > 0 and self.LIMITS[index] + "_FREQ" in self.proxyStruct[key]["limits_stat"] and \
        self.proxyStruct[key]["limits_stat"][self.LIMITS[index] + "_FREQ"] >= limit:
          ret = False
          break
        index += 1
    return ret


  # #Method checkLimits, checks available domains for current proxy
  #
  # @param key - proxy key
  # @return bool value - is available proxy by domain or not
  def checkDomains(self, key):
    ret = True
    if self.domain is not None and self.proxyStruct[key]["domains"] is not None \
    and self.domain not in self.proxyStruct[key]["domains"] and '*' not in self.proxyStruct[key]["domains"]:
      ret = False
    return ret


  # #Method commonIncrementLimits, method increments freq fields in common type container
  #
  # @param container - incoming container
  # @param key - proxy key
  def commonIncrementLimits(self, container, key):
    if key in container:
      container[key]["freq"] += 1
      if "limits_stat" in container[key] and len(container[key]["limits_stat"]) > 0:
        for elem in self.LIMITS:
          if elem + "_FREQ" in container[key]["limits_stat"]:
            container[key]["limits_stat"][elem + "_FREQ"] += 1


  # #Method incrementLimits increments freq firlds in proxyStruct and internalIndexes containers
  #
  # @param key - proxy key
  def incrementLimits(self, key):
    self.commonIncrementLimits(self.proxyStruct, key)
    self.commonIncrementLimits(self.internalIndexes, key)


  # #Method getProxy - public method which implements main proxy resolver functionality
  #
  # @return resolver proxy (as pair tuple (host, port)) or None
  def getProxy(self):
    ret = None
    if self.proxyTuple is not None:
      ret = self.proxyTuple
    elif self.proxyStruct is not None:
      for elem in sorted(self.proxyStruct.values(), key=lambda x: x["freq"] + x["priority"] * sys.maxint):
        if self.checkLimits(elem["host"]) and self.checkDomains(elem["host"]):
          if len(elem["host"].split(':')) > 1:
            ret = (elem["host"].split(':')[0], elem["host"].split(':')[1])
          else:
            ret = (elem["host"].split(':')[0], self.DEFAULT_PORT)
          self.incrementLimits(elem["host"])
          if self.indexFileName is not None:
            self.saveIndexInFile(self.indexFileName)
          break
    return ret
