"""
@package: dc
@file UrlSchema.py
@author Scorp <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import logging
import json
import random
import app.Consts as APP_CONSTS
import string
from app.Utils import ExceptionLog

logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)


## UrlSchema Class, implements functional UrlSchema custom generators
#
class UrlSchema(object):

  SHCEMA_DISABLE = 0
  SHCEMA_PREDEFINED = 1
  SHCEMA_INCREMENTAL_INT = 2
  SHCEMA_RANDOM_INT = 3
  SHCEMA_RANDOM_STR = 4

  CHAR_ASCII_LATIN = 0
  CHAR_HEXADECIMAL = 1
  CHAR_LOWER = 0
  CHAR_UPPER = 1

  MODE_ONE_URL = 0
  MODE_LIST_URLS = 1

  JSON_SUFF = ".json"


  # #Class constructor
  #
  # @param schema - incoming schema in json format
  def __init__(self, schema=None, siteId=None):
    try:
      self.globalindex = None
      self.schema = json.loads(schema)
      if "file_path" in self.schema and self.schema["file_path"] is not None:
        self.indexFileName = self.schema["file_path"]
        if siteId is not None and self.indexFileName.find(self.JSON_SUFF) != -1 and \
        self.indexFileName.find(self.JSON_SUFF) == len(self.indexFileName) - len(self.JSON_SUFF):
          if self.indexFileName[-1] != '/':
            self.indexFileName += '/'
          self.indexFileName += siteId
          self.indexFileName += self.JSON_SUFF
        self.indexStruct = self.readIndexFile(self.indexFileName)
      else:
        self.indexStruct = {}
        self.indexFileName = None
    except Exception as excp:
      ExceptionLog.handler(logger, excp, ">>> UrlSchema wrong json loads")
      self.schema = None


  # #Method readIndexFile reads from file and return param index structure
  #
  # @param fileName - incoming file name
  # @return json structure, just readed from file
  def readIndexFile(self, fileName):
    ret = {}
    fd = None
    try:
      fd = open(fileName, "r")
      if fd is not None:
        ret = json.loads(fd.read())
        fd.close()
    except Exception as excp:
      ExceptionLog.handler(logger, excp, ">>> readIndexFile error, file name = " + str(fileName))
      if fd is not None:
        fd.close()
    return ret


  # #Method schemaPredefined implements predefined schema algorithm
  #
  # @param inUrl - incoming url
  # @param parametrs - incoming schema params
  # @return processed url
  def schemaPredefined(self, inUrl, parametrs):
    for paramKey in parametrs:
      macroName = '%' + paramKey + '%'
      if inUrl.find(macroName) >= 0:
        index = 0
        if self.indexStruct is not None and paramKey in self.indexStruct:
          index = self.indexStruct[paramKey]
        if index < len(parametrs[paramKey]):
          inUrl = inUrl.replace(macroName, parametrs[paramKey][index])
        else:
          continue
        if self.indexStruct is not None:
          index += 1
          if index == len(parametrs[paramKey]):
            index = 0
          self.indexStruct[paramKey] = index
    return inUrl


  # #Method schemaIncrementalInt implements incremental int schema algorithm
  #
  # @param inUrl - incoming url
  # @param parametrs - incoming schema params
  # @return processed url
  def schemaIncrementalInt(self, inUrl, parameters):
    for paramKey in parameters:
      macroName = '%' + paramKey + '%'
      if inUrl.find(macroName) >= 0:
        index = parameters[paramKey]["min"]
        if self.indexStruct is not None and paramKey in self.indexStruct:
          index = self.indexStruct[paramKey]
        if index >= parameters[paramKey]["min"] and index <= parameters[paramKey]["max"]:
          inUrl = inUrl.replace(macroName, str(index))
        else:
          continue
        if self.indexStruct is not None:
          index += parameters[paramKey]["step"]
          if index > parameters[paramKey]["max"]:
            index = parameters[paramKey]["min"]
          self.indexStruct[paramKey] = index
    return inUrl


  # #Method schemaIncrementalInt implements random int schema algorithm
  #
  # @param inUrl - incoming url
  # @param parametrs - incoming schema params
  # @return processed url
  def schemaRandomInt(self, inUrl, parametrs):
    for paramKey in parametrs:
      macroName = '%' + paramKey + '%'
      if inUrl.find(macroName) >= 0:
        inUrl = inUrl.replace(macroName, str(random.randint(parametrs[paramKey]["min"], parametrs[paramKey]["max"])))
    return inUrl


  # #Method schemaIncrementalInt implements random string schema algorithm
  #
  # @param inUrl - incoming url
  # @param parametrs - incoming schema params
  # @return processed url
  def schemaRandomStr(self, inUrl, parametrs):
    lowAsciiSet = string.ascii_lowercase
    hexdigitsSet = ''.join([ch for ch in string.hexdigits if not ch.isupper()])
    for paramKey in parametrs:
      macroName = '%' + paramKey + '%'
      if inUrl.find(macroName) >= 0:
        valueLen = random.randint(parametrs[paramKey]["min"], parametrs[paramKey]["max"])
        valueStr = ''
        for i in xrange(0, valueLen):
          if parametrs[paramKey]["chars"] == self.CHAR_ASCII_LATIN:
            valueStr += lowAsciiSet[random.randint(0, len(lowAsciiSet) - 1)]
          elif parametrs[paramKey]["chars"] == self.CHAR_HEXADECIMAL:
            valueStr += hexdigitsSet[random.randint(0, len(hexdigitsSet) - 1)]
        if parametrs[paramKey]["case"] == self.CHAR_LOWER:
          valueStr = valueStr.lower()
        elif parametrs[paramKey]["case"] == self.CHAR_UPPER:
          valueStr = valueStr.upper()
        inUrl = inUrl.replace(macroName, valueStr)
    return inUrl


  # #Method saveIndexInFile saves indexes structute into the file
  #
  # @param fileName - incoming file name
  def saveIndexInFile(self, fileName):
    if self.indexStruct is not None and len(self.indexStruct) > 0:
      fd = None
      try:
        fd = open(fileName, "w")
        if fd is not None:
          fd.write(json.dumps(self.indexStruct))
          fd.close()
      except Exception as excp:
        ExceptionLog.handler(logger, excp, ">>> saveIndexInFile error, file name = " + str(fileName))
        if fd is not None:
          fd.close()


  # #Method generateUrlSchema main class public point, whcih
  #
  # @param inUrl - incoming url
  # @return processed url
  def generateUrlSchema(self, inUrl):
    ret = []
    itemsLen = 1
    if self.schema is not None:
      try:
        if self.schema["mode"] == self.MODE_LIST_URLS:
          itemsLen = self.schema["max_items"]
        for i in xrange(0, itemsLen):
          localRet = inUrl
          if self.schema["type"] == self.SHCEMA_PREDEFINED:
            localRet = self.schemaPredefined(inUrl, self.schema["parameters"])
          elif self.schema["type"] == self.SHCEMA_INCREMENTAL_INT:
            localRet = self.schemaIncrementalInt(inUrl, self.schema["parameters"])
          elif self.schema["type"] == self.SHCEMA_RANDOM_INT:
            localRet = self.schemaRandomInt(inUrl, self.schema["parameters"])
          elif self.schema["type"] == self.SHCEMA_RANDOM_STR:
            localRet = self.schemaRandomStr(inUrl, self.schema["parameters"])
          if localRet != inUrl and localRet not in ret:
            ret.append(localRet)
          else:
            break
      except Exception as excp:
        ExceptionLog.handler(logger, excp, ">>> generateUrlSchema methon some error")
      self.saveIndexInFile(self.indexFileName)
    logger.debug(">>> urlSchema len = " + str(len(ret)))
    return ret
