"""@package docstring
 @file scrapy_extractor.py
 @author Alexey <developers.hce@gmail.com>
 @link http://hierarchical-cluster-engine.com/
 @copyright Copyright &copy; 2013 IOIX Ukraine
 @license http://hierarchical-cluster-engine.com/license/
 @package HCE project node API
 @since 0.1
"""


import re
import json
import urllib
import logging
import types
from urlparse import urljoin
from app.SelectorWrapper import SelectorWrapper
from base_extractor import BaseExtractor
from app.Utils import varDump
import dc_processor.Constants as CONSTS
import app.Utils as Utils  # pylint: disable=F0401
import app.Consts as APP_CONSTS
import app.Utils as Utils
from app.Utils import ExceptionLog


# Logger initialization
logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)

template = {}
google_template = {}

# #ScrapyExtractor class implements data extracting using Scrapy module with prepared XPathes/Css
#
class ScrapyExtractor(BaseExtractor):

  SELF_NAME = "Scrapy extractor"

  # #class constructor
  #
  # @param config - incoming app config
  # @param templ - optionality dict with base set of extractor rules/xpathes
  # @param domain - optionality param, processing url domain
  # @param processorProperties - optionality param, incoming app processor properties
  # @param template - optionality param, incoming template set
  def __init__(self, config, templ=None, domain=None, processorProperties=None, template=None):
    BaseExtractor.__init__(self, config, templ, domain, processorProperties)
    self.merge_templates = False
    logger.debug("Properties: %s", varDump(self.properties))

    # set module rank from module's properties
    self.rankReading(self.__class__.__name__)

    if processorProperties is not None and CONSTS.TAG_MARKUP_PROP_NAME in processorProperties:
      self.innerTextTagReplacers = {}
      for key in processorProperties[CONSTS.TAG_MARKUP_PROP_NAME]:
        self.innerTextTagReplacers[key.lower()] = processorProperties[CONSTS.TAG_MARKUP_PROP_NAME][key]
    else:
      self.innerTextTagReplacers = None

    if processorProperties is not None and "SCRAPER_TAG_ITEMS_DELIMITER" in processorProperties:
      self.imgDelimiter = processorProperties["SCRAPER_TAG_ITEMS_DELIMITER"]
    else:
      self.imgDelimiter = ','
    self.name = self.SELF_NAME
    self.data["extractor"] = self.SELF_NAME
    self.sel = None
    self.resource = None
    self.default_template = None

    if template is not None:
      self.template = template
    else:
      self.template = self.templateLoad(config, templ, domain)
    logger.debug(">>> INIT Template Domain = " + str(domain))
    logger.debug(">>> INIT Template = " + str(self.template))


  # #templateLoad method which fills internal template dict with preparatory extractor rules/xpathes
  #
  # @param config - incoming app config
  # @param templ - optionality dict with base set of extractor rules/xpathes
  # @param domain - optionality param, processing url domain
  # @return result template dict
  def templateLoad(self, config, templ=None, domain=None):
    ret = {}
    self.default_template = None
    template_file = config.get("Application", "template", None)
    if template_file:
      try:
        logger.debug("Read template from file. %s", template_file)
        self.default_template = self.templatePreparer(open(template_file, "rb").read(), domain, {})
      except Exception, err:
        logger.error("Error Read template from file. %s", str(err))

    if CONSTS.TEMPLATE_KEY not in self.properties:
      ret = self.templatePreparer(self.properties[CONSTS.TEMPLATE_KEY], domain, {})
      logger.debug("template: " + str(ret))
    elif templ is not None:
      logger.debug("template: %s", str(templ))
      if type(templ) == type({}):
        ret = self.templatePreparer(None, domain, templ)
      else:
        ret = self.templatePreparer(templ, domain, {})

      # merge default template and custom one
      if self.default_template is not None and self.merge_templates:
        logger.debug("merge default template and custom one")
        default_tags = self.default_template.keys()
        custom_tags = ret.keys()
        logger.debug("tags in default template:\n%s\nin custom template:\n%s", str(default_tags), str(custom_tags))
        for tag in default_tags:
          if tag not in custom_tags:
            ret[tag] = self.default_template[tag]
            logger.debug("%s was replaced from custom template", str(tag))
    elif self.default_template is not None:
      ret = self.default_template
    else:
      logger.error("Error Read template.")
    return ret


  # #pasteLists pastes same elements in 2 incoming dicts
  #
  # @param dict1 - incoming destination dict
  # @param dict2 - incoming source dict
  def pasteLists(self, dict1, dict2):
    if type(dict1) == types.DictType and type(dict2) == types.DictType:
      for elem in dict2:
        if elem in dict1:
          dict1[elem] = dict1[elem] + dict2[elem]
        else:
          dict1[elem] = dict2[elem]


  # #Common method of prepared templates extract
  #
  # @param jsonBuf - incoming json with templates
  # @param domainCrc - current url comain crc (md5)
  # @param globalTemplate - incoming templates
  # @return template dict, that corresponds incoming domainCrc or "*"
  def templatePreparer(self, jsonBuf, domain, globalTemplate):
    ret = {}
    if len(globalTemplate) == 0:
      try:
        globalTemplate = json.loads(jsonBuf)
      except Exception, err:
        logger.error(">>> Wrong json format. %s", str(err))
    if len(globalTemplate) > 0:
      if domain is not None and domain in globalTemplate:
        if type(globalTemplate[domain]) is types.DictType:
          ret = globalTemplate[domain]
        else:
          logger.error(">>> Wrong type element for " + domain + " domain, type = " + str(type(globalTemplate[domain])))
      if domain is not None:
        if len(ret) == 0:
          while domain.find(".") != -1:
            domain = domain[domain.find(".") + 1: len(domain)]
            if domain is not None and domain in globalTemplate:
              self.pasteLists(ret, globalTemplate[domain])
          if domain is not None and domain in globalTemplate:
            self.pasteLists(ret, globalTemplate[domain])
      domain = "*"
      if domain is not None and domain in globalTemplate:
        self.pasteLists(ret, globalTemplate[domain])
    logger.debug(">>> template == " + str(ret))
    return ret


  # #extractTag method extracts for concrete tag
  #
  # @param tagName - incoming tag name
  # @param result - incoming result object
  # @param textHandler - optionality param with text processing callback function
  # @param delimiter - optionality delimiter between extracted elements
  def extractTag(self, tagName, result, textHandler=None, delimiter=' '):
    try:
      if tagName in self.template:
        for path in self.template[tagName]:
          if textHandler is not None:
            localValue = textHandler(self.sel.xpath(path), delimiter, delimiter, self.innerTextTagReplacers)
          else:
            localValue = self.sel.xpath(path).extract()
          self.addTag(result=result, tag_name=tagName, tag_value=localValue, xpath=path)
    except Exception as err:
      ExceptionLog.handler(logger, err, '>>> Some wrong in ScrapyExtractor.extractTag:', (err))


  # #extractTags public method extract data by tags and fills incoming result object
  #
  # @param resource - incoming raw data
  # @param result - incoming result object
  # @return incoming result with additionally filled fields/tags
  def extractTags(self, resource, result):
    try:
      self.resource = resource
      logger.debug("URL: %s", self.resource.url)
      self.sel = SelectorWrapper(text=resource.raw_html)

      # search engine parsing
      logger.debug("Regular parsing")
      self.extractTag(CONSTS.TAG_TITLE, result, Utils.innerText)
      self.extractTag(CONSTS.TAG_AUTHOR, result, Utils.innerText)
      self.extractTag(CONSTS.TAG_PUB_DATE, result)
      self.extractTag(CONSTS.TAG_DESCRIPTION, result, Utils.innerText)
      self.extractTag(CONSTS.TAG_DC_DATE, result)
      self.extractTag(CONSTS.TAG_MEDIA, result, Utils.innerText, self.imgDelimiter)
      self.extractTag(CONSTS.TAG_LINK, result, Utils.innerText)
      self.extractTag(CONSTS.TAG_CONTENT_UTF8_ENCODED, result, Utils.innerText)

      # for path in template["enclosure"]]
      self.extractTag(CONSTS.TAG_KEYWORDS, result, Utils.innerText)
      # Add support of html_lang tag
      self.extractTag(CONSTS.HTML_LANG, result, Utils.innerText)

      logger.debug("reslt: %s", varDump(result))
    except Exception as err:
      ExceptionLog.handler(logger, err, 'Parse error:', (err))

    return result
