#include "DataClientData.hpp"
#include "DRCEResourceExtractor.hpp"
#include "ServiceMessages.hpp"

namespace HCE {
  namespace handlers {

    DataClientData::DataClientData(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString, unsigned char dataClientMode, int64_t startedAt,
        unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, clientIdentity, connectionString, startedAt, logPriority, iniFile), DataNodeOptions(Handler::handlerProperties),
        _shiftCounter(){

      setIniFile(iniFile);
      setDataClientSockMode(dataClientMode);
      //!<Init all the rest
      initialize();
    }

    DataClientData::~DataClientData(){
      deinitialize();
    }

    int DataClientData::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << getClientIdentity() << "] with connection [" << getConnectionString() << "]" << flush;

      //!<Init sockets
      _inprocDataSock = NULL;
      _dataClientSock = NULL;
      _dataClientSockConnected = false;

      //!<Node init from file
      initNode(getIniFile(), logger);

      //!<Heartbeat at time mark
      setHeartbeatAt(s_clock()- getHeartbeatDelay());
      //!<Heartbeated time mark
      setHeartbeatedAt(s_clock());

      //!<Send out periodical stats at regular intervals
      setPstatsAt(s_clock() + STAT_INTERVAL);

      //!<Check initialization and set default value if necessary
      checkInit();

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildClientConnection(_inprocDataSock, NODE_CS_INPROC_DATA, ZMQ_DEALER, getClientIdentity());

      //!<Create dataclient socket, connect to upper node and send READY message
      rebuildDataClientSocket();

      return true;
    }

    void DataClientData::deinitialize(void){
      if(_inprocDataSock){
        delete _inprocDataSock;
      }
      _inprocDataSock = NULL;

      if(_dataClientSock){
        sendByeMessage(_dataClientSock, getClientIdentity());
        _dataClientSockConnected = false;
        delete _dataClientSock;
      }
      _dataClientSock = NULL;

      HCE::drce::ResourceUsageExtractor::shutdown();
    }

    void DataClientData::setDataClientSockMode(unsigned char dataClientSockMode){
      handlerProperties.set(HCE::handlers::properties::DATA_CLIENT_SOCK_MODE, dataClientSockMode);
    }

    unsigned char DataClientData::getDataClientSockMode(void){
      return handlerProperties.get<unsigned char>(HCE::handlers::properties::DATA_CLIENT_SOCK_MODE);
    }

    void DataClientData::setStatRequests(unsigned int statRequests){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS, statRequests);
    }

    unsigned int DataClientData::getStatRequests(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS);
    }

    void DataClientData::setStatResponses(unsigned int statResponses){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES, statResponses);
    }

    unsigned int DataClientData::getStatResponses(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_RESPONSES);
    }

    void DataClientData::setStatRequestsSize(unsigned int statRequestsSize){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_SIZE, statRequestsSize);
    }

    unsigned int DataClientData::getStatRequestsSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_SIZE);
    }

    void DataClientData::setStatResponsesSize(unsigned int statResponsesSize){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_SIZE, statResponsesSize);
    }

    unsigned int DataClientData::getStatResponsesSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_RESPONSES_SIZE);
    }

    void DataClientData::setStatRequestsProcessed(unsigned int statRequestsProcessed){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_PROCESSED, statRequestsProcessed);
    }

    unsigned int DataClientData::getStatRequestsProcessed(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_PROCESSED);
    }

    void DataClientData::setStatResponsesProcessed(unsigned int statResponsesProcessed){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_PROCESSED, statResponsesProcessed);
    }

    unsigned int DataClientData::getStatResponsesProcessed(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_RESPONSES_PROCESSED);
    }

    void DataClientData::setStatRequestsProcessedSize(unsigned int statRequestsProcessedSize){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE, statRequestsProcessedSize);
    }

    unsigned int DataClientData::getStatRequestsProcessedSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE);
    }

    void DataClientData::setStatResponsesProcessedSize(unsigned int statResponsesProcessedSize){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_PROCESSED_SIZE, statResponsesProcessedSize);
    }

    unsigned int DataClientData::getStatResponsesProcessedSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_RESPONSES_PROCESSED_SIZE);
    }

    void DataClientData::setStatQPS(float statQPS){
      statProperties.set(HCE::handlers::properties::STAT_QPS, statQPS);
    }

    float DataClientData::getStatQPS(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_QPS);
    }

    void DataClientData::setStatRPS(float statRPS){
      statProperties.set(HCE::handlers::properties::STAT_RPS, statRPS);
    }

    float DataClientData::getStatRPS(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_RPS);
    }

    void DataClientData::setStatRequestsBandwidth(float statRequestsBandwidth){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH, statRequestsBandwidth);
    }

    float DataClientData::getStatRequestsBandwidth(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH);
    }

    void DataClientData::setStatResponsesBandwidth(float statResponsesBandwidth){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH, statResponsesBandwidth);
    }

    float DataClientData::getStatResponsesBandwidth(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH);
    }

    void DataClientData::setPstatsAt(int64_t pstatsAt){
      handlerProperties.set(HCE::handlers::properties::PSTATS_AT, pstatsAt);
    }

    int64_t DataClientData::getPstatsAt(void){
      return handlerProperties.get<int64_t>(HCE::handlers::properties::PSTATS_AT);
    }

    void DataClientData::setHeartbeatAt(int64_t heartbeatAt){
      handlerProperties.set(HCE::handlers::properties::HEARTBEAT_AT, heartbeatAt);
    }

    int64_t DataClientData::getHeartbeatAt(void){
      return handlerProperties.get<int64_t>(HCE::handlers::properties::HEARTBEAT_AT);
    }

    void DataClientData::setHeartbeatedAt(int64_t heartbeatedAt){
      handlerProperties.set(HCE::handlers::properties::HEARTBEATED_AT, heartbeatedAt);
    }

    int64_t DataClientData::getHeartbeatedAt(void){
      return handlerProperties.get<int64_t>(HCE::handlers::properties::HEARTBEATED_AT);
    }

    void DataClientData::checkInit(void){
      //!<Check all stat variables
      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS))
        setStatRequests(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES))
        setStatResponses(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_SIZE))
        setStatRequestsSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_SIZE))
        setStatResponsesSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_PROCESSED))
        setStatRequestsProcessed(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_PROCESSED))
        setStatResponsesProcessed(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE))
        setStatRequestsProcessedSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_PROCESSED_SIZE))
        setStatResponsesProcessedSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_QPS))
        setStatQPS(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RPS))
        setStatRPS(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH))
        setStatRequestsBandwidth(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH))
        setStatResponsesBandwidth(0);
    }

    void DataClientData::rebuildDataClientSocket(void){
      const std::string connectionString = getConnectionString();
      if(connectionString != "" && connectionString != "0"){
        _dataClientSockConnected = rebuildClientConnection(_dataClientSock, connectionString, ZMQ_DEALER, getClientIdentity(), 0, NODE_MSG_READY);
      }else{
        logger.log(NODE_LOG_MODE_INFORMATION) << "Client connection was not rebuilt due connection string empty value [" << connectionString << "]" << flush;
      }
    }

    void DataClientData::handleExternalMessage(void){
      //!<Fetch message from dataclient
      zmsg clientMsg(*_dataClientSock);

      std::string addr((clientMsg.address())?clientMsg.address():"");

      logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::handleExternalMessage  addr: " << addr << flush;

      //!<Process message
      if(clientMsg.parts() == NODE_MSG_CMD_MIN_FIELDS){
        //!<Control message
        if(addr.compare(0, NODE_MSG_HEARTBEAT.length(), NODE_MSG_HEARTBEAT)==0){
          logger.log(NODE_LOG_MODE_DEBUG) << "HEARTBEAT message from server" << flush;

          //!<Register heartbit time mark
          setHeartbeatedAt(s_clock());

          if (getHeartbeatMode() == HEARBEAT_MODE_IMMEDIATE_AFTER_HB ||
              getHeartbeatMode() == HEARBEAT_MODE_MIXED){
            //!<Execute sending heartbeat
            sendHeartbeat();

            logger.log(NODE_LOG_MODE_TRACE) << "HEARBEAT_MODE_IMMEDIATE_AFTER_HB || HEARBEAT_MODE_MIXED" << flush;
          }

        }else{
          logger.log(NODE_LOG_MODE_TRACE) << "ERROR control message" << endl << clientMsg.dump() << flush;
        }
      }else{
        //!<Data message
        logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM dataclient" << endl << clientMsg.dump() << flush;

        //!<Create identity of dataclient's message
        std::string clientIdentity((const char*)clientMsg.pop_front().c_str());
        std::string messageId((const char*)clientMsg.pop_front().c_str());
        std::string messageBody((const char*)clientMsg.pop_front().c_str());
        std::string messageRoute((clientMsg.parts())?(const char*)clientMsg.pop_front().c_str():"");

        logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::handleExternalMessage  Recieved body: \"" << messageBody << "\"" << flush;
        logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::handleExternalMessage  Recieved route: \"" << messageRoute << "\"" << flush;

        unsigned int statRequests = getStatRequests();
        setStatRequests(++statRequests);
        setStatRequestsSize(getStatRequestsSize() + messageBody.size());

        logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from upper server [" << clientIdentity << "], message [" << messageId << "] body " << messageBody.size()
            << " bytes n=" << getStatRequests() << flush;

        //!<Create response message and set message Id
        zmsg msg(messageId.c_str());
        //!<Set message body
        msg.append(messageBody.c_str());
        //!<Set message route
        msg.append(messageRoute.c_str());

        if(getDataClientSockMode() == 0){
          //!<Send response for dataclient from end-point node
          msg.send(*_dataClientSock);
          unsigned int statResponses = getStatResponses();
          setStatResponses(++statResponses);
          setStatResponsesSize(getStatResponsesSize() + getStatRequestsSize());

          logger.log(NODE_LOG_MODE_TRACE) << " MSG TO dataclient" << endl << msg.dump() << flush;
          logger.log(NODE_LOG_MODE_DEBUG) << " RESPONSE to dataclient socket from end-point node to upper server [" << clientIdentity << "] sent, n=" << getStatResponses() << flush;

        }else{
          //!<Send response for data handler of end-point node
          msg.send(*_inprocDataSock);
          logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT to inproc_data socket" << flush;
        }

        //!<Register heartbeated time mark
        setHeartbeatedAt(s_clock());

        if (getHeartbeatMode() == HEARBEAT_MODE_ADAPTIVE){

          //!<Reset timestamp for shift counter
          _shiftCounter.reset();
          //!<Move heartbeat time forward
          setHeartbeatAt(s_clock()/* + getHeartbeatDelay()*/);
        }
      }
    }

    void DataClientData::handleInternalMessage(void){
      //!<Get message from inproc_data
      zmsg msg(*_inprocDataSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_data" << endl << msg.dump() << flush;

      //!<Re-send response message directly, not using the responses map
      msg.send(*_dataClientSock);

      unsigned int statResponses = getStatResponses();
      setStatResponses(++statResponses);
      setStatResponsesSize(getStatResponsesSize() + msg.size());

      logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE to dataclient socket from data node to upper server sent, n=" << getStatResponses() << flush;

      //!<Move heartbit time forward
      setHeartbeatAt(s_clock() + getHeartbeatDelay());
    }

    void DataClientData::heartbit(void){
      if (getHeartbeatMode() == HEARBEAT_MODE_FIXED_TIME ||
          getHeartbeatMode() == HEARBEAT_MODE_MIXED){
        if((s_clock() > getHeartbeatAt()) && _dataClientSockConnected){

          //!<Execute sending heartbeat
          sendHeartbeat();

          logger.log(NODE_LOG_MODE_TRACE) << "HEARBEAT_MODE_FIXED_TIME || HEARBEAT_MODE_MIXED" << flush;
          logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << getClientIdentity() << "] to server" << flush;
        }
      }else if(getHeartbeatMode() == HEARBEAT_MODE_ADAPTIVE){
        _shiftCounter.responceTimeStamp(getHeartbeatedAt());
        _shiftCounter.recount(getHeartbeatTimeout());

        logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::heartbit  HEARBEAT_MODE_ADAPTIVE  delta = " << (s_clock()- getHeartbeatAt()+getHeartbeatDelay())
            << " _shiftCounter.getShift()= " << _shiftCounter.getShift() << flush;

        size_t delta = s_clock()-getHeartbeatAt()+getHeartbeatDelay();
        if (delta > (getHeartbeatTimeout()+_shiftCounter.getShift()) && _dataClientSockConnected){
          //!<Execute sending heartbeat
          sendHeartbeat();
          //!<Reset shift counter
          _shiftCounter.reset();

          logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << getClientIdentity() << "] to server getHeartbeatAt() = '"
              << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(getHeartbeatAt()/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;
        }
        //!<Register heartbeated time mark
        setHeartbeatedAt(s_clock());
      }
    }

    void DataClientData::sendHeartbeat(void){
      //!<Send heartbeat with resource usage message
//      zmsg msg(NODE_MSG_HEARTBEAT.c_str());
//      msg.append(getResourceUsageMessage().c_str());
//      msg.send(*_dataClientSock);

      std::vector<ClientWorkerItem> queue;
      HCE::handlers::HeartbeatMessage heartbeatMessage(queue, getResourceUsageMessage());
      std::string message = heartbeatMessage.build(NODE_MSG_HEARTBEAT);
      s_send(*_dataClientSock, message);

      logger.log(NODE_LOG_MODE_TRACE) << "SENT HEARTBIT MESSAGE: " << message << flush;

      //!<Reset heartbit time
      setHeartbeatAt(s_clock() + getHeartbeatDelay());
    }

    std::string DataClientData::getResourceUsageMessage(void) const{
      //!<Get resource usage data
      HCE::drce::ResourceUsageExtractor::getInstance().update();
      HCE::drce::ResourceUsage resourceUsage = HCE::drce::ResourceUsageExtractor::getInstance().extract();
      HCE::drce::ResourceUsageConverter resourceUsageConverter(resourceUsage);
      return resourceUsageConverter.toString();
    }

    void DataClientData::logPeriodicStatistics(void){
      if(s_clock() > getPstatsAt()){
        setStatQPS((getStatRequests() - getStatRequestsProcessed()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatRPS(((getStatResponses() - getStatResponsesProcessed()) / (float)((STAT_INTERVAL) / 1000.0)));
        setStatRequestsBandwidth((getStatRequestsSize() - getStatRequestsProcessedSize()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatResponsesBandwidth((getStatResponsesSize() - getStatResponsesProcessedSize()) / (float)((STAT_INTERVAL) / 1000.0));

        setStatRequestsProcessed(getStatRequests());
        setStatResponsesProcessed(getStatResponses());
        setStatRequestsProcessedSize(getStatRequestsSize());
        setStatResponsesProcessedSize(getStatResponsesSize());

        logger.log(NODE_LOG_MODE_DEBUG) << "requests=" << getStatRequests() << ", responses=" << getStatResponses() << ", qps=" << getStatQPS() << ", rps=" << getStatRPS()
            << ", requestsSize=" << getStatRequestsSize() << ", responsesSize=" << getStatResponsesSize() << ", requestsBandwidth="
            << getStatRequestsBandwidth() << ", responsesBandwidth=" << getStatResponsesBandwidth() << flush;

        //!<Reset pstat time
        setPstatsAt(s_clock() + STAT_INTERVAL);
      }
    }

    std::string DataClientData::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "requests=" << getStatRequests()
          << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << getStatResponses()
          << NODE_ADMIN_COMMAND_DELIMITER << "requestsSize=" << getStatRequestsSize()
          << NODE_ADMIN_COMMAND_DELIMITER << "responsesSize=" << getStatResponsesSize()
          << NODE_ADMIN_COMMAND_DELIMITER << "clientIdentity=" << getClientIdentity()
          << NODE_ADMIN_COMMAND_DELIMITER << "dataClientMode=" << static_cast<int>(getDataClientSockMode())
          << NODE_ADMIN_COMMAND_DELIMITER << "pstatsAt=" << static_cast<uint64_t>(getPstatsAt())
          << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatAt=" << static_cast<uint64_t>(getHeartbeatAt())
          << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatedAt=" << static_cast<uint64_t>(getHeartbeatedAt())
          << NODE_ADMIN_COMMAND_DELIMITER << "statQPS=" << getStatQPS()
          << NODE_ADMIN_COMMAND_DELIMITER << "statRPS=" << getStatRPS()
          << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidth=" << getStatRequestsBandwidth()
          << NODE_ADMIN_COMMAND_DELIMITER << "statResponsesBandwidth=" << getStatResponsesBandwidth();
      return result.str();
    }

    void DataClientData::externalConnectionRefresh(void){

      if (getHeartbeatMode() != HEARBEAT_MODE_ADAPTIVE){

        if(((s_clock() - getHeartbeatedAt()) > getHeartbeatTimeout()) && getInProgress() && _dataClientSockConnected){
          logger.log(NODE_LOG_MODE_INFORMATION) << "RECONNECT client [" << getClientIdentity() << "] because HEARTBIT not received during time of "
              << (getHeartbeatTimeout()+_shiftCounter.getShift()) << " msec" << flush;

          //!<Rebuild data socket
          rebuildDataClientSocket();

          //!<Reset time mark
          setHeartbeatedAt(s_clock());

          //!<Move heartbit time forward
          setHeartbeatAt(s_clock() + getHeartbeatDelay());
        }
      }else{ // HEARBEAT_MODE_ADAPTIVE

        _shiftCounter.recount(getHeartbeatTimeout());

        logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::externalConnectionRefresh  (s_clock()-getHeartbeatedAt())= " << (s_clock()-getHeartbeatedAt())
            << " getHeartbeatTimeout()= " << getHeartbeatTimeout() << " _shiftCounter.getShift()= " << _shiftCounter.getShift() << flush;

        if(((s_clock() - getHeartbeatedAt()) > (getHeartbeatTimeout()+_shiftCounter.getShift())) && getInProgress() && _dataClientSockConnected){
          logger.log(NODE_LOG_MODE_INFORMATION) << "RECONNECT client [" << getClientIdentity() << "] because HEARTBIT not received during time of "
              << (getHeartbeatTimeout()+_shiftCounter.getShift()) << " msec" << flush;

          //!<Rebuild data socket
          rebuildDataClientSocket();

          //!<Reset time mark
          setHeartbeatedAt(s_clock());

          _shiftCounter.reset();

          //!<Move heartbit time forward
          setHeartbeatAt(s_clock() + getHeartbeatDelay());
        }
      }
    }

    std::string DataClientData::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << getStatRequests() << ",responses=" << getStatResponses()
            << ",requests_size=" << getStatRequestsSize() << ",responses_size=" << getStatResponsesSize();
        ret = result.str();
      }else if(command == NODE_MSG_REBUILD_CLIENT_CONNECTION){
        //!<Rebuild client connection
        rebuildDataClientSocket();
      }else if(command == NODE_MSG_DISCONNECT_CLIENT_CONNECTION){
        //!<Disconnect client connection
        _dataClientSock->close();
        _dataClientSockConnected = false;
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_SET){
        //!<Parse command parameters
        unsigned int heartbeatDelay = HEARTBEAT_DELAY;
        //!<Parse heartbeat delay from parameters
        std::istringstream(parameters) >> heartbeatDelay;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
        //!<Set new heartbeat delay value
        setHeartbeatDelay(heartbeatDelay*1000);
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(getHeartbeatTimeout()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_SET){
        //!<Parse command parameters
        unsigned int heartbeatTimeout = HEARTBEAT_TIMEOUT;
        //!<Parse heartbeat timeout from parameters
        std::istringstream(parameters) >> heartbeatTimeout;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(getHeartbeatTimeout()/1000);
        //!<Set new heartbeat timeout value
        setHeartbeatTimeout(heartbeatTimeout*1000);
      }else if(command == NODE_MSG_HEARTBEAT_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
      }else if(command == NODE_MSG_HEARTBEAT_MODE_SET){
        //!<Parse command parameters
        int hbMode = HEARBEAT_MODE_DEFAULT;
        //!<Parse heartbeat mode from parameters
        std::istringstream(parameters) >> hbMode;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
        //!<Set new heartbeat mode value
        setHeartbeatMode(static_cast<unsigned int>(hbMode));
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    void* DataClientData::process(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      long delay = getPollTimeout();

      while(getInProgress()){
        try{
          //!<Init polling pool
          std::vector<zmq::pollitem_t> items;
          if(_dataClientSockConnected){
            items ={{*_dataClientSock, 0, ZMQ_POLLIN, 0},{*_inprocDataSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          } else{
            items ={{*_inprocDataSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          }

          //!<Reinit delay if necessary
          if (delay){
            delay = getPollTimeout();
          }

          //!<Poll dataclient
          zmq::poll(&items[0], items.size(), delay);

          //!<Handle dataclient socket's message
          if(_dataClientSockConnected){
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_data socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleAdminMessage();
            if(getInProgress() == false){
              delay = 0;
              setInProgress(true);
              continue;
            }
          }

          if(getInProgress() == true && delay == 0){
            setInProgress(false);
            continue;
          }

          //!<Send heartbeat to server if it's time
          heartbit();

          //!<Out periodical stats
          logPeriodicStatistics();

          //!<Send periodical property values
          sendPeriodicProperties();

          //!<Dump run-time variables at regular intervals
          dumpRunTimeVariables();

          //!<Heartbited interval check and connection refresh
          externalConnectionRefresh();

        }catch(zmq::error_t& error){
          if (EINTR!=error.num()){
            std::cout << "DataClientData E: " << error.what() << std::endl;
          }
        }catch(std::exception& e){
          std::cout << "DataClientData E: " << e.what() << std::endl;
        }catch(...){
          std::cout << "DataClientData E: Unknown exception" << std::endl;
        }
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }
  }
}
