/**
 * @file DataClientData.hpp
 *
 * HCE project node dataClientData handler.
 * Implements functionality for data client handler server.
 * Implements data server functionality for data client's requests handling and HCE project data node construction.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_DATA_CLIENT_DATA_INCLUDED__
#define __HCE_DATA_CLIENT_DATA_INCLUDED__

#include <iostream>
#include <sstream>

#include "Handler.hpp"
#include "DataNodeOptions.hpp"

//Node ini
#include <Poco/Util/IniFileConfiguration.h>
#include "Poco/String.h"

namespace HCE {
  namespace handlers {

    class DataClientData: public HCE::handlers::Handler, protected HCE::handlers::DataNodeOptions {
      public:
        DataClientData(std::string name, zmq::context_t& ctx,
            const std::string& clientIdentity, std::string connectionString,
            unsigned char dataClientMode, int64_t startedAt,
            unsigned char logPriority, const std::string& iniFile="");
        ~DataClientData();

      protected:
        int initialize(void);
        void deinitialize(void);
        void* process(void);

        void rebuildDataClientSocket(void);
        void handleExternalMessage(void);
        void handleInternalMessage(void);
        void heartbit(void);
        void sendHeartbeat(void);
        void logPeriodicStatistics(void);
        void externalConnectionRefresh(void);
        std::string processAdminCommands(std::string& command, std::string& parameters);

        //!<Get resource usage message
        std::string getResourceUsageMessage(void) const;

        //!<Get property information
        std::string getPropertyInformation(void);

        //!<Check initialization and set default value if necessary
        void checkInit(void);

        //!<Sockets
        zmq::socket_t* _inprocDataSock;
        zmq::socket_t* _dataClientSock;
        bool _dataClientSockConnected = false;

        //!<Timeout shift counter
        TimeoutShiftCounter _shiftCounter;

        //!<Data client mode
        void setDataClientSockMode(unsigned char dataClientSockMode);
        unsigned char getDataClientSockMode(void);

        //!<Setters and getters for stat variables
        void setStatRequests(unsigned int statRequests);
        unsigned int getStatRequests(void);

        void setStatResponses(unsigned int statResponses);
        unsigned int getStatResponses(void);

        void setStatRequestsSize(unsigned int statRequestsSize);
        unsigned int getStatRequestsSize(void);

        void setStatResponsesSize(unsigned int statResponsesSize);
        unsigned int getStatResponsesSize(void);

        void setStatRequestsProcessed(unsigned int statRequestsProcessed);
        unsigned int getStatRequestsProcessed(void);

        void setStatResponsesProcessed(unsigned int statResponsesProcessed);
        unsigned int getStatResponsesProcessed(void);

        void setStatRequestsProcessedSize(unsigned int statRequestsProcessedSize);
        unsigned int getStatRequestsProcessedSize(void);

        void setStatResponsesProcessedSize(unsigned int statResponsesProcessedSize);
        unsigned int getStatResponsesProcessedSize(void);

        void setStatQPS(float statQPS);
        float getStatQPS(void);

        void setStatRPS(float statRPS);
        float getStatRPS(void);

        void setStatRequestsBandwidth(float statRequestsBandwidth);
        float getStatRequestsBandwidth(void);

        void setStatResponsesBandwidth(float statResponsesBandwidth);
        float getStatResponsesBandwidth(void);

        //!<Send out periodical stats at regular intervals
        void setPstatsAt(int64_t pstatsAt);
        int64_t getPstatsAt(void);

        //!<Time when heartbeat message for external connection need to be sent next time
        void setHeartbeatAt(int64_t heartbeatAt);
        int64_t getHeartbeatAt(void);

        //!<Last time when external connectionhave received heartbit message
        void setHeartbeatedAt(int64_t heartbeatedAt);
        int64_t getHeartbeatedAt(void);
    };
  }
}

#endif /* __HCE_DATA_CLIENT_DATA_INCLUDED__ */
