#include <Poco/AutoPtr.h>
#include <Poco/Configurable.h>
#include <Poco/Util/IniFileConfiguration.h>
#include <Poco/String.h>

#include "HandlerTypes.hpp"
#include "DataNodeOptions.hpp"

namespace HCE
{
namespace handlers
{
//-----------------------------------------------------------------------------
DataNodeOptions::DataNodeOptions(HandlerProperties& handlerProperties)
: properties(handlerProperties)
{
}
//-----------------------------------------------------------------------------
void DataNodeOptions::initNode(const std::string& iniFile, LoggerStream& logger)
{
  if (!iniFile.empty())
  {
    logger.log(NODE_LOG_MODE_INFORMATION) << "load node settings from '" << iniFile << "' file" << flush;

    //!<Process Node's ini file
    try
    {
      Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
      try
      {
        if (!properties.isExist(HCE::handlers::properties::HEARTBEAT_DELAY))
          setHeartbeatDelay(pConf->getInt(HCE::handlers::properties::HEARTBEAT_DELAY)*1000);

        if (!properties.isExist(HCE::handlers::properties::HEARTBEAT_TIMEOUT))
          setHeartbeatTimeout(pConf->getInt(HCE::handlers::properties::HEARTBEAT_TIMEOUT)*1000);

        if (!properties.isExist(HCE::handlers::properties::HEARTBEAT_MODE))
          setHeartbeatMode(pConf->getInt(HCE::handlers::properties::HEARTBEAT_MODE));
      }
      catch(Poco::SyntaxException& e)
      {
        //!<Log some error of config file processing
        logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node options ini file processing error: '" << e.displayText() << "'" << flush;
      }
    }
    catch(Poco::Exception& e)
    {
      //!<Log some error of config file processing
      logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node options ini file access error: '" << e.displayText() << "'" << flush;
    }
    //!<Apply default parameters if nessesary
    if (!properties.isExist(HCE::handlers::properties::HEARTBEAT_DELAY))
      setHeartbeatDelay(HCE::handlers::HEARTBEAT_DELAY);

    if (!properties.isExist(HCE::handlers::properties::HEARTBEAT_TIMEOUT))
      setHeartbeatTimeout(HCE::handlers::HEARTBEAT_TIMEOUT);

    if (!properties.isExist(HCE::handlers::properties::HEARTBEAT_MODE))
      setHeartbeatMode(HCE::handlers::HEARBEAT_MODE_DEFAULT);

    logger.log(NODE_LOG_MODE_INFORMATION) << "hb_delay = " << getHeartbeatDelay() << flush;
    logger.log(NODE_LOG_MODE_INFORMATION) << "hb_timeout = " << getHeartbeatTimeout() << flush;
    logger.log(NODE_LOG_MODE_INFORMATION) << "hb_mode = " << getHeartbeatMode() << flush;
  }
}
//-----------------------------------------------------------------------------
void DataNodeOptions::setHeartbeatDelay(unsigned int heartbeatDelay)
{
  if (heartbeatDelay >= MIN_ALLOWED_HEARTBEAT_DELAY)
    properties.set(HCE::handlers::properties::HEARTBEAT_DELAY, heartbeatDelay);
}
//-----------------------------------------------------------------------------
unsigned int DataNodeOptions::getHeartbeatDelay(void)
{
  return properties.get<unsigned int>(HCE::handlers::properties::HEARTBEAT_DELAY);
}
//-----------------------------------------------------------------------------
void DataNodeOptions::setHeartbeatTimeout(unsigned int heartbeatTimeout)
{
  if (heartbeatTimeout >= MIN_ALLOWED_HEARTBEAT_TIMEOUT)
    properties.set(HCE::handlers::properties::HEARTBEAT_TIMEOUT, heartbeatTimeout);
}
//-----------------------------------------------------------------------------
unsigned int DataNodeOptions::getHeartbeatTimeout(void)
{
  return properties.get<unsigned int>(HCE::handlers::properties::HEARTBEAT_TIMEOUT);
}
//-----------------------------------------------------------------------------
void DataNodeOptions::setHeartbeatMode(unsigned int heartbeatMode)
{
  if (heartbeatMode <= HEARBEAT_MODE_MAX_NUMBER && heartbeatMode >= HEARBEAT_MODE_DEFAULT)
    properties.set(HCE::handlers::properties::HEARTBEAT_MODE, heartbeatMode);
}
//-----------------------------------------------------------------------------
unsigned int DataNodeOptions::getHeartbeatMode(void)
{
  return properties.get<unsigned int>(HCE::handlers::properties::HEARTBEAT_MODE);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace drce
} // end namespace HCE 
