#include "DataProcessorData.hpp"

namespace HCE {
  namespace handlers {
    DataProcessorData::DataProcessorData(const std::string& nodeName, std::string name, zmq::context_t& ctx, int64_t startedAt, unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, nodeName, std::string(""), startedAt, logPriority, iniFile), _inprocDataSock(nullptr), _tasksStateNotificationSock(nullptr),
        _sphinxFO(nullptr), _drceFO(nullptr), _tasksStateNotificationFunctor(nullptr) {

      setNodeName(nodeName);
      setIniFile(iniFile);
      //!<Init all the rest
      initialize();
    }

    DataProcessorData::~DataProcessorData(){
      deinitialize();
    }

    int DataProcessorData::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize..." << flush;

      //!<Init sockets
      _inprocDataSock = nullptr;
      _tasksStateNotificationSock = nullptr;
      _tasksStateNotificationFunctor = nullptr;

      //!<Node init
      nodeInit(getIniFile());

      //!<Check initialization and set default value if necessary
      checkInit();

      //!<Sphinx FO init
      dataStorageInitSphinx(getIniFile());

      //!<DRCE FO init
      dataStorageInitDRCE(getIniFile());

      //!<Restore node properties
      restoreNodeProperties();

      //!<Restore Sphinx FO properties
      restoreSphinxProperties();

      //!<Restore DRCE FO properties
      restoreDRCEProperties();

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildServerConnection(_inprocDataSock, NODE_CS_INPROC_DATA, ZMQ_DEALER);

      return true;
    }

    void DataProcessorData::deinitialize(void){

      saveNodeProperties();
      saveSphinxProperties();
      saveDRCEProperties();

      if(_inprocDataSock){
        if(_inprocDataSock->connected()){
          _inprocDataSock->close();
        }
        delete _inprocDataSock;
        _inprocDataSock = NULL;
      }

      if(_sphinxFO){
        delete _sphinxFO;
        _sphinxFO = NULL;
      }

      if(_tasksStateNotificationSock){
        delete _tasksStateNotificationSock;
        _tasksStateNotificationSock = NULL;
      }

      if(_tasksStateNotificationFunctor){
        delete _tasksStateNotificationFunctor;
        _tasksStateNotificationFunctor = NULL;
      }

      if(_drceFO){
        delete _drceFO;
        _drceFO = NULL;
      }
    }

    void DataProcessorData::setNodeName(const std::string& nodeName){
      handlerProperties.set(HCE::handlers::properties::NODE_NAME, nodeName);
      setClientIdentity(nodeName);
    }

    std::string DataProcessorData::getNodeName(void) {
      return handlerProperties.get<std::string>(HCE::handlers::properties::NODE_NAME);
    }

    void DataProcessorData::setStatClientResponses(unsigned long long statClientResponses){
      statProperties.set(HCE::handlers::properties::STAT_CLIENT_RESPONSES, static_cast<Poco::UInt64>(statClientResponses));
    }

    unsigned long long DataProcessorData::getStatClientResponses(void){
      return static_cast<unsigned long long>(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_CLIENT_RESPONSES));
    }

    void DataProcessorData::setStatSphinxRequestsSearch(unsigned long long statSphinxRequestsSearch){
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH, static_cast<Poco::UInt64>(statSphinxRequestsSearch));
    }

    unsigned long long DataProcessorData::getStatSphinxRequestsSearch(void){
      return static_cast<unsigned long long>(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH));
    }

    void DataProcessorData::setStatSphinxRequestsAdmin(unsigned long long statSphinxRequestsAdmin){
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN, static_cast<Poco::UInt64>(statSphinxRequestsAdmin));
    }

    unsigned long long DataProcessorData::getStatSphinxRequestsAdmin(void){
      return static_cast<unsigned long long>(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN));
    }

    void DataProcessorData::setStatSphinxTimeTotalSearch(unsigned long long statSphinxTimeTotalSearch){
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_SEARCH, static_cast<Poco::UInt64>(statSphinxTimeTotalSearch));
    }

    unsigned long long DataProcessorData::getStatSphinxTimeTotalSearch(void){
      return static_cast<unsigned long long>(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_SEARCH));
    }

    void DataProcessorData::setStatSphinxTimeTotalAdmin(unsigned long long statSphinxTimeTotalAdmin){
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_ADMIN, static_cast<Poco::UInt64>(statSphinxTimeTotalAdmin));
    }

    unsigned long long DataProcessorData::getStatSphinxTimeTotalAdmin(void){
      return static_cast<unsigned long long>(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_ADMIN));
    }

    void DataProcessorData::setStatDRCERequests(unsigned long long statDRCERequests){
      statProperties.set(HCE::handlers::properties::STAT_DRCE_REQUESTS, static_cast<Poco::UInt64>(statDRCERequests));
    }

    unsigned long long DataProcessorData::getStatDRCERequests(void){
      return static_cast<unsigned long long>(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_DRCE_REQUESTS));
    }

    void DataProcessorData::setStatDRCETimeTotal(unsigned long long statDRCETimeTotal){
      statProperties.set(HCE::handlers::properties::STAT_DRCE_TIME_TOTAL, static_cast<Poco::UInt64>(statDRCETimeTotal));
    }

    unsigned long long DataProcessorData::getStatDRCETimeTotal(void){
      return static_cast<unsigned long long>(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_DRCE_TIME_TOTAL));
    }

    void DataProcessorData::setDataProcessingMode(unsigned int dataProcessingMode){
      handlerProperties.set(HCE::handlers::properties::DATA_PROCESSING_MODE, dataProcessingMode);
    }

    unsigned int DataProcessorData::getDataProcessingMode(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::DATA_PROCESSING_MODE);
    }

    void DataProcessorData::setDataProcessingFakeResults(unsigned int dataProcessingFakeResults){
      handlerProperties.set(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS, dataProcessingFakeResults);
    }

    unsigned int DataProcessorData::getDataProcessingFakeResults(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS);
    }

    void DataProcessorData::setDataProcessingFakeResultsMax(unsigned int dataProcessingFakeResultsMax){
      handlerProperties.set(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS_MAX, dataProcessingFakeResultsMax);
    }

    unsigned int DataProcessorData::getDataProcessingFakeResultsMax(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS_MAX);
    }

    void DataProcessorData::setNodeNumber(unsigned int NodeNumber){
      handlerProperties.set(HCE::handlers::properties::NODE_NUMBER, NodeNumber);
    }

    unsigned int DataProcessorData::getNodeNumber(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::NODE_NUMBER);
    }

    void DataProcessorData::setStateNotificationHost(const std::string& stateNotificationHost){
      handlerProperties.set(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_HOST, stateNotificationHost);
    }

    std::string DataProcessorData::getStateNotificationHost(void){
      return handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_HOST);
    }

    void DataProcessorData::setStateNotificationPort(const std::string& stateNotificationPort){
      handlerProperties.set(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_PORT, stateNotificationPort);
    }

    std::string DataProcessorData::getStateNotificationPort(void){
      return handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_PORT);
    }

    void DataProcessorData::checkInit(void){
      //!<Check all stat variables
      if (!statProperties.isExist(HCE::handlers::properties::STAT_CLIENT_RESPONSES))
        setStatClientResponses(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH))
        setStatSphinxRequestsSearch(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN))
        setStatSphinxRequestsAdmin(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_SEARCH))
        setStatSphinxTimeTotalSearch(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_ADMIN))
        setStatSphinxTimeTotalAdmin(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_DRCE_REQUESTS))
        setStatDRCERequests(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_DRCE_TIME_TOTAL))
        setStatDRCETimeTotal(0);
    }

    void DataProcessorData::saveNodeProperties(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Dump node properties" << flush;

      handlerProperties.set(HCE::handlers::properties::NODE_NAME, getNodeName());
      handlerProperties.set(HCE::handlers::properties::DATA_PROCESSING_MODE, getDataProcessingMode());
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_SEARCH, static_cast<Poco::UInt64>(getStatSphinxTimeTotalSearch()));
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH, static_cast<Poco::UInt64>(getStatSphinxRequestsSearch()));
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_ADMIN, static_cast<Poco::UInt64>(getStatSphinxTimeTotalAdmin()));
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN, static_cast<Poco::UInt64>(getStatSphinxRequestsAdmin()));
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH, static_cast<Poco::UInt64>(getStatSphinxRequestsSearch()));
      statProperties.set(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN, static_cast<Poco::UInt64>(getStatSphinxRequestsAdmin()));
      statProperties.set(HCE::handlers::properties::STAT_DRCE_REQUESTS, static_cast<Poco::UInt64>(getStatDRCERequests()));
      statProperties.set(HCE::handlers::properties::STAT_DRCE_TIME_TOTAL, static_cast<Poco::UInt64>(getStatDRCETimeTotal()));
      handlerProperties.set(HCE::handlers::properties::NODE_NUMBER, getNodeNumber());
    }

    void DataProcessorData::saveSphinxProperties(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Dump Sphinx FO properties" << flush;
      if (_sphinxFO){
        handlerProperties.set(HCE::handlers::properties::SPHINX_INDEX_NAME, _sphinxFO->getIndexName());
        handlerProperties.set(HCE::handlers::properties::SPHINX_HOME_DIR, _sphinxFO->getHomeDir());
        handlerProperties.set(HCE::handlers::properties::SPHINX_SEARCHD_HOST, _sphinxFO->getServerHost());
        handlerProperties.set(HCE::handlers::properties::SPHINX_SEARCHD_PORT, _sphinxFO->getServerPort());
        handlerProperties.set(HCE::handlers::properties::SPHINX_TIMEOUTED_REQUESTS, static_cast<Poco::UInt64>(_sphinxFO->getTimeoutedRequests()));
        handlerProperties.set(HCE::handlers::properties::SPHINX_RANKER, _sphinxFO->getRankingMode());
        handlerProperties.set(HCE::handlers::properties::SPHINX_RANKER_EXPRESSION, Poco::replace(_sphinxFO->getRankerExpression(), "=", "%3D"));
        handlerProperties.set(HCE::handlers::properties::SPHINX_NODE_NUMBER, _sphinxFO->getNodeNumber());
        handlerProperties.set(HCE::handlers::properties::SPHINX_NODE_NAME, _sphinxFO->getNodeName());
      }
    }

    void DataProcessorData::saveDRCEProperties(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Dump DRCE FO properties" << flush;
      if (_drceFO){
        handlerProperties.set(HCE::handlers::properties::DRCE_NODE_NAME, _drceFO->getNodeName());
        handlerProperties.set(HCE::handlers::properties::DRCE_NODE_HOST, _drceFO->getNodeHost());
        handlerProperties.set(HCE::handlers::properties::DRCE_NODE_PORT, _drceFO->getNodePort());
        handlerProperties.set(HCE::handlers::properties::DRCE_HOME_DIR, _drceFO->getHomeDir());
        handlerProperties.set(HCE::handlers::properties::DRCE_TASK_DATA_DIR, _drceFO->getTasksDataDir());
        handlerProperties.set(HCE::handlers::properties::DRCE_TASK_STATUS_DIR, _drceFO->getTasksStatusDir());
        handlerProperties.set(HCE::handlers::properties::DRCE_MAX_THREAD_COUNT, _drceFO->getMaxThreadCount());
        handlerProperties.set(HCE::handlers::properties::DRCE_TASK_QUEUE_DUMP_PERIOD, _drceFO->getTasksQueueDumpPeriod());
        handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCE_MONITOR_TIME_PERIOD, _drceFO->getResourceMonitorTimePeriod());
        handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS, _drceFO->getResourceExtractorMaxThreadsCount());
        handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES, _drceFO->getResourceExtractorMaxProcessesCount());
        handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE, _drceFO->getResourceExtractorMaxDiskSize());
        handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE, _drceFO->getResourceExtractorMaxVramSize());
        handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE, _drceFO->getResourceExtractorMaxRramSize());

        //!<Dump stat variables
        HCE::drce::DRCEFunctionalObject::Dictionary dictionary = _drceFO->dumpResourceMonitor();
        for (HCE::drce::DRCEFunctionalObject::Dictionary::iterator iter=dictionary.begin();iter!=dictionary.end();++iter)
        {
          statProperties.set((*iter).first, (*iter).second);
        }
      }
    }

    void DataProcessorData::restoreNodeProperties(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Restore node properties" << flush;

      if (handlerProperties.isExist(HCE::handlers::properties::NODE_NAME)){
        setNodeName(handlerProperties.get<std::string>(HCE::handlers::properties::NODE_NAME));
      }

      if (handlerProperties.isExist(HCE::handlers::properties::DATA_PROCESSING_MODE)){
        setDataProcessingMode(handlerProperties.get<unsigned int>(HCE::handlers::properties::DATA_PROCESSING_MODE));
      }

      if (statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_SEARCH)){
        setStatSphinxTimeTotalSearch(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_SEARCH));
      }

      if (statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH)){
        setStatSphinxRequestsSearch(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH));
      }

      if (statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_ADMIN)){
        setStatSphinxTimeTotalAdmin(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_TIME_TOTAL_ADMIN));
      }

      if (statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN)){
        setStatSphinxRequestsAdmin(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN));
      }

      if (statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH)){
        setStatSphinxRequestsSearch(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_REQUESTS_SEARCH));
      }

      if (statProperties.isExist(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN)){
        setStatSphinxRequestsAdmin(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_SPHINX_REQUESTS_ADMIN));
      }

      if (statProperties.isExist(HCE::handlers::properties::STAT_DRCE_REQUESTS)){
        setStatDRCERequests(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_DRCE_REQUESTS));
      }

      if (statProperties.isExist(HCE::handlers::properties::STAT_DRCE_TIME_TOTAL)){
        setStatDRCETimeTotal(statProperties.get<Poco::UInt64>(HCE::handlers::properties::STAT_DRCE_TIME_TOTAL));
      }

      if (handlerProperties.isExist(HCE::handlers::properties::NODE_NUMBER)){
        setNodeNumber(handlerProperties.get<unsigned int>(HCE::handlers::properties::NODE_NUMBER));
      }
    }

    void DataProcessorData::restoreSphinxProperties(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Restore Sphinx FO properties" << flush;
      if (_sphinxFO){

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_INDEX_NAME)){
          _sphinxFO->setIndexName(handlerProperties.get<std::string>(HCE::handlers::properties::SPHINX_INDEX_NAME));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_HOME_DIR)){
          _sphinxFO->setHomeDir(handlerProperties.get<std::string>(HCE::handlers::properties::SPHINX_HOME_DIR));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_SEARCHD_HOST)){
          _sphinxFO->setServerHost(handlerProperties.get<std::string>(HCE::handlers::properties::SPHINX_SEARCHD_HOST));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_SEARCHD_PORT)){
          _sphinxFO->setServerPort(handlerProperties.get<unsigned int>(HCE::handlers::properties::SPHINX_SEARCHD_PORT));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_TIMEOUTED_REQUESTS)){
          _sphinxFO->setTimeoutedRequests(handlerProperties.get<Poco::UInt64>(HCE::handlers::properties::SPHINX_TIMEOUTED_REQUESTS));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_RANKER)){
          _sphinxFO->setRankingMode(handlerProperties.get<unsigned int>(HCE::handlers::properties::SPHINX_RANKER));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_RANKER_EXPRESSION)){
          _sphinxFO->setRankerExpression(handlerProperties.get<std::string>(HCE::handlers::properties::SPHINX_RANKER_EXPRESSION));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_NODE_NUMBER)){
          _sphinxFO->setNodeNumber(handlerProperties.get<unsigned int>(HCE::handlers::properties::SPHINX_NODE_NUMBER));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::SPHINX_NODE_NAME)){
          _sphinxFO->setNodeName(handlerProperties.get<std::string>(HCE::handlers::properties::SPHINX_NODE_NAME));
        }
      }
    }

    void DataProcessorData::restoreDRCEProperties(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Restore DRCE FO properties" << flush;
      if (_drceFO){

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_NODE_NAME)){
          _drceFO->setNodeName(handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_NODE_NAME));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_NODE_HOST)){
          _drceFO->setNodeHost(handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_NODE_HOST));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_NODE_PORT)){
          _drceFO->setNodePort(handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_NODE_PORT));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_HOME_DIR)){
          _drceFO->setHomeDir(handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_HOME_DIR));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_TASK_DATA_DIR)){
          _drceFO->setTasksDataDir(handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_TASK_DATA_DIR));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_TASK_STATUS_DIR)){
          _drceFO->setTasksStatusDir(handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_TASK_STATUS_DIR));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_MAX_THREAD_COUNT)){
          _drceFO->setMaxThreadCount(handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_MAX_THREAD_COUNT));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_TASK_QUEUE_DUMP_PERIOD)){
          _drceFO->setTasksQueueDumpPeriod(handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_TASK_QUEUE_DUMP_PERIOD));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCE_MONITOR_TIME_PERIOD)){
          _drceFO->setResourceMonitorTimePeriod(handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_RESOURCE_MONITOR_TIME_PERIOD));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS)){
          _drceFO->setResourceExtractorMaxThreadsCount(handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES)){
          _drceFO->setResourceExtractorMaxProcessesCount(handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE)){
          _drceFO->setResourceExtractorMaxDiskSize(handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE)){
          _drceFO->setResourceExtractorMaxVramSize(handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE));
        }

        if (handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE)){
          _drceFO->setResourceExtractorMaxRramSize(handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE));
        }

        _drceFO->restoreResourceMonitor(statProperties.getItems());
      }
    }

    bool DataProcessorData::rebuildTasksStateNotificationConnection(){
      bool ret = true;

      if(getStateNotificationHost() != "" && getStateNotificationPort() != ""){
        //Init the tasks state notification update connection
        ret = rebuildClientConnection(_tasksStateNotificationSock, ZMQ_PROTOCOL_PREFIX_TCP + getStateNotificationHost() + ":" + getStateNotificationPort(), ZMQ_DEALER, getNodeName(), -1,
            std::string(""), std::string(""));
      }else{
        ret = false;
      }

      return ret;
    }

    void DataProcessorData::dataStorageInitSphinx(const std::string& iniFile){
      //!<Sphinx FO init parameters
      std::string sphinx_home_dir = "";
      std::string sphinx_default_index;
      bool sphinx_start_searchd = NODE_FUNCTIONAL_SPHINX_NOT_START_SEARCHD;
      bool sphinx_stop_searchd = NODE_FUNCTIONAL_SPHINX_NOT_STOP_SEARCHD;
      unsigned int sphinx_ranker = NODE_FUNCTIONAL_SPHINX_RANKER_0;
      std::string sphinx_ranker_expression = "";

      logger.log(NODE_LOG_MODE_INFORMATION) << "load Sphinx FO settings from " << iniFile << " file" << flush;

      //!<Process FO's ini file
      try{
        Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
        try{

          if (!handlerProperties.isExist(HCE::handlers::properties::SPHINX_HOME_DIR)){
            sphinx_home_dir = pConf->getString(HCE::handlers::properties::SPHINX_HOME_DIR);
            handlerProperties.set(HCE::handlers::properties::SPHINX_HOME_DIR, sphinx_home_dir);
          }else{
            sphinx_home_dir = handlerProperties.get<std::string>(HCE::handlers::properties::SPHINX_HOME_DIR);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::SPHINX_DEFAULT_INDEX)){
            sphinx_default_index = pConf->getString(HCE::handlers::properties::SPHINX_DEFAULT_INDEX);
            handlerProperties.set(HCE::handlers::properties::SPHINX_DEFAULT_INDEX, sphinx_default_index);
          }else{
            sphinx_default_index = handlerProperties.get<std::string>(HCE::handlers::properties::SPHINX_DEFAULT_INDEX);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::SPHINX_START_SEARCHD)){
            sphinx_start_searchd = pConf->getBool(HCE::handlers::properties::SPHINX_START_SEARCHD);
            handlerProperties.set(HCE::handlers::properties::SPHINX_START_SEARCHD, sphinx_start_searchd);
          }else{
            sphinx_start_searchd = handlerProperties.get<bool>(HCE::handlers::properties::SPHINX_START_SEARCHD);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::SPHINX_STOP_SEARCHD)){
            sphinx_stop_searchd = pConf->getBool(HCE::handlers::properties::SPHINX_STOP_SEARCHD);
            handlerProperties.set(HCE::handlers::properties::SPHINX_STOP_SEARCHD, sphinx_stop_searchd);
          }else{
            sphinx_stop_searchd = handlerProperties.get<bool>(HCE::handlers::properties::SPHINX_STOP_SEARCHD);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::SPHINX_RANKER)){
            sphinx_ranker = pConf->getInt(HCE::handlers::properties::SPHINX_RANKER);
            handlerProperties.set(HCE::handlers::properties::SPHINX_RANKER, sphinx_ranker);
          }else{
            sphinx_ranker = handlerProperties.get<unsigned int>(HCE::handlers::properties::SPHINX_RANKER);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::SPHINX_RANKER_EXPRESSION)){
            sphinx_ranker_expression = pConf->getString(HCE::handlers::properties::SPHINX_RANKER_EXPRESSION);
            handlerProperties.set(HCE::handlers::properties::SPHINX_RANKER_EXPRESSION, sphinx_ranker_expression);
          }else{
            sphinx_ranker_expression = handlerProperties.get<std::string>(HCE::handlers::properties::SPHINX_RANKER_EXPRESSION);
          }

        } catch(Poco::SyntaxException& e){
          //!<Log some error of config file processing
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " data node ini file processing error: " << e.displayText() << flush;
        }
      } catch(Poco::Exception& e){
        //!<Log some error of config file processing
        logger.log(NODE_LOG_MODE_ERROR) << " data node ini file access error: " << e.displayText() << flush;
      }

      //If Sphinx home dir not set, define it as current user home dir
      if(sphinx_home_dir == ""){
        sphinx_home_dir = getUserHomeDir() + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_PROJECT_ROOT_DIR;
      }else{
        if(sphinx_home_dir[0] != NODE_INI_FILE_PATH_START_CHAR){
          sphinx_home_dir = getUserHomeDir() + NODE_INI_FILE_PATH_DELIMITER_CHAR + sphinx_home_dir;
        }
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << "Sphinx FO initialization:" << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "nodeName=" << getNodeName() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "home_dir=" << sphinx_home_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "default_index=" << sphinx_default_index << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "start_searchd=" << sphinx_start_searchd << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "stop_searchd=" << sphinx_stop_searchd << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "getNodeNumber()=" << getNodeNumber() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "ranker=" << sphinx_ranker << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "ranker_expression=" << sphinx_ranker_expression << flush;

      //!<Create Sphinx Sphinx FO and init
      _sphinxFO = new HCE::sphinx::SphinxFunctionalObject(getNodeName(), getNodeNumber(), sphinx_home_dir, sphinx_default_index, sphinx_start_searchd, sphinx_stop_searchd);
      if(_sphinxFO){
        if(_sphinxFO->getErrorCode()){
          //!<Log some error of Sphinx FO
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " Sphinx FO initialization error " << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg()
              << endl << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;
        }
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << NODE_FUNCTIONAL_ERROR_SPHINX << " FO instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
        ::exit(NODE_EXIT_CODE_SPHINX_CRITICAL_ERROR);
      }

      //!<Turn checking the incoming data ON
      _sphinxFO->setCheckIncomingDocuments(true);

      //!<Set ranker
      _sphinxFO->setRankingMode(sphinx_ranker, sphinx_ranker_expression);

      logger.log(NODE_LOG_MODE_DEBUG) << "SPHO_INSTANTIATE [" << _sphinxFO->logMsg() << "]" << flush;
    }

    void DataProcessorData::dataStorageInitDRCE(const std::string& iniFile){
      //!<DRCE FO init parameters
      std::string drce_home_dir = "";
      std::string drce_tmp_dir = "";
      std::string drce_node_host = "";
      std::string drce_node_port = "";
      std::string drce_tasks_data_dir = "";
      std::string drce_tasks_status_dir = "";
      unsigned int drce_max_threads = 32;
      unsigned int drce_tasks_queue_dump_period = 10000;  // time interval period in msec.
      unsigned int drce_resources_monitor_period = 10000; // time interval period in msec.
      std::string drce_env_vars = "";
      size_t drce_resources_extractor_max_threads = 1024;
      size_t drce_resources_extractor_max_processes = 1024;
      size_t drce_resources_extractor_max_disk_size = 0;
      size_t drce_resources_extractor_max_vram_size = 0;
      size_t drce_resources_extractor_max_rram_size = 0;

      logger.log(NODE_LOG_MODE_INFORMATION) << "load DRCE FO settings from " << iniFile << " file" << flush;

      //!<Process FO's ini file
      try{
        Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
        try{

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_HOME_DIR)){
            drce_home_dir = pConf->getString(HCE::handlers::properties::DRCE_HOME_DIR);
          }else{
            drce_home_dir = handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_HOME_DIR);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_TMP_DIR)){
            drce_tmp_dir = pConf->getString(HCE::handlers::properties::DRCE_TMP_DIR);
          }else{
            drce_tmp_dir = handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_TMP_DIR);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_TASKS_DATA_DIR)){
            drce_tasks_data_dir = pConf->getString(HCE::handlers::properties::DRCE_TASKS_DATA_DIR);
            handlerProperties.set(HCE::handlers::properties::DRCE_TASKS_DATA_DIR, drce_tasks_data_dir);
          }else{
            drce_tasks_data_dir = handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_TASKS_DATA_DIR);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_TASKS_STATUS_DIR)){
            drce_tasks_status_dir = pConf->getString(HCE::handlers::properties::DRCE_TASKS_STATUS_DIR);
            handlerProperties.set(HCE::handlers::properties::DRCE_TASKS_STATUS_DIR, drce_tasks_status_dir);
          }else{
            drce_tasks_status_dir = handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_TASKS_STATUS_DIR);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_NODE_HOST)){
            drce_node_host = pConf->getString(HCE::handlers::properties::DRCE_NODE_HOST);
            handlerProperties.set(HCE::handlers::properties::DRCE_NODE_HOST, drce_node_host);
          }else{
            drce_node_host = handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_NODE_HOST);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_NODE_PORT)){
            drce_node_port = pConf->getString(HCE::handlers::properties::DRCE_NODE_PORT);
            handlerProperties.set(HCE::handlers::properties::DRCE_NODE_PORT, drce_node_port);
          }else{
            drce_node_port = handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_NODE_PORT);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_MAX_THREADS)){
            drce_max_threads = pConf->getInt(HCE::handlers::properties::DRCE_MAX_THREADS);
            handlerProperties.set(HCE::handlers::properties::DRCE_MAX_THREADS, drce_max_threads);
          }else{
            drce_max_threads = handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_MAX_THREADS);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_TASKS_QUEUE_DUMP_PERIOD)){
            drce_tasks_queue_dump_period = pConf->getInt(HCE::handlers::properties::DRCE_TASKS_QUEUE_DUMP_PERIOD);
            handlerProperties.set(HCE::handlers::properties::DRCE_TASKS_QUEUE_DUMP_PERIOD, drce_tasks_queue_dump_period);
          }else{
            drce_tasks_queue_dump_period = handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_TASKS_QUEUE_DUMP_PERIOD);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_MONITOR_PERIOD)){
            drce_resources_monitor_period = pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_MONITOR_PERIOD);
            handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_MONITOR_PERIOD, drce_resources_monitor_period);
          }else{
            drce_resources_monitor_period = handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_RESOURCES_MONITOR_PERIOD);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_HOST)){
            setStateNotificationHost(pConf->getString(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_HOST));
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_PORT)){
            setStateNotificationPort(pConf->getString(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_PORT));
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_RESET_ERROR_CODE)){
            bool drce_state_notification_reset_error_state = pConf->getBool(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_RESET_ERROR_CODE);
            handlerProperties.set(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_RESET_ERROR_CODE, drce_state_notification_reset_error_state);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_ENV_VARS)){
            drce_env_vars = pConf->getString(HCE::handlers::properties::DRCE_ENV_VARS);
            handlerProperties.set(HCE::handlers::properties::DRCE_ENV_VARS, drce_env_vars);
          }else{
            drce_env_vars = handlerProperties.get<std::string>(HCE::handlers::properties::DRCE_ENV_VARS);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS)){
            drce_resources_extractor_max_threads = pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS);
            handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS, drce_resources_extractor_max_threads);
          }else{
            drce_resources_extractor_max_threads = handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES)){
            drce_resources_extractor_max_processes = pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES);
            handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES, drce_resources_extractor_max_processes);
          }else{
            drce_resources_extractor_max_processes = handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE)){
            drce_resources_extractor_max_disk_size = pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE);
            handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE, drce_resources_extractor_max_disk_size);
          }else{
            drce_resources_extractor_max_disk_size = handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE)){
            drce_resources_extractor_max_vram_size = pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE);
            handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE, drce_resources_extractor_max_vram_size);
          }else{
            drce_resources_extractor_max_vram_size = handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE);
          }

          if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE)){
            drce_resources_extractor_max_rram_size = pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE);
            handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE, drce_resources_extractor_max_rram_size);
          }else{
            drce_resources_extractor_max_rram_size = handlerProperties.get<size_t>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE);
          }

        } catch(Poco::SyntaxException& e){
          //!<Log some error of config file processing
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " data node ini file processing error: " << e.displayText() << flush;
        }
      } catch(Poco::Exception& e){
        //!<Log some error of config file processing
        logger.log(NODE_LOG_MODE_ERROR) << " data node ini file access error: " << e.displayText() << flush;
      }

      //If DRCE home dir not set, define it as current user home dir
      if(drce_home_dir == ""){
        drce_home_dir = getUserHomeDir() + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_PROJECT_ROOT_DIR;
      }else{
        //If first path character is not slash, this means that the path specified is relative to the home dir
        if(drce_home_dir[0] != NODE_INI_FILE_PATH_START_CHAR){
          drce_home_dir = getUserHomeDir() + NODE_INI_FILE_PATH_DELIMITER_CHAR + drce_home_dir;
        }
      }

      //If drce_tasks_data_dir not set
      if(drce_tasks_data_dir == ""){
        drce_tasks_data_dir = drce_home_dir + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_DRCE_TASKS_DATA_DIR;
      }else{
        //If first path character is not slash, this means that the path specified is relative to the home dir
        if(drce_tasks_data_dir[0] != NODE_INI_FILE_PATH_START_CHAR){
          drce_tasks_data_dir = drce_home_dir + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_DRCE_TASKS_DATA_DIR;
        }
      }

      //If drce_tasks_status_dir not set
      if(drce_tasks_status_dir == ""){
        drce_tasks_status_dir = drce_home_dir + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_DRCE_TASKS_STATUS_DIR;
      }else{
        //If first path character is not slash, this means that the path specified is relative to the home dir
        if(drce_tasks_status_dir[0] != NODE_INI_FILE_PATH_START_CHAR){
          drce_tasks_status_dir = drce_home_dir + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_DRCE_TASKS_STATUS_DIR;
        }
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << "DRCE FO initialization:" << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "nodeName=" << getNodeName() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "home_dir=" << drce_home_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "tmp_dir=" << drce_tmp_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_tasks_data_dir=" << drce_tasks_data_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_tasks_status_dir=" << drce_tasks_status_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_node_host=" << drce_node_host << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_node_port=" << drce_node_port << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_max_threads=" << drce_max_threads << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_tasks_queue_dump_period=" << drce_tasks_queue_dump_period << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_resources_monitor_period=" << drce_resources_monitor_period << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "state_notification_host=" << getStateNotificationHost() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "state_notification_port=" << getStateNotificationPort() << flush;

      handlerProperties.set(HCE::handlers::properties::DRCE_HOME_DIR, drce_home_dir);

      //!<Create DRCE FO and init
      _drceFO = new HCE::drce::DRCEFunctionalObject(getNodeName(), drce_home_dir, drce_tmp_dir, drce_tasks_data_dir, drce_tasks_status_dir, drce_node_host, drce_node_port);

      if(_drceFO){
        if(_drceFO->getErrorCode()){
          //!<Log some error of DRCE FO
          logger.log(NODE_LOG_MODE_FATAL) << NODE_FUNCTIONAL_ERROR_DRCE << " DRCE FO initialization error " << _drceFO->getErrorCode() << " : " << _drceFO->getErrorMsg()
              << endl << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN << _drceFO->logMsg() << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END << flush;
          ::exit(NODE_EXIT_CODE_DRCE_CRITICAL_ERROR);
        }
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << NODE_FUNCTIONAL_ERROR_DRCE << " FO instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
        ::exit(NODE_EXIT_CODE_DRCE_CRITICAL_ERROR);
      }

      //!<Set max threads count for DRCE workers threads pool
      _drceFO->setMaxThreadCount(drce_max_threads);
      //!<Set tasks queue dump period interval (in msec.)
      _drceFO->setTasksQueueDumpPeriod(drce_tasks_queue_dump_period);
      //!<Set resource monitor time period (in msec.)
      _drceFO->setResourceMonitorTimePeriod(drce_resources_monitor_period);
      //!<Set mandatory environment variables
      _drceFO->setEnvironments(drce_env_vars);
      //!<Set max threads count for resource extractor
      _drceFO->setResourceExtractorMaxThreadsCount(drce_resources_extractor_max_threads);
      //!<Set max processes count for resource extractor
      _drceFO->setResourceExtractorMaxProcessesCount(drce_resources_extractor_max_processes);
      //!<Set max disk size for resource extractor
      if (drce_resources_extractor_max_disk_size)
        _drceFO->setResourceExtractorMaxDiskSize(drce_resources_extractor_max_disk_size);
      //!<Set max virtual memory size for resource extractor
      if (drce_resources_extractor_max_vram_size)
        _drceFO->setResourceExtractorMaxVramSize(drce_resources_extractor_max_vram_size);
      //!<Set max rss memory size for resource extractor
      if (drce_resources_extractor_max_rram_size)
        _drceFO->setResourceExtractorMaxRramSize(drce_resources_extractor_max_rram_size);

      //!<Create connection socket and connect to tasks state notifications service
      recoverNotificationConnection();
      //!<Output DRCE log message to HCE logger
      logger.log(NODE_LOG_MODE_DEBUG) << "DRCE_FO_INSTANTIATE [" << _drceFO->logMsg() << "]" << flush;
    }

    bool DataProcessorData::recoverNotificationConnection(void){
      bool result = false;
      //!<Create connection socket and connect to tasks state notifications service
      if(rebuildTasksStateNotificationConnection()){
        logger.log(NODE_LOG_MODE_DEBUG) << "Tasks state notification connection created!" << flush;
        //!<Create instance or the notification functor
        _tasksStateNotificationFunctor = new TasksStateNotificationFunctor(_tasksStateNotificationSock);
        //!<Set notification function to DRCE FO instance
        _drceFO->setNotificationFunctor(*_tasksStateNotificationFunctor);
        //!<Set flag reset error state for notification
        _drceFO->setResetErrorCodeStateNotification(handlerProperties.get<bool>(HCE::handlers::properties::DRCE_STATE_NOTIFICATION_RESET_ERROR_CODE));
        logger.log(NODE_LOG_MODE_DEBUG) << "Tasks state notification functor is set!" << flush;
        result = true;
      }
      return result;
    }

    std::string DataProcessorData::getUserHomeDir(void){
      std::string homeDir;

      struct passwd *pw = getpwuid(getuid());
      const char *homedir = pw->pw_dir;
      if(homedir){
        homeDir = homedir;
      }

      return homeDir;
    }

    void DataProcessorData::nodeInit(const std::string& iniFile){
      //!<Node init parameters
      unsigned int processingMode = NODE_DATA_PROCESSING_MODE_NORMAL;

      logger.log(NODE_LOG_MODE_INFORMATION) << "load node settings from " << iniFile << " file" << flush;

      //!<Process Node's ini file
      try{
        Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
        try{
          //Set node number from ini
          if (!handlerProperties.isExist(HCE::handlers::properties::NODE_NUMBER))
            setNodeNumber(pConf->getInt(HCE::handlers::properties::NODE_NUMBER));

          //Set data processing mode from ini
          if (!handlerProperties.isExist(HCE::handlers::properties::DATA_PROCESSING_MODE))
            processingMode = pConf->getInt(HCE::handlers::properties::DATA_PROCESSING_MODE);

          setDataProcessingMode(processingMode);

          if(processingMode > 0){
            if (!handlerProperties.isExist(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS)){
              setDataProcessingFakeResults(pConf->getInt(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS));
            }
            if (!handlerProperties.isExist(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS_MAX)){
              setDataProcessingFakeResultsMax(pConf->getInt(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS_MAX));
            }
          }
        } catch(Poco::SyntaxException& e){
          //!<Log some error of config file processing
          logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node ini file processing error: " << e.displayText() << flush;
        }
      } catch(Poco::Exception& e){
        //!<Log some error of config file processing
        logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node ini file access error: " << e.displayText() << flush;
      }

      if (!handlerProperties.isExist(HCE::handlers::properties::NODE_NUMBER))
        setNodeNumber(0);

      if (!handlerProperties.isExist(HCE::handlers::properties::DATA_PROCESSING_MODE))
        setDataProcessingMode(0);

      if (!handlerProperties.isExist(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS)){
        setDataProcessingFakeResults(0);
      }

      if (!handlerProperties.isExist(HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS_MAX)){
        setDataProcessingFakeResultsMax(0);
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << HCE::handlers::properties::NODE_NUMBER << ": " << getNodeNumber() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << HCE::handlers::properties::DATA_PROCESSING_MODE << ": " << getDataProcessingMode() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS << ": " << getDataProcessingFakeResults() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << HCE::handlers::properties::DATA_PROCESSING_FAKE_RESULTS_MAX << ": " << getDataProcessingFakeResultsMax() << flush;
    }

    std::string DataProcessorData::processDataRequest(const std::string& messageBody){
      std::string ret;

      if(messageBody.size() > 0 && messageBody[0] == NODE_JSON_BEGIN_CHAR){
        logger.log(NODE_LOG_MODE_TRACE) << "JSON processing..." << flush;

        //!<Possible json message
        ret = processDataRequestJson(messageBody);
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "PLAIN_TEXT processing..." << flush;

        //!<Plain text message
        ret = processDataRequestPlainText(messageBody);
      }

      return ret;
    }

    std::string DataProcessorData::processDataRequestJson(const std::string& messageBody){
      //Response json string
      std::string ret;

      //Create default empty cover object for results
      HCE::JsonMessageCover msgCoverOutEmpty;

      //Serialize default empty cover object
      if(!msgCoverOutEmpty.serialize(ret)){
        logger.log(NODE_LOG_MODE_DEBUG) << "Default empty cover object serialize error " << msgCoverOutEmpty.getErrorCode() << " : " << msgCoverOutEmpty.getErrorMsg() << flush;
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "Request json : " << messageBody << flush;

      //Unserialize request json
      HCE::JsonMessageCover msgCoverIn(messageBody);

      if(msgCoverIn.getErrorCode()){
        logger.log(NODE_LOG_MODE_ERROR) << "Data request cover json deserialization error : " << msgCoverIn.getErrorMsg() << flush;
      }else{
        logger.log(NODE_LOG_MODE_DEBUG) << "Json request object from cover: [" << msgCoverIn.getData() << "]" << flush;

        if(msgCoverIn.getType() == HCE::types::MessageType::mtSphinx){
          //Process Sphinx request type
          std::string sphinxResultDataJson = processDataRequestSphinx(msgCoverIn.getData());

          //Create cover object for results
          HCE::JsonMessageCover msgCoverOut(msgCoverIn.getType(), sphinxResultDataJson, msgCoverIn.getTTL());

          //Serialize output cover object
          if(!msgCoverOut.serialize(ret)){
            //!<If serialization error return default result to reducer
            logger.log(NODE_LOG_MODE_ERROR) << "Cover object for Sphinx request serialize error " << msgCoverOut.getErrorCode() << " : " << msgCoverOut.getErrorMsg() << flush;
          }
        }else{
          if(msgCoverIn.getType() == HCE::types::MessageType::mtDrce){
            //Process DRCE request type
            std::string drceResultDataJson = processDataRequestDRCE(msgCoverIn.getData());

            //Create cover object for results
            HCE::JsonMessageCover msgCoverOut(msgCoverIn.getType(), drceResultDataJson, msgCoverIn.getTTL());

            //Serialize output cover object
            if(!msgCoverOut.serialize(ret)){
              //!<If serialization error return default result to reducer
              logger.log(NODE_LOG_MODE_ERROR) << "Cover object for DRCE request serialize error " << msgCoverOut.getErrorCode() << " : " << msgCoverOut.getErrorMsg() << flush;
            }
          }else{
            logger.log(NODE_LOG_MODE_ERROR) << "Data request type error : type " << static_cast<int>(msgCoverIn.getType()) << " not supported" << flush;
          }
        }
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "Result json : " << ret << flush;

      return ret;
    }

    std::string DataProcessorData::processDataRequestSphinx(const std::string& messageBody){
      //Create empty Sphinx result object to return it for farther processing by default
      HCE::sphinx::SphinxDefaultJSON defaultJson;
      std::string ret = defaultJson.getJSON();

      //Create message cover object and fill it with default results json to return it to upper level processing (reducing) in case of some error happened
      HCE::JsonMessageCover msgCover(HCE::types::MessageType::mtSphinx, ret);
      if(!msgCover.serialize(ret)){
        //!<If serialization error return default result to reducer
        logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " cover object from default results - serialization error "
            << msgCover.getErrorCode() << " : " << msgCover.getErrorMsg() << flush;
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "Sphinx data handler request, json=[" << messageBody << "]" << flush;

      //If node data processing mode is normal - enter FO usage, else in simulation mode - return default results
      if(getDataProcessingMode() == NODE_DATA_PROCESSING_MODE_NORMAL){
        //!<Create input object for Sphinx FO
        HCE::sphinx::SphinxInputJsonMessage inputMsg(messageBody);
        if(!inputMsg.isError()){
          if(_sphinxFO->getErrorCode()){
            //!<Return empty string to reducer in case of searcher state error
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " searcher state error " << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg()
                << endl << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;

            //!<Reset searcher error state to get next message to start from scratch
            _sphinxFO->resetError();
          }else{
            //!<Execute search request
            HCE::sphinx::SphinxOutputJsonMessage outputMessage = _sphinxFO->Process(inputMsg);

            //TODO temporary for filters debug reason, need to be removed in release
            logger.log(NODE_LOG_MODE_DEBUG) << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;
            //TODO

            if(_sphinxFO->getErrorCode()){
              //!<Log some error of Sphinx FO
              logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " search process error " << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg()
                  << endl << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;
              //!<Reset searcher error state to get next message to start from scratch
              _sphinxFO->resetError();
            }

            if(outputMessage.getErrorCode()){
              //!<Log some error of search request execution
              logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " output message error " << outputMessage.getErrorCode() << " : " << outputMessage.getErrorMessage() << flush;
            }

            logger.log(NODE_LOG_MODE_DEBUG) << "Sphinx time:" << outputMessage.getTime() << flush;

            //!<Return Sphinx results data json string of data from Sphinx FO
            ret = outputMessage.getData();

            //!<Register Sphinx request time
            setStatSphinxTimeTotalSearch(getStatSphinxTimeTotalSearch() + outputMessage.getTime());
          }
        }else{
          //!<Return empty json string to reducer in case of de-serialisation error to direct the processing on json type
          logger.log(NODE_LOG_MODE_DEBUG) << NODE_FUNCTIONAL_ERROR_SPHINX << " deserialization error " << inputMsg.getErrorCode() << " : " << inputMsg.getErrorMsg() << flush;
        }
      }else{
        //Fake processing mode, random generator of results set
        //Create message cover object and fill it with default results json to return it to upper level processing (reducing) in case of some error happened
        if(defaultJson.makeJSON(getDataProcessingFakeResults(), 1, getDataProcessingFakeResultsMax())){
          //Return Sphinx results data json string of data from fake results generator
          ret = defaultJson.getJSON();
        }else{
          //TODO
          logger.log(NODE_LOG_MODE_ERROR) << "Error create fake results " << getDataProcessingFakeResults() << ", zero results returned for Sphinx" << flush;
        }
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "Sphinx results json : " << ret << flush;

      //Register Sphinx requests number stat
      unsigned long long statSphinxRequestsSearch = getStatSphinxRequestsSearch();
      setStatSphinxRequestsSearch(++statSphinxRequestsSearch);

      return ret;
    }

    std::string DataProcessorData::processDataRequestDRCE(const std::string& messageBody){
      //Create empty DRCE result object to return it for farther processing by default
      HCE::drce::DRCEDefaultJSON defaultJson;
      std::string ret = defaultJson.getJSON();

      logger.log(NODE_LOG_MODE_DEBUG) << "DRCE data handler request, json=[" << messageBody << "]" << flush;

      //If node data processing mode is normal - enter FO usage, else in simulation mode - return default results
      if(getDataProcessingMode() == NODE_DATA_PROCESSING_MODE_NORMAL){
        //!<Create input object for DRCE FO
        HCE::drce::DRCEInputJsonMessage inputMsg(messageBody);
        if(!inputMsg.isError()){
          if(_drceFO->getErrorCode()){
            //!<Return empty string to reducer in case of DRCE state error
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " DRCE state error " << _drceFO->getErrorCode() << " : " << _drceFO->getErrorMsg()
                << endl << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN << _drceFO->logMsg() << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END << flush;

            //!<Reset DRCE error state to get next message to start from scratch
            _drceFO->resetError();
          }else{
            //!<Execute DRCE request
            HCE::drce::DRCEOutputJsonMessage outputMessage = _drceFO->Process(inputMsg);

            //TODO temporary for filters debug reason, need to be removed in release
            logger.log(NODE_LOG_MODE_DEBUG) << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN << _drceFO->logMsg() << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END << flush;

            if(_drceFO->getErrorCode()){
              //!<Log some error of DRCE FO
              logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " execution error " << _drceFO->getErrorCode() << " : " << _drceFO->getErrorMsg()
                  << endl << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN << _drceFO->logMsg() << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END << flush;

              //!<Reset DRCE error state to get next message to start from scratch
              _drceFO->resetError();
            }

            if(outputMessage.getErrorCode()){
              //!<Log some error of DRCE request execution
              logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " output message error " << outputMessage.getErrorCode() << " : " << outputMessage.getErrorMsg() << flush;
            }
            logger.log(NODE_LOG_MODE_DEBUG) << "DRCE time:" << _drceFO->getRequestTime() << flush;

            //!<Return DRCE results data json string of data from the DRCEOutputJsonMessage object
            outputMessage.serialize(ret);

            //!<Register DRCE request time
            setStatDRCETimeTotal(getStatDRCETimeTotal() + _drceFO->getRequestTime());
          }
        }else{
          //!<Return empty json string to reducer in case of de-serialisation error to direct the processing on json type
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " deserialization error " << inputMsg.getErrorCode() << " : " << inputMsg.getErrorMsg() << flush;
          //!<Fill error code and error message for default json
          defaultJson.setJsonErrorCode(inputMsg.getErrorCode());
          defaultJson.setJsonErrorMessage(inputMsg.getErrorMsg());
          defaultJson.setJsonRequestType(inputMsg.getRequestType());
          //defaultJson.setJsonRequestId(inputMsg.getRequestId());
          defaultJson.setJsonNodeHost(_drceFO->getNodeHost());
          defaultJson.setJsonNodePort(_drceFO->getNodePort());
          defaultJson.setJsonNodeName(_drceFO->getNodeName());
          defaultJson.setJsonState(DRCE_REQUEST_STATE_ERROR_SET_AS_NEW_TASK);
          defaultJson.makeJSON();
          ret = defaultJson.getJSON();
        }
      }else{
        //Fake processing mode, random generator of results set
        //Create message cover object and fill it with default results json to return it to upper level processing (reducing) in case of some error happened
        if(defaultJson.makeJSON(getDataProcessingFakeResults())){
          //Return DRCE results data json string of data from fake results generator
          ret = defaultJson.getJSON();
        }else{
          logger.log(NODE_LOG_MODE_ERROR) << "Error create fake results " << getDataProcessingFakeResults() << ", zero results returned for DRCE" << flush;
        }
      }

      //TODO: test, need to be removed
      //ret = DRCE_FAKE_RESPONSE;

      logger.log(NODE_LOG_MODE_DEBUG) << "DRCE results json : " << ret << flush;

      //!<Register DRCE requests number stat
      unsigned long long statDRCERequests = getStatDRCERequests();
      setStatDRCERequests(++statDRCERequests);

      return ret;
    }

    std::string DataProcessorData::processDataRequestPlainText(const std::string& messageBody){
      //!<Parse request and fetch items number
      unsigned int items_number;
      std::istringstream(messageBody) >> items_number;
      //!<Generate items
      std::vector<unsigned long long> items(items_number);
      //!<Randomize items values
      if(items_number > 0){
        for(unsigned long long i = 0; i < items.size(); ++i){
          items[i] = (rand() % RAND_MAX) + 1;
        }
      }else{
        items.resize(1);
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "DATA items requested:" << items_number << ", items generated:" << items.size() << flush;

      //!<Sord items
      sort(items.begin(), items.end(), std::less<unsigned long long>());

      logger.log(NODE_LOG_MODE_DEBUG) << "DATA sorted" << flush;
      //!<Serialize items
      std::ostringstream oss;
      std::copy(items.begin(), items.end() - 1, std::ostream_iterator<int>(oss, " "));
      oss << items.back();
      std::string body(oss.str());

      return body;
    }

    void DataProcessorData::handleInternalMessage(void){
      //!<Fetch message
      zmsg msg(*_inprocDataSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_data" << endl << msg.dump() << flush;
      //!<Get identity and body of request message message
      std::string messageId((const char*)msg.pop_front().c_str());
      std::string messageBody((const char*)msg.pop_front().c_str());
      std::string messageRoute((msg.parts())?(const char*)msg.pop_front().c_str():"");

      unsigned long long statClientResponses = getStatClientResponses();
      setStatClientResponses(++statClientResponses);

      logger.log(NODE_LOG_MODE_DEBUG) << "DATA PROCESSOR RECEIVED message [" << messageId << "] body " << messageBody.size() << " bytes n=" << getStatClientResponses() << flush;

      //!<Process data request
      std::string body = processDataRequest(messageBody);

      //!<Create response message
      msg.append(messageId.c_str());
      msg.append(body.c_str());
      msg.append(messageRoute.c_str());

      logger.log(NODE_LOG_MODE_TRACE) << "DATA PROCESSOR RESPONSE message [" << messageId << "] body [" << body << "]" << flush;

      //!<Send response
      msg.send(*_inprocDataSock);

      logger.log(NODE_LOG_MODE_DEBUG) << "DATA PROCESSOR RESPONSE message [" << messageId << "] body " << body.size() << " bytes sent" << flush;
    }

    std::string DataProcessorData::processAdminCommands(std::string & command, std::string & parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << getStatClientResponses() << NODE_ADMIN_COMMAND_DELIMITER << "avgSphinxSearchTime="
            << (getStatSphinxTimeTotalSearch() / (getStatSphinxRequestsSearch() + 1)) << NODE_ADMIN_COMMAND_DELIMITER << "avgSphinxAdminTime="
            << (getStatSphinxTimeTotalAdmin() / (getStatSphinxRequestsAdmin() + 1)) << NODE_ADMIN_COMMAND_DELIMITER << "sphinxSearchRequests=" << getStatSphinxRequestsSearch()
            << NODE_ADMIN_COMMAND_DELIMITER << "sphinxAdminRequests=" << getStatSphinxRequestsAdmin() << NODE_ADMIN_COMMAND_DELIMITER << "DRCERequests=" << getStatDRCERequests()
            << NODE_ADMIN_COMMAND_DELIMITER << "avgDRCETime=" << (getStatDRCETimeTotal() / (getStatDRCERequests() + 1));
        ret = result.str();
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else if(command == NODE_MSG_SET_DATA_PROCESSING_MODE){
        std::stringstream result;

        //!<Parse command parameters and fetch processing mode value
        unsigned int processingMode;
        std::istringstream(parameters) >> processingMode;

        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << std::to_string(getDataProcessingMode()) << NODE_ADMIN_COMMAND_DELIMITER << getName();

        //Change data processing mode
        setDataProcessingMode(processingMode);

        ret = result.str();
      }else if(command == NODE_MSG_SPHINX){
        //!<Sphinx functional object command
        ret = processAdminCommandsSphinx(parameters);
      }else if(command == NODE_MSG_DRCE){
        //!<DRCE functional object command
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + processDataRequestDRCE(parameters);
        //ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + DRCE_FAKE_RESPONSE;
      }else if((command == NODE_MSG_DRCE_SET_HOST) || (command == NODE_MSG_DRCE_GET_HOST)){
        //!<Prepare response message
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + _drceFO->getNodeHost() + NODE_ADMIN_COMMAND_DELIMITER;
        if(command == NODE_MSG_DRCE_SET_HOST){
          //Change DRCE FO host
          _drceFO->setNodeHost(parameters);
        }
      }else if((command == NODE_MSG_DRCE_SET_PORT) || (command == NODE_MSG_DRCE_GET_PORT)){
        //!<Prepare response message
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + _drceFO->getNodePort() + NODE_ADMIN_COMMAND_DELIMITER;
        if(command == NODE_MSG_DRCE_SET_PORT){
          //Change DRCE FO port
          _drceFO->setNodePort(parameters);
        }
      }else if(command == NODE_MSG_DRCE_GET_TASKS){
        ret = _drceFO->getCurrentTasksQueue();
      }else if(command == NODE_MSG_DRCE_GET_TASKS_INFO){
        ret = _drceFO->getListInfoAllTasks();
      }else if(command == NODE_MSG_RESOURCE_USAGE){
        ret = _drceFO->getHostResourceUsage();
      }else if(command == NODE_MSG_RECOVER_NOTIFICATION_CONNECTION){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << std::boolalpha << recoverNotificationConnection() << NODE_ADMIN_COMMAND_DELIMITER;
        ret = result.str();
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    std::string DataProcessorData::getPropertyInformation(void){
      std::stringstream result;
      result  << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "responses=" << getStatClientResponses()
          << NODE_ADMIN_COMMAND_DELIMITER << "iniFile=" << getIniFile()
          << NODE_ADMIN_COMMAND_DELIMITER << "nodeName=" << getNodeName()
          << NODE_ADMIN_COMMAND_DELIMITER << "dataProcessingMode=" << getDataProcessingMode()
          << NODE_ADMIN_COMMAND_DELIMITER << "avgSphinxSearchTime=" << (getStatSphinxTimeTotalSearch() / (getStatSphinxRequestsSearch() + 1))
          << NODE_ADMIN_COMMAND_DELIMITER << "avgSphinxAdminTime=" << (getStatSphinxTimeTotalAdmin() / (getStatSphinxRequestsAdmin() + 1))
          << NODE_ADMIN_COMMAND_DELIMITER << "sphinxSearchRequests=" << getStatSphinxRequestsSearch()
          << NODE_ADMIN_COMMAND_DELIMITER << "sphinxAdminRequests=" << getStatSphinxRequestsAdmin()
          << NODE_ADMIN_COMMAND_DELIMITER << "sphinxDataStorageProperties=" << getDataStorageSphinxProperties()
          << NODE_ADMIN_COMMAND_DELIMITER << "drceDataStorageProperties=" << getDataStorageDRCEProperties()
          << NODE_ADMIN_COMMAND_DELIMITER << "DRCERequests=" << getStatDRCERequests()
          << NODE_ADMIN_COMMAND_DELIMITER << "avgDRCETime=" << (getStatDRCETimeTotal() / (getStatDRCERequests() + 1))
          << NODE_ADMIN_COMMAND_DELIMITER << "nodeNumber=" << getNodeNumber();
      return result.str();
    }

    std::string DataProcessorData::getDataStorageSphinxProperties(void){
      std::stringstream ret;
      ret << "isConnected" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->isConnected() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "indexName"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getIndexName() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "dataDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getDataDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "etcDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getEtcDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "homeDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getHomeDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "logDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getLogDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "runDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getRunDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "searchdHost"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getServerHost() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "searchdPort"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getServerPort() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "timeoutedRequests"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getTimeoutedRequests() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "ranker"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getRankingMode() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "ranker_expression"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << Poco::replace(_sphinxFO->getRankerExpression(), "=", "%3D") << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "nodeNumber"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getNodeNumber() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "nodeName"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getNodeName();

      return ret.str();
    }

    std::string DataProcessorData::getDataStorageDRCEProperties(void){
      std::stringstream ret;
      HCE::drce::StatVariables statVariables = _drceFO->getStatVariables();
      ret << "nodeName" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getNodeName() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "nodeHost" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getNodeHost() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "nodePort" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getNodePort() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "homeDir" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getHomeDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "tasksDataDir" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getTasksDataDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "tasksStatusDir" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getTasksStatusDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "maxThreadCount" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getMaxThreadCount() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "tasksQueueDumpPeriod" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getTasksQueueDumpPeriod() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "resourceMonitorTimePeriod" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getResourceMonitorTimePeriod() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "asyncTasksCount" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getAsyncTasksCount() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "requestTime" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getRequestTime() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "timeAsyncTask.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeAsyncTask.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "timeAsyncTask.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeAsyncTask.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "timeAsyncTask.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeAsyncTask.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "timeSyncTask.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeSyncTask.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "timeSyncTask.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeSyncTask.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "timeSyncTask.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeSyncTask.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "sizeInputBufferSyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferSyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeInputBufferSyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferSyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeInputBufferSyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferSyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "sizeOutputBufferSyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferSyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeOutputBufferSyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferSyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeOutputBufferSyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferSyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "sizeInputBufferAsyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferAsyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeInputBufferAsyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferAsyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeInputBufferAsyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferAsyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "sizeOutputBufferAsyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferAsyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeOutputBufferAsyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferAsyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeOutputBufferAsyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferAsyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "cpuUsageAsyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.cpuUsageAsyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "cpuUsageAsyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.cpuUsageAsyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "cpuUsageAsyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.cpuUsageAsyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "memoryUsageAsyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.memoryUsageAsyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "memoryUsageAsyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.memoryUsageAsyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "memoryUsageAsyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.memoryUsageAsyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "countSyncTasksForMinute.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasksForMinute.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countSyncTasksForMinute.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasksForMinute.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countSyncTasksForMinute.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasksForMinute.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "countAsyncTasksForMinute.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasksForMinute.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countAsyncTasksForMinute.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasksForMinute.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countAsyncTasksForMinute.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasksForMinute.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "countSyncTasks" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasks << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countAsyncTasks" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasks << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countSyncTasksFail" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasksFail << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countAsyncTasksFail" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasksFail;

      return ret.str();
    }

    std::string DataProcessorData::processAdminCommandsSphinx(std::string & command){
      std::string ret = NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN_JSON_MSG;

      logger.log(NODE_LOG_MODE_DEBUG) << NODE_MSG_SPHINX << " admin command requested [" << command << "]" << flush;

      //!<Create input object for Sphinx FO
      HCE::sphinx::SphinxInputJsonMessage inputMsg(command);
      if(!inputMsg.isError()){
        if(_sphinxFO->getErrorCode()){
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " before process indexer state error "
              << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg() << flush;
          //!<Reset searcher error state to get next message to start from scratch
          _sphinxFO->resetError();
        }else{
          //!<Execute admin request
          HCE::sphinx::SphinxOutputJsonMessage outputMessage = _sphinxFO->Process(inputMsg);
          //Sphinx FO error
          if(_sphinxFO->getErrorCode()){
            //!<Log some error of Sphinx FO
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " after process indexer state error " << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg()
                << endl << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;
            //!<Reset indexer error state to get next message to start from scratch
            _sphinxFO->resetError();
          }
          //Output message error
          if(outputMessage.getErrorCode()){
            //!<Log some error of indexer request execution
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " output message error " << outputMessage.getErrorCode() << " : " << outputMessage.getErrorMessage() << flush;
          }
          //Get json message from output object
          if(!outputMessage.serialize(ret)){
            //!<Log some error of indexer request execution
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " output message serialization error " << outputMessage.getErrorCode() << " : "
                << outputMessage.getErrorMessage() << flush;

            //!<Return default response with general error
            ret = NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN_JSON_MSG;
          }
          //!<Register Sphinx request time
          setStatSphinxTimeTotalAdmin(getStatSphinxTimeTotalAdmin() + outputMessage.getTime());
        }
      }else{
        logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " deserialization error " << inputMsg.getErrorCode() << " : " << inputMsg.getErrorMsg() << flush;
      }

      //!<Register Sphinx admin requests number stat
      unsigned long long statSphinxRequestsAdmin = getStatSphinxRequestsAdmin();
      setStatSphinxRequestsAdmin(++statSphinxRequestsAdmin);

      return ret;
    }

    void* DataProcessorData::process(void){
      if(_inprocDataSock && _inprocAdminSock){
        logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

        //!<Enter main loop of processing of message
        while(getInProgress()){
          try{
            //!<Init polling pool
            zmq::pollitem_t items[] = {{*_inprocDataSock, 0, ZMQ_POLLIN, 0}, {*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};

            //!<Poll inproc_data server
            zmq::poll(items, 2, getPollTimeout());

            //!<Handle income data message
            if(items[0].revents & ZMQ_POLLIN){
              handleInternalMessage();
            }

            //!<Handle inproc_admin's socket message
            if(items[1].revents & ZMQ_POLLIN){
              handleAdminMessage();
            }

            //!<Send periodical property values
            sendPeriodicProperties();

            //!<Dump run-time variables at regular intervals
            if (dumpRunTimeVariables()){
              saveNodeProperties();
              saveSphinxProperties();
              saveDRCEProperties();
              dumpProperties();
            }

          }catch(zmq::error_t& error){
            if (EINTR!=error.num()){
              std::cout << "DataProcessorData E: " << error.what() << std::endl;
            }
          }catch(std::exception& e){
            std::cout << "DataProcessorData E: " << e.what() << std::endl;
          }catch(...){
            std::cout << "DataProcessorData E: Unknown exception" << std::endl;
          }
        }

        logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << "DataProcessorData::process() aborted because not all required sockets created : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
        ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
      }

      return NULL;
    }

  }
}
