/**
 * @file DataProcessorData.hpp
 *
 * HCE project dataprocessor_data handler.\n
 * Implements functionality for data processor handler.\n
 * Implements data processing functionality of data and admin requests handling for data node.\n
 * Container for data processor functional objects FO of data storage(s) or end-user data management functionality blackboxes.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_DATA_PROCESSOR_DATA_INCLUDED__
#define __HCE_DATA_PROCESSOR_DATA_INCLUDED__

#include <iostream>
#include <sstream>
#include <iterator>

#include "Handler.hpp"

//Node ini
#include <Poco/Util/IniFileConfiguration.h>
#include "Poco/String.h"

//Sphinx FO
#include "SphinxInputJsonMessage.hpp"
#include "SphinxOutputJsonMessage.hpp"
#include "SphinxFunctionalObject.hpp"
//DRCE FO
#include "DRCEDefaultJSON.hpp"
#include "DRCEInputJsonMessage.hpp"
#include "DRCEOutputJsonMessage.hpp"
#include "DRCEFunctionalObject.hpp"
//Reducer input object
#include "PocoJSONReducingInputMessageConvertor.hpp"
#include <Poco/SharedPtr.h>
#include "Exceptions.hpp"
//Message cover object
#include "JsonMessageCover.hpp"

//User home directory path
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>

//Task state notification
#include <Poco/UUID.h>

namespace HCE {
  namespace handlers {

    class DataProcessorData: public HCE::handlers::Handler {
      public:
        DataProcessorData(const std::string& nodeName, std::string name, zmq::context_t& ctx, int64_t startedAt, unsigned char logPriority, const std::string& iniFile);
        ~DataProcessorData();

      protected:
        int initialize(void);
        void deinitialize(void);
        void* process(void);

        void handleInternalMessage(void);
        std::string processDataRequest(const std::string& messageBody);
        std::string processAdminCommands(std::string& command, std::string& parameters);
        std::string processAdminCommandsSphinx(std::string& command);

        //!<Node name setter
        void setNodeName(const std::string& nodeName);
        //!<Node name getter
        std::string getNodeName(void);

        //!<Plain text requests data processor. Used for test purposes as a minimal simple data generator and processor as well as messages handling algorithms
        std::string processDataRequestPlainText(const std::string& messageBody);

        //!<Json requests data processor. Used to check the request target type and chose proper metod to handle: Sphinx, sql, etc...
        std::string processDataRequestJson(const std::string& messageBody);

        //!<%Sphinx requests data processor by Sphinx FO
        std::string processDataRequestSphinx(const std::string& messageBody);

        //!<%Sphinx FO initialization from ini file
        void dataStorageInitSphinx(const std::string& iniFile);

        //!<%Node initialization from ini file
        void nodeInit(const std::string& iniFile);

        //!<Check initialization and set default value if necessary
        void checkInit(void);

        //!<Prepare dump node properties
        void saveNodeProperties(void);

        //!<Prepare dump properties for Sphinx FO
        void saveSphinxProperties(void);

        //!<Prepare dump properties for DRCE FO
        void saveDRCEProperties(void);

        //!<Restore node properties
        void restoreNodeProperties(void);

        //!<Restore properties for Sphinx FO
        void restoreSphinxProperties(void);

        //!<Restore properties for DRCE FO
        void restoreDRCEProperties(void);

        //!<%Get current user home dir
        static std::string getUserHomeDir(void);

        //!<%Get data storage properties for Sphinx FO
        std::string getDataStorageSphinxProperties(void);

        //!<%Get data storage properties for DRCE FO
        std::string getDataStorageDRCEProperties(void);

        //!<%DRCE requests data processor by DRCE FO
        std::string processDataRequestDRCE(const std::string& messageBody);

        //!<%DRCE FO initialization from ini file
        void dataStorageInitDRCE(const std::string& iniFile);

        //!<Rebuilds tasks state notification connection
        bool rebuildTasksStateNotificationConnection(void);

        //!<Recover notification connection
        bool recoverNotificationConnection(void);

        //!<Get property information
        std::string getPropertyInformation(void);

        //!<%ZMQ Sockets
        zmq::socket_t* _inprocDataSock;
        zmq::socket_t* _tasksStateNotificationSock;

        //!<Setters and getters for stat variables
        void setStatClientResponses(unsigned long long statClientResponses);
        unsigned long long getStatClientResponses(void);

        void setStatSphinxRequestsSearch(unsigned long long statSphinxRequestsSearch);
        unsigned long long getStatSphinxRequestsSearch(void);

        void setStatSphinxRequestsAdmin(unsigned long long statSphinxRequestsAdmin);
        unsigned long long getStatSphinxRequestsAdmin(void);

        void setStatSphinxTimeTotalSearch(unsigned long long statSphinxTimeTotalSearch);
        unsigned long long getStatSphinxTimeTotalSearch(void);

        void setStatSphinxTimeTotalAdmin(unsigned long long statSphinxTimeTotalAdmin);
        unsigned long long getStatSphinxTimeTotalAdmin(void);

        void setStatDRCERequests(unsigned long long statDRCERequests);
        unsigned long long getStatDRCERequests(void);

        void setStatDRCETimeTotal(unsigned long long statDRCETimeTotal);
        unsigned long long getStatDRCETimeTotal(void);

        //!<%Sphinx FO instance
        HCE::sphinx::SphinxFunctionalObject* _sphinxFO;

        //!<%DRCE FO instance
        HCE::drce::DRCEFunctionalObject* _drceFO;

        //!<%Node data processing mode, 0 - regular with involving of FO, 1 - fake w/o involving FO and random results generation
        void setDataProcessingMode(unsigned int dataProcessingMode);
        unsigned int getDataProcessingMode(void);

        void setDataProcessingFakeResults(unsigned int dataProcessingFakeResults);
        unsigned int getDataProcessingFakeResults(void);

        void setDataProcessingFakeResultsMax(unsigned int dataProcessingFakeResultsMax);
        unsigned int getDataProcessingFakeResultsMax(void);

        //!<Node number
        void setNodeNumber(unsigned int NodeNumber);
        unsigned int getNodeNumber(void);

        //!<Tasks state notification update service
        void setStateNotificationHost(const std::string& stateNotificationHost);
        std::string getStateNotificationHost(void);

        void setStateNotificationPort(const std::string& stateNotificationPort);
        std::string getStateNotificationPort(void);

        class TasksStateNotificationFunctor {
          public:
            TasksStateNotificationFunctor(zmq::socket_t* _ssock):_msgId(Poco::UUID::x500().toString()),_updateConnectionSock( _ssock){
              //std::cout << "\n\n***Constructor::TasksStateNotificationFunctor***\n\n";
              //_updateConnectionSock = _ssock;
              //_msgId = Poco::UUID::x500().toString();
            }

            ~TasksStateNotificationFunctor(){
              if(_updateConnectionSock){
                //std::cout << "\n\n***Destructor::TasksStateNotificationFunctor***\n\n";
              }
            }

            std::string operator()(const std::string& data){
//              std::cout << "!!!TasksStateNotificationFunctor!!!" << std::endl;
//              std::cout << data << std::endl;
//              std::cout << "!!!TasksStateNotificationFunctor!!!" << std::endl;

              if(_updateConnectionSock){
                //Create new message with Id
                zmsg msg(_msgId.c_str());
                //!<Insert response buffer
                msg.append(data.c_str());

                //!<Send notification message
                msg.send(*_updateConnectionSock);
              }

              return "";
            }

            //Message Id
            std::string _msgId;
            //!<%ZMQ Socket
            zmq::socket_t* _updateConnectionSock;
        };

        TasksStateNotificationFunctor* _tasksStateNotificationFunctor;
    };

  }
}

#endif /* __HCE_DATA_PROCESSOR_DATA_INCLUDED__ */
