#include <stdlib.h>
#include <time.h>
#include <random>

#include "DataServerProxy.hpp"
#include "ServiceMessages.hpp"
//DRCE FO
#include "DRCEFunctionalObject.hpp"
#include "DRCEMessageConst.hpp"
#include "DRCEError.hpp"
//Sphinx FO
#include "SphinxFunctionalObject.hpp"
#include "SphinxMessageConst.hpp"
#include "SphinxError.hpp"
//Message cover object
#include "JsonMessageCover.hpp"
//Resource extractor
#include "DRCEResourceExtractor.hpp"
//Resource usage checker
#include "ResourceUsageLimitsChecker.hpp"

namespace HCE {
  namespace handlers {

    DataServerProxy::DataServerProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity,
        std::string connectionString, unsigned char nodeMode, int64_t startedAt,
        unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, clientIdentity, connectionString, startedAt, logPriority, iniFile), DataNodeOptions(Handler::handlerProperties),
        _previousClientNumberRandom(0), _previousClientNumberResourceUsage(0), calculatorAlgorithmData(), resourceUsageLimitsData(),
        typeCalculatorAlgorithm(CalculatorAlgorithm::Types::caDefault), _resourceUsageManager(logger){

      setNodeMode(nodeMode);
      //!<Init all the rest
      initialize();
    }

    DataServerProxy::~DataServerProxy(){
      deinitialize();
    }

    int DataServerProxy::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << getClientIdentity() << "] with connection [" << getConnectionString() << "]" << flush;

      //!<Init sockets
      _inprocProxySock = NULL;
      _inprocReducerInSock = NULL;
      _dataServerSock = NULL;

      //!<Node init from file
      initNode(getIniFile(), logger);

      //!<Heartbeat at time mark
      _heartbeatAt = s_clock() - getHeartbeatDelay();

      //!<Send out periodical stats at regular intervals
      _pstatsAt = s_clock() + STAT_INTERVAL;

      //!<Node init
      nodeInit(getIniFile());

      //!<Check initialization and set default value if necessary
      checkInit();

      _clientsManager = new ClientsQueueManager(&_clientsQueue, 0, &_resourceUsageManager);
      //!<Apply heartbeat timeout value to client manager
      _clientsManager->setHeartbeatTimeout(getHeartbeatTimeout());
      //!<Apply max resources usage collected size value to client manager
      _clientsManager->setMaxResourcesUsageCollectedSize(getResourcesUsageCollectedSize());

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildServerConnection(_inprocProxySock, NODE_CS_INPROC_PROXY, ZMQ_DEALER);
      //!<Rebuild inproc_reducer_in connection
      rebuildClientConnection(_inprocReducerInSock, NODE_CS_INPROC_REDUCER_IN, ZMQ_DEALER);
      //!<Rebuild external server connection
      rebuildServerConnection(_dataServerSock, getConnectionString(), ZMQ_ROUTER);

      return true;
    }

    void DataServerProxy::deinitialize(void){
      if(_inprocProxySock){
        delete _inprocProxySock;
      }
      if(_inprocReducerInSock){
        delete _inprocReducerInSock;
      }
      if(_dataServerSock){
        delete _dataServerSock;
      }
      _inprocProxySock = NULL;
      _inprocReducerInSock = NULL;
      _dataServerSock = NULL;
      _clientsQueue.clear();
      if(_clientsManager){
        delete _clientsManager;
      }
      HCE::drce::ResourceUsageExtractor::shutdown();
    }

    std::string DataServerProxy::getNodeModeName(unsigned int nodeMode){
      std::string result;
      switch(nodeMode){
        case NODE_MODE_PROXY: result = "NODE_MODE_PROXY";   //!< Multicast requests balancing (supposes proportional sharding of documents as sharding model)
          break;
        case NODE_MODE_SHARD: result = "NODE_MODE_SHARD";   //!< Round robin requests balancing (supposes mirroring of data documents as sharding model)
          break;
        case NODE_MODE_DATA: result = "NODE_MODE_DATA";     //!< Node in data mode, i.e. - end of chain in cluster tree
          break;
        case NODE_MODE_ROUTER: result = "NODE_MODE_ROUTER"; //!< Multicast requests balancing
          break;
        case NODE_MODE_SHARD_RND: result = "NODE_MODE_SHARD_RND"; //!< Random single-cast balancing
          break;
        case NODE_MODE_SHARD_RU: result = "NODE_MODE_SHARD_RU"; //!< Resources usage balancing (supposes choose of data node based resources usage)
          break;
        default: result = "Unknown node mode !!!";
          break;
      }
      return result;
    }

    void DataServerProxy::setNodeMode(unsigned int nodeMode){
      handlerProperties.set(HCE::handlers::properties::NODE_MODE, nodeMode);
    }

    unsigned int DataServerProxy::getNodeMode(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::NODE_MODE);
    }

    void DataServerProxy::setResourcesUsageAlg(unsigned int resourcesUsageAlg){
      typeCalculatorAlgorithm = static_cast<CalculatorAlgorithm::Types>(resourcesUsageAlg);
      handlerProperties.set(HCE::handlers::properties::RESOURCES_USAGE_ALG, resourcesUsageAlg);
    }

    unsigned int DataServerProxy::getResourcesUsageAlg(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::RESOURCES_USAGE_ALG);
    }

    void DataServerProxy::setResourcesUsageAlg0Weights(const std::string& resourcesUsageAlg0Weights){
      if (!resourcesUsageAlg0Weights.empty()) {
        HCE::handlers::DictionaryConverter<std::string, double> converter(resourcesUsageAlg0Weights);
        calculatorAlgorithmData = converter.toObject().cast<HCE::handlers::CalculatorAlgorithmData>();
        handlerProperties.set(HCE::handlers::properties::RESOURCES_USAGE_ALG0_WEIGHTS, resourcesUsageAlg0Weights);
      }
    }

    std::string DataServerProxy::getResourcesUsageAlg0Weights(void){
      return handlerProperties.get<std::string>(HCE::handlers::properties::RESOURCES_USAGE_ALG0_WEIGHTS);
    }

    void DataServerProxy::setResourcesUsageOrder(unsigned int resourcesUsageOrder){
      calculatorAlgorithmData.setOrder(resourcesUsageOrder);
      handlerProperties.set(HCE::handlers::properties::RESOURCES_USAGE_USAGE_ORDER, resourcesUsageOrder);
    }

    unsigned int DataServerProxy::getResourcesUsageOrder(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::RESOURCES_USAGE_USAGE_ORDER);
    }

    void DataServerProxy::setResourcesUsageLimits(const std::string& resourcesUsageLimits){
      if (!resourcesUsageLimits.empty()) {
        HCE::handlers::DictionaryConverter<std::string, double> converter(resourcesUsageLimits);
        resourceUsageLimitsData = converter.toObject().cast<HCE::handlers::ResourceUsageLimitsData>();
        handlerProperties.set(HCE::handlers::properties::RESOURCES_USAGE_LIMITS, resourcesUsageLimits);
      }
    }

    std::string DataServerProxy::getResourcesUsageLimits(void){
      return handlerProperties.get<std::string>(HCE::handlers::properties::RESOURCES_USAGE_LIMITS);
    }

    void DataServerProxy::setResourcesUsageCollectedSize(unsigned int resourcesUsageCollectedSize){
      handlerProperties.set(HCE::handlers::properties::RESOURCES_USAGE_COLLECTED_SIZE, resourcesUsageCollectedSize);
    }

    unsigned int DataServerProxy::getResourcesUsageCollectedSize(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::RESOURCES_USAGE_COLLECTED_SIZE);
    }

    void DataServerProxy::setResourcesUsageMaxAllowedRange(double resourcesUsageMaxAllowedRange){
      handlerProperties.set(HCE::handlers::properties::RESOURCES_USAGE_MAX_ALLOWED_RANGE, resourcesUsageMaxAllowedRange);
    }

    double DataServerProxy::getResourcesUsageMaxAllowedRange(void){
      return handlerProperties.get<double>(HCE::handlers::properties::RESOURCES_USAGE_MAX_ALLOWED_RANGE);
    }

    void DataServerProxy::setResourcesExtractorMaxThreads(unsigned int resourcesExtractorMaxThreads){
      handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS, resourcesExtractorMaxThreads);
      HCE::drce::ResourceUsageExtractor::getInstance().setMaxThreadsCount(resourcesExtractorMaxThreads);
    }

    unsigned int DataServerProxy::getResourcesExtractorMaxThreads(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS);
    }

    void DataServerProxy::setResourcesExtractorMaxProcesses(unsigned int resourcesExtractorMaxProcesses){
      handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES, resourcesExtractorMaxProcesses);
      HCE::drce::ResourceUsageExtractor::getInstance().setMaxProcessesCount(resourcesExtractorMaxProcesses);
    }

    unsigned int DataServerProxy::getResourcesExtractorMaxProcesses(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES);
    }

    void DataServerProxy::setResourcesExtractorMaxDiskSize(unsigned int resourcesExtractorMaxDiskSize){
      handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE, resourcesExtractorMaxDiskSize);
      HCE::drce::ResourceUsageExtractor::getInstance().setMaxDiskSize(resourcesExtractorMaxDiskSize);
    }

    unsigned int DataServerProxy::getResourcesExtractorMaxDiskSize(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE);
    }

    void DataServerProxy::setResourcesExtractorMaxVramSize(unsigned int resourcesExtractorMaxVramSize){
      handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE, resourcesExtractorMaxVramSize);
      HCE::drce::ResourceUsageExtractor::getInstance().setMaxVramSize(resourcesExtractorMaxVramSize);
    }

    unsigned int DataServerProxy::getResourcesExtractorMaxVramSize(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE);
    }

    void DataServerProxy::setResourcesExtractorMaxRramSize(unsigned int resourcesExtractorMaxRramSize){
      handlerProperties.set(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE, resourcesExtractorMaxRramSize);
      HCE::drce::ResourceUsageExtractor::getInstance().setMaxRramSize(resourcesExtractorMaxRramSize);
    }

    unsigned int DataServerProxy::getResourcesExtractorMaxRramSize(void){
      return handlerProperties.get<unsigned int>(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE);
    }

    //!<Setters and getters for stat variables
    void DataServerProxy::setStatClientRequests(unsigned int statClientRequests){
      statProperties.set(HCE::handlers::properties::STAT_CLIENT_REQUESTS, statClientRequests);
    }

    unsigned int DataServerProxy::getStatClientRequests(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_CLIENT_REQUESTS);
    }

    void DataServerProxy::setStatClientResponses(unsigned int statClientResponses){
      statProperties.set(HCE::handlers::properties::STAT_CLIENT_RESPONSES, statClientResponses);
    }

    unsigned int DataServerProxy::getStatClientResponses(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_CLIENT_RESPONSES);
    }

    void DataServerProxy::setStatRequests(unsigned int statRequests){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS, statRequests);
    }

    unsigned int DataServerProxy::getStatRequests(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS);
    }

    void DataServerProxy::setStatClientRequestsSize(unsigned int statClientRequestsSize){
      statProperties.set(HCE::handlers::properties::STAT_CLIENT_REQUESTS_SIZE, statClientRequestsSize);
    }

    unsigned int DataServerProxy::getStatClientRequestsSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_CLIENT_REQUESTS_SIZE);
    }

    void DataServerProxy::setStatClientResponsesSize(unsigned int statClientResponsesSize){
      statProperties.set(HCE::handlers::properties::STAT_CLIENT_RESPONSES_SIZE, statClientResponsesSize);
    }

    unsigned int DataServerProxy::getStatClientResponsesSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_CLIENT_RESPONSES_SIZE);
    }

    void DataServerProxy::setStatRequestsSize(unsigned int statRequestsSize){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_SIZE, statRequestsSize);
    }

    unsigned int DataServerProxy::getStatRequestsSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_SIZE);
    }

    void DataServerProxy::setPeriodicStatClientRequests(unsigned int periodicStatClientRequests){
      statProperties.set(HCE::handlers::properties::PEREODIC_STAT_CLIENT_REQUESTS, periodicStatClientRequests);
    }

    unsigned int DataServerProxy::getPeriodicStatClientRequests(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::PEREODIC_STAT_CLIENT_REQUESTS);
    }

    void DataServerProxy::setPeriodicStatClientResponses(unsigned int periodicStatClientResponses){
      statProperties.set(HCE::handlers::properties::PEREODIC_STAT_CLIENT_RESPONSES, periodicStatClientResponses);
    }

    unsigned int DataServerProxy::getPeriodicStatClientResponses(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::PEREODIC_STAT_CLIENT_RESPONSES);
    }

    void DataServerProxy::setStatRequestsProcessed(unsigned int statRequestsProcessed){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_PROCESSED, statRequestsProcessed);
    }

    unsigned int DataServerProxy::getStatRequestsProcessed(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_PROCESSED);
    }

    void DataServerProxy::setStatRequestsProcessedSize(unsigned int statRequestsProcessedSize){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE, statRequestsProcessedSize);
    }

    unsigned int DataServerProxy::getStatRequestsProcessedSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE);
    }

    void DataServerProxy::setPeriodicStatClientRequestsSize(unsigned int periodicStatClientRequestsSize){
      statProperties.set(HCE::handlers::properties::PEREODIC_STAT_CLIENT_REQUESTS_SIZE, periodicStatClientRequestsSize);
    }

    unsigned int DataServerProxy::getPeriodicStatClientRequestsSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::PEREODIC_STAT_CLIENT_REQUESTS_SIZE);
    }

    void DataServerProxy::setPeriodicStatClientResponsesSize(unsigned int periodicStatClientResponsesSize){
      statProperties.set(HCE::handlers::properties::PEREODIC_STAT_CLIENT_RESPONSES_SIZE, periodicStatClientResponsesSize);
    }

    unsigned int DataServerProxy::getPeriodicStatClientResponsesSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::PEREODIC_STAT_CLIENT_RESPONSES_SIZE);
    }

    void DataServerProxy::setStatQPSClient(float statQPSClient){
      statProperties.set(HCE::handlers::properties::STAT_QPS_CLIENT, statQPSClient);
    }

    float DataServerProxy::getStatQPSClient(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_QPS_CLIENT);
    }

    void DataServerProxy::setStatQPSServer(float statQPSServer){
      statProperties.set(HCE::handlers::properties::STAT_QPS_SERVER, statQPSServer);
    }

    float DataServerProxy::getStatQPSServer(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_QPS_SERVER);
    }

    void DataServerProxy::setStatRPSClient(float statRPSClient){
      statProperties.set(HCE::handlers::properties::STAT_RPS_CLIENT, statRPSClient);
    }

    float DataServerProxy::getStatRPSClient(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_RPS_CLIENT);
    }

    void DataServerProxy::setStatRPSServer(float statRPSServer){
      statProperties.set(HCE::handlers::properties::STAT_RPS_SERVER, statRPSServer);
    }

    float DataServerProxy::getStatRPSServer(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_RPS_SERVER);
    }

    void DataServerProxy::setStatRequestsBandwidthClient(float statRequestsBandwidthClient){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH_CLIENT, statRequestsBandwidthClient);
    }

    float DataServerProxy::getStatRequestsBandwidthClient(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH_CLIENT);
    }

    void DataServerProxy::setStatRequestsBandwidthServer(float statRequestsBandwidthServer){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH_SERVER, statRequestsBandwidthServer);
    }

    float DataServerProxy::getStatRequestsBandwidthServer(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH_SERVER);
    }

    void DataServerProxy::setStatResponsesBandwidthClient(float statResponsesBandwidthClient){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH_CLIENT, statResponsesBandwidthClient);
    }

    float DataServerProxy::getStatResponsesBandwidthClient(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH_CLIENT);
    }

    void DataServerProxy::setStatResponsesBandwidthServer(float statResponsesBandwidthServer){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH_SERVER, statResponsesBandwidthServer);
    }

    float DataServerProxy::getStatResponsesBandwidthServer(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH_SERVER);
    }

    void DataServerProxy::handleInternalMessage(void){
      //!<Fetch message from inprocontrol
      zmsg msg(*_inprocProxySock);
      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_proxy" << endl << msg.dump() << flush;

      if (msg.parts() == NODE_MSG_CMD_MIN_FIELDS){

        char* addrp = msg.address();
        std::string addr ((addrp)?addrp:"");
        if (addr.compare(0, NODE_MSG_ROUTES.length(), NODE_MSG_ROUTES)==0){
          std::vector<ClientWorkerItem> clientQueue;
          RouteMessage routeMessage(clientQueue);
          routeMessage.extract(NODE_MSG_ROUTES, addr);

          _clientsManager->append(clientQueue);

          RouteMessage routeMsg(_clientsQueue);

          s_send(*_inprocReducerInSock, routeMsg.build(NODE_MSG_ROUTES));
        }
      }else{
        //!<Create identity, messageId and body of dataclient's message
        std::string clientIdentity((const char*)msg.pop_front().c_str());
        std::string messageId((const char*)msg.pop_front().c_str());
        std::string messageBody((const char*)msg.pop_front().c_str());
        std::string messageRoute((msg.parts())?(const char*)msg.pop_front().c_str():"");

        unsigned int statRequests = getStatRequests();
        logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from upper server [" << clientIdentity << "], message [" << messageId << "] body " << messageBody.size()
            << " bytes, n=" << ++statRequests << flush;
        setStatRequests(statRequests);

        if(_clientsQueue.empty()){
          //!<No clients connected, make response message and send response for reducer
          sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNoClients, messageId, messageRoute);
          logger.log(NODE_LOG_MODE_DEBUG) << "No client`s connection registered!" << flush;
          logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
        }else{
          //!<Get allowed clients names
          std::vector<std::string> routeNames = std::forward<std::vector<std::string> >(RouteMessage::extractRouteNames(_clientsQueue));

          unsigned int nodeMode(getNodeMode());
          CalculatorAlgorithm::Types resourcesUsageAlgorithm = CalculatorAlgorithm::Types::caDefault;
          CalculatorAlgorithmData algorithmData(calculatorAlgorithmData);
          ResourceUsageLimitsData limitsData(resourceUsageLimitsData);

          logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  messageRoute: '" << messageRoute << "'" << flush;

          if (!messageRoute.empty() && nodeMode != NODE_MODE_ROUTER){
            logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  >>> (!messageRoute.empty() && nodeMode != NODE_MODE_ROUTER) <<<" << flush;

            //!<Extract route JSON from routeMessage
            std::string json(messageRoute);
            size_t pos = json.find_last_of('}');

            logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  pos = " << pos << flush;

            if (pos!=std::string::npos){
              json = json.substr(0, pos+1);

              logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  json(1): '" << json << "'" << flush;
            }else{
              json.clear();
            }
            //!<If JSON was successful extracted, update route names
            if (!json.empty()){
              logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  json(2): '" << json << "'" << flush;

              //!<Get nodes list from route message
              RequestRouteMessage requestRouteMessage(json); //deserialize route json
              if (!requestRouteMessage.isError()){
                unsigned int managerRole = requestRouteMessage.getManagerRole();

                logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  received value of managerRole: "
                    << managerRole << " it's '" << getNodeModeName(managerRole) << "'" << flush;

                std::vector<std::string> identityNames = std::forward<std::vector<std::string> >(requestRouteMessage.getNodesNames());

                logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  =============" << flush;
                for (const std::string& identityName : identityNames){
                  logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  identityName: " << identityName << flush;
                }

                //!<Set current resource usage algorithm
                resourcesUsageAlgorithm = (static_cast<CalculatorAlgorithm::Types>(requestRouteMessage.getResourcesUsageAlgorithm()) != typeCalculatorAlgorithm)?
                    static_cast<CalculatorAlgorithm::Types>(requestRouteMessage.getResourcesUsageAlgorithm()):typeCalculatorAlgorithm;
                //!<Set current resource usage algorithm weights
                if (!requestRouteMessage.getResourcesUsageAlgorithmWeights().empty()){
                  HCE::handlers::DictionaryConverter<std::string, double> converter(requestRouteMessage.getResourcesUsageAlgorithmWeights());
                  algorithmData = converter.toObject().cast<HCE::handlers::CalculatorAlgorithmData>();
                }
                //!<Set current resource usage limits
                if (!requestRouteMessage.getResourcesUsageLimits().empty()){
                  HCE::handlers::DictionaryConverter<std::string, double> converter(requestRouteMessage.getResourcesUsageLimits());
                  limitsData = converter.toObject().cast<HCE::handlers::ResourceUsageLimitsData>();
                }
                //!<Check allowed data node mode
                if (managerRole == NODE_MODE_PROXY || managerRole == NODE_MODE_SHARD || managerRole == NODE_MODE_SHARD_RND || managerRole == NODE_MODE_SHARD_RU)
                  nodeMode = managerRole;

                logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  NODE MODE: " << getNodeModeName(nodeMode) << flush;

                logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  routeNames.size() = " << routeNames.size() << flush;
                logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  identityNames.size() = " << identityNames.size() << flush;

                if (!identityNames.empty() && nodeMode != NODE_MODE_SHARD){
                  std::vector<std::string> routes;
                  for (const std::string& routeName:routeNames){
                    for (const std::string& identityName:identityNames){
                      logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  routeName: " << routeName
                          << " identityName: " << identityName << " isEqual: " << std::boolalpha << (routeName==identityName) << flush;
                      if (routeName==identityName){
                        routes.push_back(identityName);
                      }
                    }
                  }
                  routeNames = routes;
                }else if(nodeMode == NODE_MODE_SHARD){
                  std::vector<std::string> routes;
                  for (const ClientWorkerItem& clientWorker:_clientsQueue){
                    logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  clientWorker.identity: " << clientWorker.identity << flush;
                    routes.push_back(clientWorker.identity);
                  }
                  routeNames = routes;
                }
              }else{
                logger.log(NODE_LOG_MODE_ERROR) << "!!! DataServerProxy::handleInternalMessage  ErrorMsg: " << requestRouteMessage.getErrorMsg() << flush;
                logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  messageRoute: " << messageRoute << flush;
              }
            }
          }

          logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  getNodeMode(): " << getNodeModeName(getNodeMode()) << flush;
          logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  using nodeMode: " << getNodeModeName(nodeMode) << flush;
          logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  routeNames.size() = " << routeNames.size() << flush;

          for (size_t i=0;i<routeNames.size();++i){
            logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  routeNames[" << i << "] = " << routeNames[i] << flush;
          }

          if(routeNames.empty() && nodeMode != NODE_MODE_SHARD){
          //!<No clients connected, make response message and send response for reducer
          sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNoClients, messageId, messageRoute);
          logger.log(NODE_LOG_MODE_DEBUG) << "No valid route names!" << flush;
          logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;

          //!<Some client(s) registered
          }else if(nodeMode == NODE_MODE_PROXY || nodeMode == NODE_MODE_ROUTER){

            std::stringstream routes;
            routes << messageRoute << ROUTE_LIST_COUNT_DELIMITER << _clientsQueue.size();
            //!<Re-send data request for all clients connected to dataserver
            for (size_t i=0;i<_clientsQueue.size();++i){
              //!<Create and send response message
              sendResponseMessage(_clientsQueue[i].identity, messageId, messageBody, routes.str());

              unsigned int statClientRequests = getStatClientRequests();
              setStatClientRequests(++statClientRequests);
              setStatClientRequestsSize(getStatClientRequestsSize() + messageBody.size());

              logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT_ALL to dataserver socket for client [" << _clientsQueue[i].identity
                  << "], message [" << messageId << "] body " << messageBody.size() << ", n=" << getStatClientRequests() << flush;
            }
          }else{

            if(nodeMode == NODE_MODE_SHARD || nodeMode == NODE_MODE_SHARD_RND || nodeMode == NODE_MODE_SHARD_RU) {
              std::string identity;
              //!<Make local copy
              std::vector<ClientWorkerItem> clientsQueue;
              for (size_t i=0;i<routeNames.size();++i){
                for (size_t k=0;k<_clientsQueue.size();++k){
                  if (routeNames[i] == _clientsQueue[k].identity){
                    clientsQueue.push_back(ClientWorkerItem(_clientsQueue[k]));
                    break;
                  }
                }
              }

              if (clientsQueue.empty()){
                //!<No clients connected, make response message and send response for reducer
                sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNoClients, messageId, messageRoute);
                logger.log(NODE_LOG_MODE_DEBUG) << "No valid route names!" << flush;
                logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
              }else{

                if (nodeMode == NODE_MODE_SHARD_RND){
                  //!<Get identity use random algorithm
                  identity = getIdentityRandom(_clientsQueue, _previousClientNumberRandom);
                }else if(nodeMode == NODE_MODE_SHARD_RU){
                  //!<Get identity use resource usage algorithm
                  identity = getIdentityResourceUsage(clientsQueue, _previousClientNumberResourceUsage, resourcesUsageAlgorithm, algorithmData);
                  if (identity.empty()){
                    identity = getIdentityRandom(_clientsQueue, _previousClientNumberResourceUsage);
                  }
                  _clientsManager->incrementFrequency(identity);
                  logger.log(NODE_LOG_MODE_DEBUG) << "Choosen identity: '" << identity << "'" << flush;
                }else{
                  logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  using nodeMode: "
                      << nodeMode << " it's '" << getNodeModeName(nodeMode) << "'" << flush;

                  //!<Sort clients by frequency of queries done
                  std::sort(clientsQueue.begin(), clientsQueue.end(), ClientWorkerItem::sortByFrequencyPredicate);
                  //!<Inc freq of best (with minimal freq)
                  _clientsManager->incrementFrequency(clientsQueue[0].identity);
                  //!<Save identity for next use
                  if (!clientsQueue.empty())
                    identity = clientsQueue[0].identity;
                }


                if (identity.empty()){
                  //!<No clients connected, make response message and send response for reducer
                  sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNoClients, messageId, messageRoute);
                  logger.log(NODE_LOG_MODE_DEBUG) << "Empty client name cannot be use for routing balance!" << flush;
                  logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
                }else{

                  logger.log(NODE_LOG_MODE_DEBUG) << "resourcesUsageAlgorithm = " << (unsigned int)resourcesUsageAlgorithm << flush;
                  logger.log(NODE_LOG_MODE_DEBUG) << "algorithmData = " << HCE::handlers::DictionaryConverter<std::string, double>(algorithmData).toString() << flush;
                  logger.log(NODE_LOG_MODE_DEBUG) << "limitsData = " << HCE::handlers::DictionaryConverter<std::string, double>(limitsData).toString() << flush;

                  ResourceUsageCollection resourceUsageCollection = _clientsManager->getResources(identity);

                  HCE::handlers::DictionaryConverter<std::string, double> converter(_resourceUsageManager.calculateAverage(identity, resourceUsageCollection));
                  HCE::handlers::CalculatorNodeResourceData resourceData = converter.toObject().cast<HCE::handlers::CalculatorNodeResourceData>(identity);

                  std::string errorMsg;
                  HCE::handlers::ResourceUsageLimitsChecker resourceUsageLimitsChecker(limitsData);
                  if (!resourceUsageLimitsChecker.safeCheck(resourceData, errorMsg)) {
                    //!<Node has not allowed resource usage limits
                    sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNotEnoughResourceLimits, messageId, messageRoute, errorMsg);
                    logger.log(NODE_LOG_MODE_DEBUG) << "Node '" << identity << "' has not allowed resources usage limits, message skipped" << endl << msg.dump() << flush;
                  }else{
                    std::stringstream routes;
                    routes << messageRoute << ROUTE_LIST_COUNT_DELIMITER << 1;
                    //!<Create and send response message
                    sendResponseMessage(identity, messageId, messageBody, routes.str());

                    unsigned int statClientRequests = getStatClientRequests();
                    setStatClientRequests(++statClientRequests);
                    setStatClientRequestsSize(getStatClientRequestsSize() + messageBody.size());

                    logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT_RR to inproc_proxy socket from [" << identity << "], message [" << messageId << "] body "
                        << messageBody.size() << ", n=" << getStatClientRequests() << flush;
                  }
                }
              }
            }else{
              //!<Node mode not supported, make response message and send response for reducer
              sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertBadNodeMode, messageId, messageRoute);
              logger.log(NODE_LOG_MODE_DEBUG) << "Node mode value " << getNodeMode() << " not supported, message skipped" << endl << msg.dump() << flush;
            }
          }
        }
      }
    }

    size_t DataServerProxy::getRandomNodeNumber(size_t maxNumber){
      //!<Initialize random seed
      std::random_device rd;
      std::mt19937_64 gen(rd());
      std::uniform_real_distribution<double> dis(0, maxNumber);
      //!<Generate index number limited clients queue size
      return static_cast<size_t>(dis(gen));
    }

    std::string DataServerProxy::getIdentityRandom(std::vector<ClientWorkerItem>& clients, size_t& previousClientNumber){
      logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::getIdentityRandom clients count = "
          << clients.size() << " previousClientNumber = " << previousClientNumber << flush;

      std::string identity;
      size_t queueIndex = 0;
      do{
        queueIndex = getRandomNodeNumber(clients.size());

        logger.log(NODE_LOG_MODE_TRACE) << "getRandomNodeNumber() return queueIndex = " << queueIndex << flush;

        if (clients.empty() || clients.size()==1){
          logger.log(NODE_LOG_MODE_TRACE) << "Do-while broken because: clients count==1" << flush;
          break;
        }

      }while(queueIndex==previousClientNumber);

      if (!clients.empty()){

        std::vector<ClientWorkerItem> clientsQueue(clients);
        if (clientsQueue.size() > 1){
          //!<Erase previous client
          clientsQueue.erase(clientsQueue.begin()+previousClientNumber);
        }

        if (!clientsQueue.empty()){
          //!<Sort clients by frequency of queries done
          std::sort(clientsQueue.begin(), clientsQueue.end(), ClientWorkerItem::sortByFrequencyPredicate);
          //!<Save more greater node's identity name
          identity = clientsQueue[0].identity;

          for (queueIndex=0;queueIndex<clients.size();++queueIndex){
            if (clients[queueIndex].identity == identity){
              //!<Save previous client number
              previousClientNumber = queueIndex;
              break;
            }
          }
        }
      }
      //!<Return identity name
      return identity;
    }

    std::string DataServerProxy::getIdentityResourceUsage(std::vector<ClientWorkerItem>& clients,
        size_t& previousClientNumber,
        CalculatorAlgorithm::Types typeAlgorithm,
        const CalculatorAlgorithmData& algorithmData){

      logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::getIdentityResourceUsage  NODE_MODE_SHARD_RU  client count = " << clients.size() << flush;

      HCE::handlers::NodeResourceDataList nodeResourceDataList;
      for (size_t i=0;i<clients.size();++i){
        if (!clients[i].resources.empty()){
          //!<Calculate average value of resources usage
          HCE::handlers::DictionaryConverter<std::string, double> converter(_resourceUsageManager.calculateAverage(clients[i].identity, clients[i].resources));
          HCE::handlers::CalculatorNodeResourceData resourceData = converter.toObject().cast<HCE::handlers::CalculatorNodeResourceData, std::string>(clients[i].identity);
          nodeResourceDataList.addItem(std::forward<HCE::handlers::CalculatorNodeResourceData>(resourceData));

          logger.log(NODE_LOG_MODE_DEBUG) << i << ") [ " << clients[i].identity << " ] avg: '" << converter.toString() << "'" << flush;
        }
      }

      HCE::handlers::CalculatorNodeExecution calculatorNodeExecution(typeAlgorithm);
      std::string identity = calculatorNodeExecution.calculate(algorithmData, nodeResourceDataList, _resourceUsageManager);
      if (!identity.empty()){
        ResourceUsageCollection resourceUsageCollection = _clientsManager->getResources(identity);
        std::vector<double> listWeights = calculatorNodeExecution.getListWeights(algorithmData, resourceUsageCollection);
        _resourceUsageManager.setListWeights(identity, listWeights);

        if (_resourceUsageManager.isAllowedRange(identity, getResourcesUsageMaxAllowedRange())){
          identity.clear();
        }
      }

      for (size_t i=0;i<clients.size();++i){
        if (clients[i].identity==identity){
          previousClientNumber = i;
          break;
        }
      }
      return identity;
    }

    void DataServerProxy::sendEmptyResponseMessage(zmsg& msg,
        const std::string& coverMessageBody,
        EmptyResponseType responseType,
        const std::string& messageId,
        const std::string& messageRoute,
        const std::string& errorMsg){
      //!<Make response message and send response for reducer
      std::string messageBody = makeEmptyResponseMessage(coverMessageBody, responseType, messageRoute, errorMsg);
      //!<Insert message Id back
      msg.append(messageId.c_str());
      //!<Insert body back
      msg.append(messageBody.c_str());
      //!<Insert route back
      msg.append(messageRoute.c_str());
      //!<Create current clients number
      std::stringstream clientsNumberStream;
      clientsNumberStream << _clientsQueue.size();
      //!<Insert clients number field
      msg.append(clientsNumberStream.str().c_str());
      //!<Send message
      msg.send(*_inprocReducerInSock);
    }

    void DataServerProxy::sendResponseMessage(const std::string& identity, const std::string& messageId, const std::string& messageBody, const std::string& messageRoute){
      //!<Create response message and fill with fields
      zmsg clientMsg(getClientIdentity().c_str());
      clientMsg.wrap(identity.c_str(), NULL);
      clientMsg.append(messageId.c_str());
      clientMsg.append(messageBody.c_str());
      clientMsg.append(messageRoute.c_str());
      //!<Send response
      clientMsg.send(*_dataServerSock);
    }

    void DataServerProxy::nodeInit(const std::string& iniFile){

      if (!iniFile.empty()){
        logger.log(NODE_LOG_MODE_INFORMATION) << "load node settings from '" << iniFile << "' file" << flush;

        //!<Process Node's ini file
        try{
          Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
          try{
            //!<Set type of calcuator algorithm
            if (!handlerProperties.isExist(HCE::handlers::properties::RESOURCES_USAGE_ALG)){
              setResourcesUsageAlg(static_cast<unsigned int>(pConf->getInt(HCE::handlers::properties::RESOURCES_USAGE_ALG)));
            }
            //!<Set weights for default algorithm
            if (!handlerProperties.isExist(HCE::handlers::properties::RESOURCES_USAGE_ALG0_WEIGHTS)){
              setResourcesUsageAlg0Weights(pConf->getString(HCE::handlers::properties::RESOURCES_USAGE_ALG0_WEIGHTS));
            }

            //!<Set order for calculator algorithm data
            if (!handlerProperties.isExist(HCE::handlers::properties::RESOURCES_USAGE_USAGE_ORDER)){
              setResourcesUsageOrder(pConf->getInt(HCE::handlers::properties::RESOURCES_USAGE_USAGE_ORDER));
            }
            //!<Set resources usage limits
            if (!handlerProperties.isExist(HCE::handlers::properties::RESOURCES_USAGE_LIMITS)){
              setResourcesUsageLimits(pConf->getString(HCE::handlers::properties::RESOURCES_USAGE_LIMITS));
            }

            //!<Set resources usage collected size
            if (!handlerProperties.isExist(HCE::handlers::properties::RESOURCES_USAGE_COLLECTED_SIZE)){
              setResourcesUsageCollectedSize(pConf->getInt(HCE::handlers::properties::RESOURCES_USAGE_COLLECTED_SIZE));
            }

            //!<Set resource usage max allowed range (in percent)
            if (!handlerProperties.isExist(HCE::handlers::properties::RESOURCES_USAGE_MAX_ALLOWED_RANGE)){
              setResourcesUsageMaxAllowedRange(pConf->getInt(HCE::handlers::properties::RESOURCES_USAGE_MAX_ALLOWED_RANGE));
            }

            //!<Extract threads and processes counts from configuration file
            if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS)){
              setResourcesExtractorMaxThreads(pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS));
            }

            if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES)){
              setResourcesExtractorMaxProcesses(pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES));
            }

            if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE)){
              setResourcesExtractorMaxDiskSize(pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE));
            }

            if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE)){
              setResourcesExtractorMaxVramSize(pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE));
            }

            if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE)){
              setResourcesExtractorMaxRramSize(pConf->getInt(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE));
            }

          } catch(Poco::SyntaxException& e){
            //!<Log some error of config file processing
            logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node ini file processing error: " << e.displayText() << flush;
          }
        } catch(Poco::Exception& e){
          //!<Log some error of config file processing
          logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node ini file access error: " << e.displayText() << flush;
        }

        //!<Filling not initialized values use default values
        if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS)){
          setResourcesExtractorMaxThreads(HCE::handlers::DRCE_RESOURCES_EXTRACTOR_MAX_THREADS);
        }

        if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES)){
          setResourcesExtractorMaxThreads(HCE::handlers::DRCE_RESOURCES_EXTRACTOR_MAX_PROCESSES);
        }

        if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE)){
          setResourcesExtractorMaxThreads(HCE::handlers::DRCE_RESOURCES_EXTRACTOR_MAX_DISK_SIZE);
        }

        if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE)){
          setResourcesExtractorMaxThreads(HCE::handlers::DRCE_RESOURCES_EXTRACTOR_MAX_VRAM_SIZE);
        }

        if (!handlerProperties.isExist(HCE::handlers::properties::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE)){
          setResourcesExtractorMaxThreads(HCE::handlers::DRCE_RESOURCES_EXTRACTOR_MAX_RRAM_SIZE);
        }

        if (!handlerProperties.isExist(HCE::handlers::properties::RESOURCES_USAGE_COLLECTED_SIZE)){
          setResourcesUsageCollectedSize(HCE::handlers::RESOURCES_USAGE_COLLECTED_SIZE);
        }

        if (!handlerProperties.isExist(HCE::handlers::properties::RESOURCES_USAGE_MAX_ALLOWED_RANGE)){
          setResourcesUsageMaxAllowedRange(HCE::handlers::RESOURCES_USAGE_MAX_ALLOWED_RANGE);
        }

        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_order: " << calculatorAlgorithmData.getOrder() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_alg: " << static_cast<unsigned int>(typeCalculatorAlgorithm) << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_alg0_weights: " << getResourcesUsageAlg0Weights() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_limits: " << getResourcesUsageLimits() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_collected_size: " << getResourcesUsageCollectedSize() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_max_allowed_range: " << getResourcesUsageMaxAllowedRange() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_threads: " << getResourcesExtractorMaxThreads() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_processes: " << getResourcesExtractorMaxProcesses() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_disk_size: " << getResourcesExtractorMaxDiskSize() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_vram_size: " << getResourcesExtractorMaxVramSize() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_rram_size: " << getResourcesExtractorMaxRramSize() << flush;
      }
    }

    void DataServerProxy::checkInit(void){
      //!<Check all stat variables
      if (!statProperties.isExist(HCE::handlers::properties::STAT_CLIENT_REQUESTS))
        setStatClientRequests(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_CLIENT_RESPONSES))
        setStatClientResponses(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS))
        setStatRequests(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_CLIENT_REQUESTS_SIZE))
        setStatClientRequestsSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_CLIENT_RESPONSES_SIZE))
        setStatClientResponsesSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_SIZE))
        setStatRequestsSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::PEREODIC_STAT_CLIENT_REQUESTS))
        setPeriodicStatClientRequests(0);

      if (!statProperties.isExist(HCE::handlers::properties::PEREODIC_STAT_CLIENT_RESPONSES))
        setPeriodicStatClientResponses(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_PROCESSED))
        setStatRequestsProcessed(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE))
        setStatRequestsProcessedSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::PEREODIC_STAT_CLIENT_REQUESTS_SIZE))
        setPeriodicStatClientRequestsSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::PEREODIC_STAT_CLIENT_RESPONSES_SIZE))
        setPeriodicStatClientResponsesSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_QPS_CLIENT))
        setStatQPSClient(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_QPS_SERVER))
        setStatQPSServer(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RPS_CLIENT))
        setStatRPSClient(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RPS_SERVER))
        setStatRPSServer(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH_CLIENT))
        setStatRequestsBandwidthClient(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH_SERVER))
        setStatRequestsBandwidthServer(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH_CLIENT))
        setStatResponsesBandwidthClient(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH_SERVER))
        setStatResponsesBandwidthServer(0);
    }

    void DataServerProxy::registerClient(const std::string& identity){
      _clientsManager->remove(identity);
      _clientsManager->append(identity, false);

      logger.log(NODE_LOG_MODE_DEBUG) << "READY message from client [" << identity << "] received" << flush;
    }

    void DataServerProxy::processControlMessage(zmsg& msg){

      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());

      std::string addr((msg.address())?msg.address():"");
      std::string body((msg.body())?msg.body():"");

      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage  clientIdentity: " << clientIdentity << flush;
      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage            addr: " << addr << flush;
      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage            body: " << body << flush;

      if(addr == NODE_MSG_READY){
        registerClient(clientIdentity);
      }else if(addr.compare(0, NODE_MSG_HEARTBEAT.length(), NODE_MSG_HEARTBEAT)==0){

        std::vector<ClientWorkerItem> clientQueue;
        HCE::handlers::HeartbeatMessage heartbeatMessage(clientQueue);
        heartbeatMessage.extract(addr, NODE_MSG_HEARTBEAT);
        if (heartbeatMessage.isError())
          logger.log(NODE_LOG_MODE_ERROR) << "Operation extract heartbeat message has error: " << heartbeatMessage.getErrorMsg() << flush;
        else
        {
          _clientsManager->refresh(clientIdentity, clientQueue, heartbeatMessage.getResources());

          logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage  _clientsManager->refresh(" << clientIdentity << ")" << flush;

          RouteMessage routeMsg(_clientsQueue);
          std::string route = routeMsg.build(NODE_MSG_ROUTES);
          if (routeMsg.isError())
            logger.log(NODE_LOG_MODE_ERROR) << "Operation build route message has error: " << routeMsg.getErrorMsg() << flush;
          else
          {
            logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage  route: " << route << flush;
            s_send(*_inprocProxySock, route);
          }
        }
        logger.log(NODE_LOG_MODE_DEBUG) << "HEARTBEAT message from client [" << clientIdentity << "] received" << flush;

      }else if(addr == NODE_MSG_BYE){
        //!<Delete client from queue
        _clientsManager->remove(clientIdentity);

        logger.log(NODE_LOG_MODE_DEBUG) << "BYE message from client [" << clientIdentity << "] received" << flush;
      }else{
        //!<Invalid message type
        logger.log(NODE_LOG_MODE_DEBUG) << "INVALID command message from client [" << clientIdentity << "]" << flush;
        logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
      }
    }

    void DataServerProxy::processDataMessage(zmsg& msg){
      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());

      if(!_clientsManager->refresh(clientIdentity)){
        logger.log(NODE_LOG_MODE_DEBUG) << "MSG STRUCTURE ERROR wrong client identity [" << clientIdentity << "]"
            << endl << "Clients:[" << _clientsManager->getDump() << "]" << flush;
        logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
      }

      //!<Get message Id
      std::string messageId((const char*)msg.pop_front().c_str());
      //!<Get message body
      std::string messageBody((const char*)msg.pop_front().c_str());
      setStatClientResponsesSize(getStatClientResponsesSize() + messageBody.size());
      logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED RESPONSE from down client [" << clientIdentity << "], message [" << messageId << "] body " << messageBody.size()
          << " bytes n=" << getStatClientResponses() << flush;

      std::string messageRoute((msg.parts())?(const char*)msg.pop_front().c_str():"");

      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processDataMessage  messageRoute: " << messageRoute << flush;

      size_t pos = messageRoute.find_last_of(ROUTE_LIST_COUNT_DELIMITER);
      //!<Create current clients number
      std::stringstream clientsNumberStream;

      if (pos!=std::string::npos && (pos+1)!=std::string::npos) {
        unsigned int clientsNumber = 0;
        std::istringstream(messageRoute.substr(pos+1)) >> clientsNumber;
        if (clientsNumber && clientsNumber < _clientsQueue.size()) {
          clientsNumberStream << clientsNumber;
        }
        else
          clientsNumberStream << _clientsQueue.size();

        messageRoute = messageRoute.substr(0, pos);
      }else{
        clientsNumberStream << _clientsQueue.size();
      }

      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processDataMessage  messageRoute: '" << messageRoute << "', clientsNumberStream: '" << clientsNumberStream.str() << "'" << flush;

      //!<Insert message Id back
      msg.append(messageId.c_str());
      //!<Insert body back
      msg.append(messageBody.c_str());

      //!<Insert route back
      msg.append(messageRoute.c_str());

      //!<Insert clients number field
      msg.append(clientsNumberStream.str().c_str());

      //!<Send message
      msg.send(*_inprocReducerInSock);

      logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE SENT to inproc_reducer_in socket, message [" << messageId << "] body " << messageBody.size() << " bytes sent" << flush;
    }

    std::string DataServerProxy::makeEmptyResponseMessage(const std::string& messageBody,
        EmptyResponseType responseType,
        const std::string& messageRoute,
        const std::string& errorMsg){

      HCE::JsonMessageCover msgCover(messageBody);
      std::string resultMessageBody;
      std::stringstream errorMessage;

      switch(msgCover.getType()){
        case HCE::types::MessageType::mtSphinx: {

            unsigned int errorCode = HCE::sphinx::NO_ERROR;

            if (responseType==EmptyResponseType::ertNoClients){
              //!<No target node clients available
              errorCode = HCE::sphinx::ERROR_TARGET_CLIENTS_NOT_AVAILABLE;
              errorMessage << HCE::sphinx::message_const::targetClientsNotAvailable;
            }else if(responseType==EmptyResponseType::ertBadNodeMode){
              //!<Node mode not supported
              errorCode = HCE::sphinx::ERROR_NODE_MODE_NOT_SUPPORTED;
              errorMessage << HCE::sphinx::message_const::nodeNotSupported;
            }else if(responseType==EmptyResponseType::ertNotEnoughResourceLimits){
              //!<Chosen node has not allowed resource usage limits
              errorCode = HCE::sphinx::ERROR_NODE_HAS_NOT_ENOUGH_RESOURCE_LIMITS;
              errorMessage << HCE::sphinx::message_const::nodeHasNotEnoughResources;
            }

            if (!errorMsg.empty()){
              errorMessage << ". " << errorMsg;
            }
            resultMessageBody = HCE::sphinx::SphinxFunctionalObject::makeErrorMessage(msgCover.getData(),
                errorCode,
                errorMessage.str());
          }
          break;
        case HCE::types::MessageType::mtDrce: {

            unsigned int errorCode = HCE::drce::NO_ERROR;

            if (responseType==EmptyResponseType::ertNoClients){
              //!<No target node clients available
              errorCode = HCE::drce::ERROR_TARGET_CLIENTS_NOT_AVAILABLE;
              errorMessage  << HCE::drce::message_const::targetClientsNotAvailable;
            }
            else if(responseType==EmptyResponseType::ertBadNodeMode){
              //!<Node mode not supported
              errorCode = HCE::drce::ERROR_NODE_MODE_NOT_SUPPORTED;
              errorMessage << HCE::drce::message_const::nodeNotSupported;
            }else if(responseType==EmptyResponseType::ertNotEnoughResourceLimits){
              //!<Chosen node has not allowed resource usage limits
              errorCode = HCE::drce::ERROR_NODE_HAS_NOT_ENOUGH_RESOURCE_LIMITS;
              errorMessage << HCE::drce::message_const::nodeHasNotEnoughResources;
            }

            if (!errorMsg.empty()){
              errorMessage << ". " << errorMsg;
            }
            resultMessageBody = HCE::drce::DRCEFunctionalObject::makeErrorMessage(msgCover.getData(),
                HCE::drce::DRCETaskRequest::TaskState::NOT_SET_AS_NEW,
                errorCode,
                errorMessage.str());
          }
          break;
        default:;
      }

      std::string jsonMsg;
      msgCover.setData(resultMessageBody);
      if(!msgCover.serialize(jsonMsg)){
        logger.log(NODE_LOG_MODE_ERROR) << "Serialization error of cover object from empty results " << msgCover.getErrorCode() << " : " << msgCover.getErrorMsg() << flush;
      }

      return jsonMsg;
    }

    void DataServerProxy::handleExternalMessage(void){
      //!<Fetch message from dataserver
      zmsg msg(*_dataServerSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM dataserver" << endl << msg.dump() << flush;

      //!<Process dataserver's message
      if(msg.parts() == NODE_MSG_DATA_MIN_FIELDS){
        //!<Control message
        processControlMessage(msg);
      }else if(msg.parts() > NODE_MSG_DATA_MIN_FIELDS){
        //!<Inc data responses counter
        unsigned int statClientResponses = getStatClientResponses();
        setStatClientResponses(++statClientResponses);
        //!<Data message
        processDataMessage(msg);
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "ERROR of message structure, wrong fields number " << msg.parts() << endl << msg.dump() << flush;
      }
    }

    std::string DataServerProxy::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "clientRequests=" << getStatClientRequests() << ",clientResponses=" << getStatClientResponses() << ",clientRequestsSize="
            << getStatClientRequestsSize() << ",clientResponsesSize=" << getStatClientResponsesSize() << ",Requests=" << getStatRequests() << ",requestsSize=" << getStatRequestsSize();
        ret = result.str();
      }else if(command == NODE_MSG_rebuildServerConnection){
        //!<Rebuild server connection
        rebuildServerConnection(_dataServerSock, getConnectionString(), ZMQ_ROUTER);
      }else if(command == NODE_MSG_DISCONNECT_SERVER_CONNECTION){
        //!<Disconnect server connection
        _dataServerSock->close();
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else if(command == NODE_MSG_MANAGER_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(getNodeMode());
      }else if(command == NODE_MSG_MANAGER_MODE_SET){
        //!<Parse command parameters and fetch manager mode value
        unsigned int nodeMode = getNodeMode();
        //Parse manager mode from parameters
        std::istringstream(parameters) >> nodeMode;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(getNodeMode());
        //Set new manager mode
        setNodeMode(nodeMode);
      }else if(command == NODE_MSG_MANAGER_RESOURCES_COLLECTED_SIZE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_RESOURCES_COLLECTED_SIZE + "=" + std::to_string(_resourceUsageManager.getMaxSize());
      }else if(command == NODE_MSG_MANAGER_RESOURCES_COLLECTED_SIZE_SET){
        //!<Parse command parameters and fetch resources usage collected size
        unsigned int resourcesUsageCollectedSize = getResourcesUsageCollectedSize();
        //Parse resources usage collected size from parameters
        std::istringstream(parameters) >> resourcesUsageCollectedSize;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_RESOURCES_COLLECTED_SIZE + "=" + std::to_string(_resourceUsageManager.getMaxSize());
        //Set new resources usage collected size
        _resourceUsageManager.setMaxSize(resourcesUsageCollectedSize);
        _clientsManager->setMaxResourcesUsageCollectedSize(resourcesUsageCollectedSize);
        setResourcesUsageCollectedSize(resourcesUsageCollectedSize);
      }else if(command == NODE_MSG_MANAGER_PURGE_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_clientsManager->getPurgeMode()));
      }else if(command == NODE_MSG_MANAGER_PURGE_MODE_SET){
        //!<Parse command parameters
        int purgeMode = 0;
        //!<Parse purge mode from parameters
        std::istringstream(parameters) >> purgeMode;
        //!<Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_clientsManager->getPurgeMode()));
        //!<Set new purge mode
        _clientsManager->setPurgeMode(purgeMode);
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_SET){
        //!<Parse command parameters
        int hbDelay = HEARTBEAT_DELAY;
        //!<Parse heartbeat delay from parameters
        std::istringstream(parameters) >> hbDelay;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
        //!<Set new heartbeat delay value
        setHeartbeatDelay(static_cast<unsigned int>(hbDelay*1000));
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(_clientsManager->getHeartbeatTimeout()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_SET){
        //!<Parse command parameters
        int hbTimeout = HEARTBEAT_TIMEOUT;
        //!<Parse heartbeat timeout from parameters
        std::istringstream(parameters) >> hbTimeout;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(_clientsManager->getHeartbeatTimeout()/1000);
        //!<Set new heartbeat timeout value
        setHeartbeatTimeout(static_cast<unsigned int>(hbTimeout*1000));
        _clientsManager->setHeartbeatTimeout(getHeartbeatTimeout());
      }else if(command == NODE_MSG_HEARTBEAT_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
      }else if(command == NODE_MSG_HEARTBEAT_MODE_SET){
        //!<Parse command parameters
        int hbMode = HEARBEAT_MODE_DEFAULT;
        //!<Parse heartbeat mode from parameters
        std::istringstream(parameters) >> hbMode;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
        //!<Set new heartbeat mode value
        setHeartbeatMode(static_cast<unsigned int>(hbMode));
      }else if(command == NODE_MSG_ROUTES){
        //!<Return cluster node routes list
        std::vector<ClientWorkerItem> clientsQueue;
        RouteMessage routeMessage(clientsQueue);
        ClientsQueueManager clientsManager(&clientsQueue, 0);
        clientsManager.refresh(getClientIdentity(), _clientsQueue);
        ret = routeMessage.build();
      }else if(command == NODE_MSG_RESOURCE_USAGE){
        //!<Return cluster node resource usage
        std::string json = "";
        HCE::drce::ResourceUsageExtractor::getInstance().update();
        HCE::drce::ResourceUsage resourceUsage = HCE::drce::ResourceUsageExtractor::getInstance().extract();
        HCE::drce::ResourceUsageSerializator resourceUsageSerializator(resourceUsage);
        resourceUsageSerializator.serialize(json);
        if (resourceUsageSerializator.isError())
          logger.log(NODE_LOG_MODE_ERROR) << "Get resource usage error: "<< resourceUsageSerializator.getErrorMsg() << HCE::flush;
        ret = json;
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }
      return ret;
    }

    //!<Get property information
    std::string DataServerProxy::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "clientRequests=" << getStatClientRequests()
          << NODE_ADMIN_COMMAND_DELIMITER << "clientResponses=" << getStatClientResponses()
          << NODE_ADMIN_COMMAND_DELIMITER << "clientRequestsSize=" << getStatClientRequestsSize()
          << NODE_ADMIN_COMMAND_DELIMITER << "clientResponsesSize=" << getStatClientResponsesSize()
          << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << getStatRequests()
          << NODE_ADMIN_COMMAND_DELIMITER << "requestsSize=" << getStatRequestsSize()
          << NODE_ADMIN_COMMAND_DELIMITER << "nodeMode=" << static_cast<int>(getNodeMode())
          << NODE_ADMIN_COMMAND_DELIMITER << "clients=" << _clientsManager->getDump()
          << NODE_ADMIN_COMMAND_DELIMITER << "statQPSClient=" << getStatQPSClient()
          << NODE_ADMIN_COMMAND_DELIMITER << "statQPSServer=" << getStatQPSServer()
          << NODE_ADMIN_COMMAND_DELIMITER << "statRPSClient=" << getStatRPSClient()
          << NODE_ADMIN_COMMAND_DELIMITER << "statRPSServer=" << getStatRPSServer()
          << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidthClient=" << getStatRequestsBandwidthClient()
          << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidthServer=" << getStatRequestsBandwidthServer()
          << NODE_ADMIN_COMMAND_DELIMITER << "statResponsesBandwidthClient=" << getStatResponsesBandwidthClient()
          << NODE_ADMIN_COMMAND_DELIMITER << "statResponsesBandwidthServer=" << getStatResponsesBandwidthServer();
      return result.str();
    }

    void DataServerProxy::logPeriodicStatistics(void){
      //!<Out periodical stats
      if(s_clock() > _pstatsAt){

        setStatQPSClient((getStatClientRequests() - getPeriodicStatClientRequests()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatQPSServer((getStatRequests() - getStatRequestsProcessed()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatRPSClient((getStatClientResponses() - getPeriodicStatClientResponses()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatRPSServer(0);
        setStatRequestsBandwidthClient((getStatClientRequestsSize() - getPeriodicStatClientRequestsSize()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatRequestsBandwidthServer((getStatRequestsSize() - getStatRequestsProcessedSize()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatResponsesBandwidthClient((getStatClientResponsesSize() - getPeriodicStatClientResponsesSize()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatResponsesBandwidthServer(0);

        setStatRequestsProcessed(getStatRequests());
        setStatRequestsProcessedSize(getStatRequestsSize());
        setPeriodicStatClientRequests(getStatClientRequests());
        setPeriodicStatClientResponses(getStatClientResponses());
        setPeriodicStatClientRequestsSize(getStatClientRequestsSize());
        setPeriodicStatClientResponsesSize(getStatClientResponsesSize());

        logger.log(NODE_LOG_MODE_DEBUG) << "SRequests=" << getStatRequests() << ", CRequests=" << getStatClientRequests() << ", CResponses=" << getStatClientResponses()
            << ", SQPS=" << getStatQPSServer() << ", CQPS=" << getStatQPSClient() << ", CRPS=" << getStatRPSClient() << ", SRequestsSize=" << getStatRequestsSize()
            << ", CRequestsSize=" << getStatClientRequestsSize() << ", CResponsesSize=" << getStatClientResponsesSize()
            << ", SRequestsBandwidth=" << getStatRequestsBandwidthServer() << ", CRequestsBandwidth=" << getStatRequestsBandwidthClient()
            << ", CResponsesBandwidth=" << getStatResponsesBandwidthClient() << flush;

        //Shift next log stat action time
        _pstatsAt = s_clock() + STAT_INTERVAL;
      }
    }

    void DataServerProxy::heartbit(void){

      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::heartbit  HeartbeatMode: " << getHeartbeatMode() << " _heartbeatAt = '"
          << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;

      //!<Send heartbeats to idle clients if it's time
      if(s_clock() > _heartbeatAt){
        if (getHeartbeatMode() == HEARBEAT_MODE_ADAPTIVE){

          for(std::vector<ClientWorkerItem>::iterator it = _clientsQueue.begin(); it < _clientsQueue.end(); ++it){
            it->shiftCounter.responceTimeStamp(it->rtimestamp);
            it->shiftCounter.recount(getHeartbeatTimeout());

            logger.log(NODE_LOG_MODE_TRACE) << "[ " << it->identity << " ] delta = " << abs(s_clock()-_heartbeatAt+getHeartbeatDelay())
                << " _shiftCounter.getShift()= " << it->shiftCounter.getShift() << flush;

            size_t delta = s_clock()-_heartbeatAt+getHeartbeatDelay();
            if (delta > (getHeartbeatTimeout()+it->shiftCounter.getShift())){

              //!<Execute sending heartbeat
              sendHeartbeat(it->identity);
              //!<Reset response shift counter
              it->shiftCounter.reset();

              logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << getClientIdentity() << "] to server _heartbeatAt = '"
                  << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;
            }
          }
         //!<Send heartbeats to idle clients if it's time
        }else{

          for(std::vector<ClientWorkerItem>::iterator it = _clientsQueue.begin(); it != _clientsQueue.end(); ++it){
            //!<Execute sending heartbeat
            sendHeartbeat(it->identity);
            logger.log(NODE_LOG_MODE_DEBUG) << "HEARTBEAT message for client [" << it->identity.c_str() << "] sent by timer reason" << flush;
          }
        }
        //!<Reset heartbit time
        _heartbeatAt = s_clock() + getHeartbeatDelay();
      }
    }

    void DataServerProxy::sendHeartbeat(const std::string& identity){
      //!<Execute sending heartbeat
      zmsg msg(NODE_MSG_HEARTBEAT.c_str());
      msg.wrap(identity.c_str(), NULL);
      msg.send(*_dataServerSock);
    }

    void* DataServerProxy::process(void){
      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      while(getInProgress()){
        try {
          std::vector<zmq::pollitem_t> items;
          if(_dataServerSock->connected()){
            items ={{*_dataServerSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          } else{
            items ={{*_inprocProxySock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          }

          //!<Poll dataserver
          zmq::poll(&items[0], items.size(), getPollTimeout());

          //!<Handle dataserver socket's message if socket is connected
          if(_dataServerSock->connected()){
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_proxy socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleAdminMessage();
          }

          //!<Heartbeats for clients
          heartbit();

          //!<Log stats
          logPeriodicStatistics();

          //!<Send periodical property values
          sendPeriodicProperties();

          //!<Dump run-time variables at regular intervals
          dumpRunTimeVariables();

          //!<Remove expired clients from queue
          std::string r = _clientsManager->purge();
          if(r != ""){
            logger.log(NODE_LOG_MODE_INFORMATION) << "PURGED_CLIENTS: " << r << flush;
          }
        }catch(zmq::error_t& error){
          if (EINTR!=error.num()){
            std::cout << "DataServerProxy E: " << error.what() << std::endl;
          }
        }catch(std::exception& e){
          std::cout << "DataServerProxy E: " << e.what() << std::endl;
        }catch(...){
          std::cout << "DataServerProxy E: Unknown exception" << std::endl;
        }
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }

  }
}
