/**
 * @file DataServerProxy.hpp
 *
 * HCE project DataServerProxy handler.
 * Implements functionality for data server proxy handler.
 * Contains client data requests server used in data and proxy nodes.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_DATASERVER_PROXY_INCLUDED__
#define __HCE_DATASERVER_PROXY_INCLUDED__

#include <sstream>
#include <map>

#include "Handler.hpp"
#include "DataNodeOptions.hpp"

//Node ini
#include <Poco/Util/IniFileConfiguration.h>
#include "Poco/String.h"
//Message cover object
#include "JsonMessageCover.hpp"
//Calculator balancing nodes for execution
#include "CalculatorNodeExecution.hpp"
//Checker of resource usage limits
#include "ResourceUsageLimitsChecker.hpp"

namespace HCE {
  namespace handlers {

    class DataServerProxy: public HCE::handlers::Handler, protected HCE::handlers::DataNodeOptions {
      public:
        DataServerProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString,
            unsigned char nodeMode, int64_t startedAt, unsigned char logPriority, const std::string& iniFile="");
        ~DataServerProxy();

      protected:
        int initialize(void);
        void deinitialize(void);
        void* process(void);

        void handleInternalMessage(void);
        void handleExternalMessage(void);

        std::string processAdminCommands(std::string& command, std::string& parameters);

        void logPeriodicStatistics(void);
        void heartbit(void);
        void sendHeartbeat(const std::string& identity);
        void registerClient(const std::string& identity);

        void processControlMessage(zmsg& msg);
        void processDataMessage(zmsg& msg);

        //!<Get property information
        std::string getPropertyInformation(void);

        enum class EmptyResponseType : std::int8_t
        {
          ertNoClients=0,
          ertBadNodeMode=1,
          ertNotEnoughResourceLimits=2
        };
        //!<Make empty response message according to response type
        std::string makeEmptyResponseMessage(const std::string& messageBody,
            EmptyResponseType responseType,
            const std::string& messageRoute,
            const std::string& errorMsg="");

        //!<Send empty response message to reducer
        void sendEmptyResponseMessage(zmsg& msg,
            const std::string& coverMessageBody,
            EmptyResponseType responseType,
            const std::string& messageId,
            const std::string& messageRoute,
            const std::string& errorMsg="");

        //!<Make and send response message
        void sendResponseMessage(const std::string& identity,
            const std::string& messageId,
            const std::string& messageBody,
            const std::string& messageRoute);

        //!<Node initialization from ini file
        void nodeInit(const std::string& iniFile);

        //!<Check initialization and set default value if necessary
        void checkInit(void);

        //!<ZMQ Sockets
        zmq::socket_t* _inprocProxySock;
        zmq::socket_t* _inprocReducerInSock;
        zmq::socket_t* _dataServerSock;

        //!<Previous client number for random and resource usage algorithms
        size_t _previousClientNumberRandom;
        size_t _previousClientNumberResourceUsage;

        //!<Random client number genertor
        size_t getRandomNodeNumber(size_t maxNumber);
        //!<Get identity use random algorithm
        std::string getIdentityRandom(std::vector<ClientWorkerItem>& clients,
            size_t& previousClientNumber);
        //!<Get identity use resource usage algorithm
        std::string getIdentityResourceUsage(std::vector<ClientWorkerItem>& clients,
            size_t& previousClientNumber,
            CalculatorAlgorithm::Types typeAlgorithm,
            const CalculatorAlgorithmData& algorithmData);

        //!<Algorithm data for balancing calculator
        CalculatorAlgorithmData calculatorAlgorithmData;
        //!<Resource usage limits data
        ResourceUsageLimitsData resourceUsageLimitsData;
        //!<Type of algorithm for balancing calculator
        CalculatorAlgorithm::Types typeCalculatorAlgorithm;

        //!<Resource usage manager
        ResourceUsageManager _resourceUsageManager;

        //!<Queue of connected clients
        std::vector<ClientWorkerItem> _clientsQueue;
        ClientsQueueManager* _clientsManager;

        //!<Heartbeat at time mark
        int64_t _heartbeatAt;

        //!<Node node as string
        std::string getNodeModeName(unsigned int nodeMode);

        //!<Node mode: router, shard manager, replica manager or replica.
        void setNodeMode(unsigned int nodeMode);
        unsigned int getNodeMode(void);

        //!<Resource usage algorithm
        void setResourcesUsageAlg(unsigned int resourcesUsageAlg);
        unsigned int getResourcesUsageAlg(void);

        //!<Resource usage alg0 weights (default algorithm)
        void setResourcesUsageAlg0Weights(const std::string& resourcesUsageAlg0Weights);
        std::string getResourcesUsageAlg0Weights(void);

        //!<Resource usage order
        void setResourcesUsageOrder(unsigned int resourcesUsageOrder);
        unsigned int getResourcesUsageOrder(void);

        //!<Resource usage limits
        void setResourcesUsageLimits(const std::string& resourcesUsageLimits);
        std::string getResourcesUsageLimits(void);

        //!<Resource usage collected size
        void setResourcesUsageCollectedSize(unsigned int resourcesUsageCollectedSize);
        unsigned int getResourcesUsageCollectedSize(void);

        //!<Resource usage max allowed range for calculation of weight in percent
        void setResourcesUsageMaxAllowedRange(double resourcesUsageMaxAllowedRange);
        double getResourcesUsageMaxAllowedRange(void);

        //!<DRCE resources extractor max threads
        void setResourcesExtractorMaxThreads(unsigned int resourcesExtractorMaxThreads);
        unsigned int getResourcesExtractorMaxThreads(void);

        //!<DRCE resources extractor max processes
        void setResourcesExtractorMaxProcesses(unsigned int resourcesExtractorMaxProcesses);
        unsigned int getResourcesExtractorMaxProcesses(void);

        //!<DRCE resources extractor max disk size
        void setResourcesExtractorMaxDiskSize(unsigned int resourcesExtractorMaxDiskSize);
        unsigned int getResourcesExtractorMaxDiskSize(void);

        //!<DRCE resources extractor max vram size
        void setResourcesExtractorMaxVramSize(unsigned int resourcesExtractorMaxVramSize);
        unsigned int getResourcesExtractorMaxVramSize(void);

        //!<DRCE resources extractor max rram size
        void setResourcesExtractorMaxRramSize(unsigned int resourcesExtractorMaxRramSize);
        unsigned int getResourcesExtractorMaxRramSize(void);

        //!<Setters and getters for stat variables
        void setStatClientRequests(unsigned int statClientRequests);
        unsigned int getStatClientRequests(void);

        void setStatClientResponses(unsigned int statClientResponses);
        unsigned int getStatClientResponses(void);

        void setStatRequests(unsigned int statRequests);
        unsigned int getStatRequests(void);

        void setStatClientRequestsSize(unsigned int statRequestsSize);
        unsigned int getStatClientRequestsSize(void);

        void setStatClientResponsesSize(unsigned int statResponsesSize);
        unsigned int getStatClientResponsesSize(void);

        void setStatRequestsSize(unsigned int statRequestsSize);
        unsigned int getStatRequestsSize(void);

        void setPeriodicStatClientRequests(unsigned int periodicStatClientRequests);
        unsigned int getPeriodicStatClientRequests(void);

        void setPeriodicStatClientResponses(unsigned int periodicStatClientResponses);
        unsigned int getPeriodicStatClientResponses(void);

        void setStatRequestsProcessed(unsigned int statRequestsProcessed);
        unsigned int getStatRequestsProcessed(void);

        void setStatRequestsProcessedSize(unsigned int statRequestsProcessedSize);
        unsigned int getStatRequestsProcessedSize(void);

        void setPeriodicStatClientRequestsSize(unsigned int periodicStatClientRequestsSize);
        unsigned int getPeriodicStatClientRequestsSize(void);

        void setPeriodicStatClientResponsesSize(unsigned int periodicStatClientResponsesSize);
        unsigned int getPeriodicStatClientResponsesSize(void);

        void setStatQPSClient(float statQPSClient);
        float getStatQPSClient(void);

        void setStatQPSServer(float statQPSServer);
        float getStatQPSServer(void);

        void setStatRPSClient(float statRPSClient);
        float getStatRPSClient(void);

        void setStatRPSServer(float statRPSServer);
        float getStatRPSServer(void);

        void setStatRequestsBandwidthClient(float statRequestsBandwidthClient);
        float getStatRequestsBandwidthClient(void);

        void setStatRequestsBandwidthServer(float statRequestsBandwidthServer);
        float getStatRequestsBandwidthServer(void);

        void setStatResponsesBandwidthClient(float statResponsesBandwidthClient);
        float getStatResponsesBandwidthClient(void);

        void setStatResponsesBandwidthServer(float statResponsesBandwidthServer);
        float getStatResponsesBandwidthServer(void);

        //!<Send out periodical stats at regular intervals
        int64_t _pstatsAt = s_clock() + HEARTBEAT_DELAY;
    };
  }
}

#endif /* __HCE_DATASERVER_PROXY_INCLUDED__ */
