/**
 * @Handler.hpp
 *
 * HCE project node handler's parent.
 * Implements basic functionality common for any kind of node handler.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_HANDLER_INCLUDED__
#define __HCE_HANDLER_INCLUDED__

#include <bitset>

#include "zmsg.hpp"
#include "HandlerTypes.hpp"
#include "ClientsQueueManager.hpp"
#include "AdminCommandParameters.hpp"

//POCO date time formatter dep.
#include <Poco/Timestamp.h>
#include <Poco/DateTimeFormatter.h>
#include <Poco/Logger.h>
#include <Poco/AutoPtr.h>
#include <Poco/Configurable.h>
#include <Poco/Util/IniFileConfiguration.h>
#include <Poco/String.h>
#include <Poco/Path.h>
#include <Poco/File.h>

#include "LoggerStream.hpp"
#include "ServiceMessages.hpp"
#include "HandlerProperties.hpp"

namespace HCE {
  namespace handlers {

    class Handler {
      public:
        Handler(const std::string& name, zmq::context_t& context, const std::string& clientIdentity, const std::string& connectionString,
            int64_t startedAt, unsigned char logPriority, const std::string& iniFile="") :
            _inprocAdminSock(NULL), handlerProperties(), statProperties(), _context(context), _logMask(NODE_LOG_DEAFULT_MASK),
            _logTimeFormat(NODE_LOG_DEAFULT_TIME_FORMAT), log(), logger(name), _inProgress(true),
            _propertyIntervalAt(s_clock()-PROPERTY_INTERVAL), _dumpIntervalAt(s_clock()+DUMP_INTERVAL), _dumpDir(HCE::handlers::DUMP_DIR) {
          //!<Set default value to allowed dump restore
          setDumpAllowRestore(DUMP_ALLOWED_RESTORE);
          setPropertyInterval(PROPERTY_INTERVAL);
          setDumpInterval(DUMP_INTERVAL);
          //!<Read initialization data from file
          initHandler(iniFile, logger);

          setName(name);
          setClientIdentity(clientIdentity);
          setStartedAt(startedAt);
          setLogPriority(logPriority);
          setBindTriesMaxNumber(NODE_CONNECTION_REBUILD_MAX_TRIES);
          setBindTriesDelay(NODE_CONNECTION_REBUILD_DELAY_DEFAULT);
          setConnectionString(connectionString);
          setPollTimeout(POLL_TIMEOUT);
          //!<Check is allowed dump restore and restore if necessary
          if (getDumpAllowRestore() > 0){
            //!<Load handler properties values
            if (!handlerProperties.load(getDumpFileName())){
              logger.log(HCE::handlers::NODE_LOG_MODE_ERROR) << "Load handler properties has error: " << handlerProperties.getErrorMsg() << flush;
            }else{
              logger.log(HCE::handlers::NODE_LOG_MODE_DEBUG) << "Load handler properties - SUCCESS" << flush;
            }
            //!<Load stistic properties values
            if (!statProperties.load(getStatFileName())){
              logger.log(HCE::handlers::NODE_LOG_MODE_ERROR) << "Load stat properties has error: " << statProperties.getErrorMsg() << flush;
            }else{
              logger.log(HCE::handlers::NODE_LOG_MODE_DEBUG) << "Load stat properties - SUCCESS" << flush;
            }
          }
        }

        virtual ~Handler(){
          //!<Save handler properties values
          if (!handlerProperties.save(getDumpFileName())){
            logger.log(HCE::handlers::NODE_LOG_MODE_DEBUG) << "Save handler properties has error: " << handlerProperties.getErrorMsg() << flush;
          }

          //!<Save stistic properties values
          if (!statProperties.save(getStatFileName())){
            logger.log(HCE::handlers::NODE_LOG_MODE_DEBUG) << "Save stat properties has error: " << statProperties.getErrorMsg() << flush;
          }

          if(_inprocAdminSock){
            if(_inprocAdminSock->connected()){
              _inprocAdminSock->close();
            }
            delete _inprocAdminSock;
            _inprocAdminSock = NULL;
          }
        }

        static void* main(void *that){
          return (static_cast<Handler*>(that))->process();
        }

        void shutdown(void){
          _inProgress = false;
        }

        //!<Get handler instance name
        std::string getName(void) {
          return handlerProperties.get<std::string>(HCE::handlers::properties::NAME);
        }

      protected:

        //!<Set handler instance name
        void setName(const std::string& name){
          handlerProperties.set(HCE::handlers::properties::NAME, name);
        }

        //!<Set client ID
        void setClientIdentity(const std::string& clientIdentity){
          handlerProperties.set(HCE::handlers::properties::CLIENT_IDENTITY, clientIdentity);
        }

        //!<Get client ID
        std::string getClientIdentity(void) {
          return handlerProperties.get<std::string>(HCE::handlers::properties::CLIENT_IDENTITY);
        }

        //!<Set start time mark
        void setStartedAt(int64_t startedAt) {
          handlerProperties.set(HCE::handlers::properties::STARTED_AT, startedAt);
        }

        //!<Get start time mark
        int64_t getStartedAt(void) {
          return handlerProperties.get<int64_t>(HCE::handlers::properties::STARTED_AT);
        }

        //!<Set log priority level
        void setLogPriority(unsigned int logPriority) {
          handlerProperties.set(HCE::handlers::properties::LOG_PRIORITY, logPriority);
        }

        //!<Get log priority level
        unsigned int getLogPriority(void) {
          return handlerProperties.get<unsigned int>(HCE::handlers::properties::LOG_PRIORITY);
        }

        //!<Set ini file name
        void setIniFile(const std::string& iniFile){
          handlerProperties.set(HCE::handlers::properties::INI_FILE, iniFile);
        }

        //!<Get ini file name
        std::string getIniFile(void){
          return handlerProperties.get<std::string>(HCE::handlers::properties::INI_FILE);
        }

        //!<Set ZMQ connections rebuild tries max number
        void setBindTriesMaxNumber(const unsigned int& bindTriesMaxNumber){
          handlerProperties.set(HCE::handlers::properties::CONNECTION_REBUILD_TRIES_MAX, bindTriesMaxNumber);
        }

        //!<Get ZMQ connections rebuild tries max number
        unsigned int getBindTriesMaxNumber(void) {
          return handlerProperties.get<unsigned int>(HCE::handlers::properties::CONNECTION_REBUILD_TRIES_MAX);
        }

        //!<Set ZMQ connections rebuild tries delay
        void setBindTriesDelay(const unsigned int& bindTriesDelay){
          handlerProperties.set(HCE::handlers::properties::CONNECTION_REBUILD_TRIES_DELAY, bindTriesDelay);
        }

        //!<Get ZMQ connections rebuild tries delay
        unsigned int getBindTriesDelay(void) {
          return handlerProperties.get<unsigned int>(HCE::handlers::properties::CONNECTION_REBUILD_TRIES_DELAY);
        }

        //!<Set ZMQ connection string
        void setConnectionString(const std::string& connectionString){
          handlerProperties.set(HCE::handlers::properties::CONNECTION_STRING, connectionString);
        }

        //!<Get ZMQ connection string
        std::string getConnectionString(void) {
          return handlerProperties.get<std::string>(HCE::handlers::properties::CONNECTION_STRING);
        }

        //!<Set flag for polling condition
        void setInProgress(bool in_progress){
          _inProgress = in_progress;
        }

        //!<Get flag for polling condition
        bool getInProgress(void) const{
          return _inProgress;
        }

        //!<Set poll timeout value
        void setPollTimeout(unsigned int pollTimeout){
          if (pollTimeout>=MIN_ALLOWED_POLL_TIMEOUT){
            handlerProperties.set(HCE::handlers::properties::POLL_TIMEOUT, pollTimeout);
          }
        }

        //!<Get poll timeout value
        unsigned int getPollTimeout(void) {
          return handlerProperties.get<unsigned int>(HCE::handlers::properties::POLL_TIMEOUT);
        }

        //!<Set property interval between resend Property values
        void setPropertyInterval(unsigned int propertyInterval){
          if (propertyInterval >= MIN_ALLOWED_PROPERTY_INTERVAL){
              handlerProperties.set(HCE::handlers::properties::PROPERTY_INTERVAL, propertyInterval);
          }
        }

        //!<Get property interval between resend Property values
        unsigned int getPropertyInterval(void) {
          return handlerProperties.get<unsigned int>(HCE::handlers::properties::PROPERTY_INTERVAL);
        }

        //!<Set interval between dump run-time variables
        void setDumpInterval(unsigned int dumpInterval){
          if (dumpInterval >= MIN_ALLOWED_DUMP_INTERVAL){
            handlerProperties.set(HCE::handlers::properties::DUMP_INTERVAL, dumpInterval);
            _dumpIntervalAt = s_clock() + dumpInterval;
          }
        }

        //!<Get interval between dump run-time variables
        unsigned int getDumpInterval(void) {
          return handlerProperties.get<unsigned int>(HCE::handlers::properties::DUMP_INTERVAL);
        }

        //!<Set flag for allowed restore variables use dump
        void setDumpAllowRestore(unsigned int dumpAllowRestore){
          handlerProperties.set(HCE::handlers::properties::DUMP_ALLOW_RESTORE, dumpAllowRestore);
        }

        //!<Get interval between dump run-time variables
        unsigned int getDumpAllowRestore(void) {
          return handlerProperties.get<unsigned int>(HCE::handlers::properties::DUMP_ALLOW_RESTORE);
        }

        void logFlush(unsigned int priority = NODE_LOG_MODE_DEFAULT){
          if(priority == getLogPriority() || priority == NODE_LOG_MODE_DEFAULT){
            //Fill time mark
            std::string timeMark = Poco::DateTimeFormatter::format(Poco::Timestamp(static_cast<Poco::Timestamp::TimeVal>(s_clock() * 1000)), _logTimeFormat);

            //Backup log message
            std::string message = log.str();

            //Reset log accumulator
            log.str(std::string());
            log.clear();

            //Create log context
            if(_logMask[0]){
              //Out time diff
              log << std::setfill('0') << std::setw(8) << ((unsigned int)(s_clock() - getStartedAt())) << " ";
            }
            //Out three more log components
            log << ((_logMask[1] ? timeMark : "") + (_logMask[2] ? "[" + getName() + "] " : "") + (_logMask[3] ? message : "")) << std::endl;

            //Out final log context
            std::cout << log.str();
            std::cout.flush();
          }

          //Reset log accumulator
          log.str(std::string());
          log.clear();
        }

        //!<Node init from file
        void initHandler(const std::string& iniFile, LoggerStream& logger){
          if (!iniFile.empty()){
            logger.log(NODE_LOG_MODE_INFORMATION) << "load node settings from '" << iniFile << "' file" << flush;

            setIniFile(iniFile);

            std::string dumpDir = HCE::handlers::DUMP_DIR;
            //!<Process Node's ini file
            try
            {
              Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
              try
              {
                setDumpAllowRestore(pConf->getInt(HCE::handlers::properties::DUMP_ALLOW_RESTORE, DUMP_ALLOWED_RESTORE));
                setPropertyInterval(pConf->getInt(HCE::handlers::properties::PROPERTY_INTERVAL)*1000);
                setDumpInterval(pConf->getInt(HCE::handlers::properties::DUMP_INTERVAL)*1000);
                dumpDir = pConf->getString(HCE::handlers::properties::DUMP_DIR);
              }
              catch(Poco::SyntaxException& e)
              {
                //!<Log some error of config file processing
                logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node options ini file processing error: '" << e.displayText() << "'" << flush;
              }
            }
            catch(Poco::Exception& e)
            {
              //!<Log some error of config file processing
              logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node options ini file access error: '" << e.displayText() << "'" << flush;
            }
            //!<Apply parameters use default values
            if (!handlerProperties.isExist(HCE::handlers::properties::PROPERTY_INTERVAL))
              setPropertyInterval(PROPERTY_INTERVAL);

            if (!handlerProperties.isExist(HCE::handlers::properties::DUMP_INTERVAL))
              setDumpInterval(DUMP_INTERVAL);

            if (!handlerProperties.isExist(HCE::handlers::properties::DUMP_ALLOW_RESTORE))
              setDumpAllowRestore(DUMP_ALLOWED_RESTORE);

            _dumpDir = dumpDir;

            logger.log(NODE_LOG_MODE_INFORMATION) << HCE::handlers::properties::PROPERTY_INTERVAL << " = " << getPropertyInterval() << flush;
            logger.log(NODE_LOG_MODE_INFORMATION) << HCE::handlers::properties::DUMP_INTERVAL << " = " << getDumpInterval() << flush;
            logger.log(NODE_LOG_MODE_INFORMATION) << HCE::handlers::properties::DUMP_ALLOW_RESTORE << " = " << getDumpAllowRestore() << flush;
            logger.log(NODE_LOG_MODE_INFORMATION) << HCE::handlers::properties::DUMP_DIR << " = " << _dumpDir << flush;
          }
        }

        void handleAdminMessage(void){
          //!<Fetch request message from inproc_admin socket
          zmsg msg(*_inprocAdminSock);

          logger.log(NODE_LOG_MODE_TRACE) << " MSG FROM inproc_admin\n" << msg.dump() << flush;

          //!<Process message
          processAdminMessage(msg);

          //!<Send response message
          msg.send(*_inprocAdminSock);
        }

        void processAdminMessage(zmsg& msg){
          //!<Fetch message Id
          std::string messageId((const char*)msg.pop_front().c_str());
          //!<Fetch message command
          std::string messageCommand((const char*)msg.pop_front().c_str());
          //!<Fetch message command parameters
          std::string messageParams((const char*)msg.pop_front().c_str());
          //!<Fetch message source Identity
          std::string messageSrcIdentity((const char*)msg.pop_front().c_str());

          logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED admin message from inproc_admin id [" << messageId << "], body [" << messageCommand << "], params ["
              << messageParams << "], source identity [" << messageSrcIdentity << "]" << flush;

          //!<Create response
          std::string response = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER;

          if(messageCommand == NODE_MSG_ECHO){
            response += NODE_MSG_ECHO + NODE_ADMIN_COMMAND_DELIMITER + getName();
          }else if(messageCommand == NODE_MSG_LOG_LEVEL_GET){
            response += NODE_ADMIN_PROPERTY_NAME_LOG + "=" + std::to_string(getLogPriority()) + NODE_ADMIN_COMMAND_DELIMITER + getName();
          }else if(messageCommand == NODE_MSG_LOG_LEVEL_SET){
            //!<Parse command parameters and fetch log level value
            unsigned int logp = getLogPriority();

            std::istringstream(messageParams) >> logp;
            response += NODE_ADMIN_PROPERTY_NAME_LOG + "=" + std::to_string(getLogPriority()) + NODE_ADMIN_COMMAND_DELIMITER + getName();
            setLogPriority(logp);
            Poco::Logger::root().setLevel(getLogPriority());
          }else if(messageCommand == NODE_MSG_POLL_TIMEOUT_GET){
            response += NODE_ADMIN_PROPERTY_POLL_TIMEOUT + "=" + std::to_string(getPollTimeout()/1000);
          }else if(messageCommand == NODE_MSG_POLL_TIMEOUT_SET){
            //!<Parse command parameters and fetch log level value
            unsigned int pollTimeout = getPollTimeout();
            //Parse log priority from parameters
            std::istringstream(messageParams) >> pollTimeout;
            //Make response
            response += NODE_ADMIN_PROPERTY_POLL_TIMEOUT + "=" + std::to_string(getPollTimeout()/1000);
            //Set new poll timeout value
            setPollTimeout(pollTimeout*1000);
          }else if(messageCommand == NODE_MSG_PROPERTY_INTERVAL_GET){
            response += NODE_ADMIN_PROPERTY_PROPERTY_INTERVAL + "=" + std::to_string(getPropertyInterval()/1000);
          }else if(messageCommand == NODE_MSG_PROPERTY_INTERVAL_SET){
            //!<Parse command parameters and fetch property interval value
            unsigned int propertyInterval = getPropertyInterval();
            //Parse property interval from parameters
            std::istringstream(messageParams) >> propertyInterval;
            //Make response
            response += NODE_ADMIN_PROPERTY_PROPERTY_INTERVAL + "=" + std::to_string(getPropertyInterval()/1000);
            //Set new property interval value
            setPropertyInterval(propertyInterval*1000);
          }else if(messageCommand == NODE_MSG_DUMP_INTERVAL_GET){
            response += NODE_ADMIN_PROPERTY_DUMP_INTERVAL + "=" + std::to_string(getDumpInterval()/1000);
          }else if(messageCommand == NODE_MSG_DUMP_INTERVAL_SET){
            //!<Parse command parameters and fetch dump interval value
            unsigned int dumpInterval = getDumpInterval();
            //Parse interval value from parameters
            std::istringstream(messageParams) >> dumpInterval;
            //Make response
            response += NODE_ADMIN_PROPERTY_DUMP_INTERVAL + "=" + std::to_string(getDumpInterval()/1000);
            //Set new dump interval value
            setDumpInterval(dumpInterval*1000);
          }else if(messageCommand == NODE_MSG_STOP){
            //!<Stop handler main loop as a result of processing of the "Shutdown" command for "Admin" handler
            setInProgress(false);
          }else{
            //!<Call admin commands processing redefined in the derivative class instance
            response = processAdminCommands(messageCommand, messageParams);
          }

          //!<Insert message Id
          msg.append(messageId.c_str());
          //!<Insert response
          msg.append(response.c_str());
          //!<Insert messageSrcIdentity
          msg.append(messageSrcIdentity.c_str());
        }

        virtual std::string processAdminCommands(std::string& command, std::string& parameters){
          std::string ret("DEFAULT stub for command processor!");

          return ret;
        }

        void rebuildAdminClientSocket(void){
          if(_inprocAdminSock){
            if(_inprocAdminSock->connected()){
              _inprocAdminSock->close();
            }
            delete _inprocAdminSock;
            _inprocAdminSock = NULL;
          }

          //!<Create inproc_admin socket
          _inprocAdminSock = new zmq::socket_t(_context, ZMQ_DEALER);
          if(_inprocAdminSock){
            //!<Set identity for client socket
            _inprocAdminSock->setsockopt(ZMQ_IDENTITY, getName().c_str(), getName().size());
            //!<Connect to inproc_admin socket
            _inprocAdminSock->connect(NODE_CS_INPROC_ADMIN.c_str());
            logger.log(NODE_LOG_MODE_DEBUG) << "Connected inproc_admin as [" << getName() << "]" << flush;
            //!<Send ready message
            logger.log(NODE_LOG_MODE_DEBUG) << "Sending ready message to admin..." << flush;
            s_send(*_inprocAdminSock, NODE_MSG_READY);
            logger.log(NODE_LOG_MODE_DEBUG) << "Sent ready message to admin" << flush;
          }else{
            logger.log(NODE_LOG_MODE_FATAL) << "Admin client socket instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
            ::abort();
          }
        }

        bool rebuildServerConnection(zmq::socket_t*& sock_p, const std::string& connectionString, const int type){
          bool bound = false;

          //!<Delete socket if already exists
          if(sock_p != NULL){
            //TODO: experimental
            if(sock_p->connected()){
              sock_p->close();
              s_sleep(100);
            }
            delete sock_p;
            sock_p = NULL;
          }

          //!<Create server's socket
          sock_p = new zmq::socket_t(_context, type);
          if(sock_p != NULL){
            //!<Try to bind
            unsigned int counter = 0;
            while(counter < getBindTriesMaxNumber()){
              try{
                sock_p->bind(connectionString.c_str());
                bound = true;
                break;
              } catch(zmq::error_t &e){
                logger.log(NODE_LOG_MODE_ERROR) << "Bind exeption : " << e.what() << flush;
                counter++;
                s_sleep(getBindTriesDelay());
              }
            }
          }else{
            logger.log(NODE_LOG_MODE_FATAL) << "Server connection socket type " << type << " instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
            ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
          }

          logger.log(NODE_LOG_MODE_DEBUG) << "Bind to [" << connectionString << "] ";
          if(bound){
            logger.log(NODE_LOG_MODE_DEBUG) << "success";
          }else{
            logger.log(NODE_LOG_MODE_DEBUG) << "error";
          }
          logger.log(NODE_LOG_MODE_DEBUG) << flush;

          return bound;
        }

        void rebuildClientConnection(zmq::socket_t*& sock_p, const std::string& connectionString, const int type, const std::string& identity=""){
          rebuildClientConnection(sock_p, connectionString, type, identity, -1);
        }

        bool rebuildClientConnection(zmq::socket_t*& sock_p, const std::string& connectionString, const int type, const std::string& identity, int linger, const std::string& readyMessage =
            "", const std::string& byeMessage = NODE_MSG_BYE){
          bool connected = false;

          //!<Delete socket if already exists
          if(sock_p != NULL){
            if(sock_p->connected()){
              if(byeMessage != ""){
                //!<Send BYE message to the server (especially for the Paranoid Pirate Protocol implementations)
                sendByeMessage(sock_p, identity, byeMessage);
              }

              logger.log(NODE_LOG_MODE_DEBUG) << "rebuildServerConnection()::close();" << flush;
              sock_p->close();
            }
            delete sock_p;
            sock_p = NULL;
          }

          //!<Create socket
          sock_p = new zmq::socket_t(_context, type);
          if(sock_p != NULL){
            if(identity != ""){
              //!<Set identity to route responses from router
              sock_p->setsockopt(ZMQ_IDENTITY, identity.c_str(), identity.size());
            }

            //!<Try to connect
            unsigned int counter = 0;
            while(counter < getBindTriesMaxNumber()){
              try{
                sock_p->connect(connectionString.c_str());
                connected = true;
                break;
              } catch(zmq::error_t &e){
                logger.log(NODE_LOG_MODE_ERROR) << "Connect [" << connectionString << "] warning : " << e.what() << flush;
                counter++;
                s_sleep(getBindTriesDelay());
              }
            }
          }else{
            logger.log(NODE_LOG_MODE_FATAL) << "Client connection socket type " << type << " instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
            ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
          }

          logger.log(NODE_LOG_MODE_INFORMATION) << "Connect to [" << connectionString << "] ";
          if(connected){
            logger.log(NODE_LOG_MODE_INFORMATION) << "success" << flush;

            if(linger != -1){
              //!<Set linger for the socket (0 - to not wait at close time)
              sock_p->setsockopt(ZMQ_LINGER, &linger, sizeof(linger));
              logger.log(NODE_LOG_MODE_DEBUG) << "Linger value " << linger << " for [" << connectionString << "] set" << flush;
            }

            if(readyMessage != ""){
              //!<Send READY message to the server (especially for the Paranoid Pirate Protocol implementations)
              s_send(*sock_p, readyMessage);
              logger.log(NODE_LOG_MODE_INFORMATION) << "READY message sent from client [" << identity << "] to [" << connectionString << "]" << flush;
            }
          }else{
            logger.log(NODE_LOG_MODE_DEBUG) << "error" << flush;
          }

          return connected;
        }

        void sendByeMessage(zmq::socket_t*& sock_p, const std::string& identity="", const std::string& byeMessage = NODE_MSG_BYE){
          //!<Send BYE message to the server (especially for the Paranoid Pirate Protocol implementations)
          s_send(*sock_p, byeMessage);
          logger.log(NODE_LOG_MODE_INFORMATION) << "BYE message sent from client [" << ((identity.empty())?getName():identity) << "]" << flush;
        }

        std::string getProperties(void) {
          std::stringstream result;
          result << "name=" << getName() << NODE_ADMIN_COMMAND_DELIMITER << "startedAt=" << getStartedAt() << NODE_ADMIN_COMMAND_DELIMITER << "logPriority=" << static_cast<int>(getLogPriority())
              << NODE_ADMIN_COMMAND_DELIMITER << "connectionRebuildTriesMax=" << getBindTriesMaxNumber() << NODE_ADMIN_COMMAND_DELIMITER << "connectionRebuildTriesDelay="
              << getBindTriesDelay() << NODE_ADMIN_COMMAND_DELIMITER << "connectionString=" << getConnectionString() << NODE_ADMIN_COMMAND_DELIMITER;
          return result.str();
        }

        //!<Get property information
        virtual std::string getPropertyInformation(void){
          std::string ret("DEFAULT stub for property informaion!");
          return ret;
        }

        //!<Send periodical property values
        void sendPeriodicProperties(void){
          if(s_clock() > _propertyIntervalAt){

            HCE::handlers::PropertyMessage propertyMessage(getName(), getPropertyInformation());
            std::string message = propertyMessage.build();
            if (!propertyMessage.isError()){
              s_send(*_inprocAdminSock, message);

              if (getName() == "DataServerProxy"){
                //!<Log periodic properties sending
                logger.log(NODE_LOG_MODE_TRACE) << "Send 'Periodic Properties' from DataServerProxy to _inprocAdminSock" << flush;
              }
            }else{
              logger.log(NODE_LOG_MODE_ERROR) << "Operation build property message has error: " << propertyMessage.getErrorMsg() << flush;
            }
            _propertyIntervalAt = s_clock() + getPropertyInterval();
          }
        }

        //!<Safe create directories if necessary
        void createDirectories(const std::string& name){
          Poco::File dirPath(name);
          try{
            if (!dirPath.exists()){
              dirPath.createDirectories();
            }
          }catch(Poco::Exception& e){
            logger.log(NODE_LOG_MODE_DEBUG) << "Create directories '" << name << "' return: " << e.message() << flush;
          }
        }

        //!<Get file name use for dump and restore properties values
        std::string getDumpFileName(void) {
          createDirectories(_dumpDir);
          std::ostringstream ostr;
          ostr << _dumpDir << Poco::Path::separator() << getClientIdentity() << "_" << getName() << ".json";
          return ostr.str();
        }

        //!<Get file name use for dump and restore statistic values
        std::string getStatFileName(void) {
          createDirectories(_dumpDir);
          std::ostringstream ostr;
          ostr << _dumpDir << Poco::Path::separator() << getClientIdentity() << "_" << getName() << ".stat_vars";
          return ostr.str();
        }

        //!<Dump properties variables
        bool dumpProperties(void){
          //!<Save handler properties values
          if (!handlerProperties.save(getDumpFileName())){
            logger.log(HCE::handlers::NODE_LOG_MODE_DEBUG) << "Save handler properties has error: " << handlerProperties.getErrorMsg() << flush;
          }
          //!<Save stistic properties values
          if (!statProperties.save(getStatFileName())){
            logger.log(HCE::handlers::NODE_LOG_MODE_DEBUG) << "Save stat properties has error: " << statProperties.getErrorMsg() << flush;
          }
          return (!handlerProperties.isError() && !statProperties.isError());
        }

        //!<Dump run-time variables at regular intervals
        bool  dumpRunTimeVariables(void){
          bool result = false;
          if(s_clock() > _dumpIntervalAt && getDumpInterval() > 0){
            result = dumpProperties();
            _dumpIntervalAt = s_clock() + getDumpInterval();
          }
          return result;
        }

        virtual int initialize(void) = 0;
        virtual void deinitialize(void) = 0;
        virtual void* process(void) = 0;

        void handleInternalMessage(void);
        void handleExternalMessage(void);

        void processControlMessage(zmsg& msg);
        void processDataMessage(zmsg& msg);

        void logPeriodicStatistics(void);
        void heartbit(void);
        void registerClient(const std::string& identity);

        //!<Admin socket
        zmq::socket_t* _inprocAdminSock;

        //!<Storage for run-time property variables
        HandlerProperties handlerProperties;

        //!<Storage for run-time statistic variables
        HandlerProperties statProperties;

        //!<ZMQ context
        zmq::context_t& _context;

        //Log components mask
        std::bitset<4> _logMask;  // remove ??
        //Log time format POCO specification
        std::string _logTimeFormat; // remove ??
        //Log stream buffer (use only in method logFlush)
        std::stringstream log; // remove ??

        //Logger stream use now as log buffer
        LoggerStream logger;

        //!<Main loop of sockets polling condition
        bool _inProgress;

        //!<Send out periodical property values to Admin class at regular intervals
        int64_t _propertyIntervalAt;

        //!<Dump run-time variables at regular intervals
        int64_t _dumpIntervalAt;

        //!<Dump dir name
        std::string _dumpDir;
    };

  }
}

#endif /* __HCE_HANDLER_INCLUDED__ */
