#include "RouterServerProxy.hpp"
#include "ServiceMessages.hpp"

namespace HCE {
  namespace handlers {

    RouterServerProxy::RouterServerProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString,
        int64_t startedAt, unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, clientIdentity, connectionString, startedAt, logPriority, iniFile){

      //!<Init all the rest
      initialize();
    }

    RouterServerProxy::~RouterServerProxy(){
      deinitialize();
    }

    int RouterServerProxy::initialize(void){
      //!<Init sockets
      _inprocProxySock = NULL;
      _inprocReducerOutSock = NULL;
      _routerServerSock = NULL;

      //!<Check initialization and set default value if necessary
      checkInit();

      //!<Send out periodical stats at regular intervals
      setPstatsAt(s_clock() + LOG_STATISTIC_PERIOD);

      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << getClientIdentity() << "] with connection [" << getConnectionString() << "]" << flush;

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildClientConnection(_inprocProxySock, NODE_CS_INPROC_PROXY, ZMQ_DEALER);
      //!<Rebuild inproc_reducer_out connection
      rebuildClientConnection(_inprocReducerOutSock, NODE_CS_INPROC_REDUCER_OUT, ZMQ_DEALER);
      //!<Rebuild external server connection
      rebuildServerConnection(_routerServerSock, getConnectionString(), ZMQ_ROUTER);

      return true;
    }

    void RouterServerProxy::deinitialize(void){
      if(_inprocProxySock){
        delete _inprocProxySock;
      }
      if(_inprocReducerOutSock){
        delete _inprocReducerOutSock;
      }
      if(_routerServerSock){
        delete _routerServerSock;
      }
      _inprocProxySock = NULL;
      _inprocReducerOutSock = NULL;
      _routerServerSock = NULL;
      _responses.clear();
    }

    void RouterServerProxy::setStatRequests(unsigned int statRequests){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS, statRequests);
    }

    unsigned int RouterServerProxy::getStatRequests(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS);
    }

    void RouterServerProxy::setStatResponses(unsigned int statResponses){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES, statResponses);
    }

    unsigned int RouterServerProxy::getStatResponses(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_RESPONSES);
    }

    void RouterServerProxy::setStatRequestsSize(unsigned int statRequestsSize){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_SIZE, statRequestsSize);
    }

    unsigned int RouterServerProxy::getStatRequestsSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_SIZE);
    }

    void RouterServerProxy::setStatResponsesSize(unsigned int statResponsesSize){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_SIZE, statResponsesSize);
    }

    unsigned int RouterServerProxy::getStatResponsesSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_RESPONSES_SIZE);
    }

    void RouterServerProxy::setStatRequestsProcessed(unsigned int statRequestsProcessed){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_PROCESSED, statRequestsProcessed);
    }

    unsigned int RouterServerProxy::getStatRequestsProcessed(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_PROCESSED);
    }

    void RouterServerProxy::setStatResponsesProcessed(unsigned int statResponsesProcessed){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_PROCESSED, statResponsesProcessed);
    }

    unsigned int RouterServerProxy::getStatResponsesProcessed(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_RESPONSES_PROCESSED);
    }

    void RouterServerProxy::setStatRequestsProcessedSize(unsigned int statRequestsProcessedSize){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE, statRequestsProcessedSize);
    }

    unsigned int RouterServerProxy::getStatRequestsProcessedSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE);
    }

    void RouterServerProxy::setStatResponsesProcessedSize(unsigned int statResponsesProcessedSize){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_PROCESSED_SIZE, statResponsesProcessedSize);
    }

    unsigned int RouterServerProxy::getStatResponsesProcessedSize(void){
      return statProperties.get<unsigned int>(HCE::handlers::properties::STAT_RESPONSES_PROCESSED_SIZE);
    }

    void RouterServerProxy::setStatQPS(float statQPS){
      statProperties.set(HCE::handlers::properties::STAT_QPS, statQPS);
    }

    float RouterServerProxy::getStatQPS(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_QPS);
    }

    void RouterServerProxy::setStatRPS(float statRPS){
      statProperties.set(HCE::handlers::properties::STAT_RPS, statRPS);
    }

    float RouterServerProxy::getStatRPS(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_RPS);
    }

    void RouterServerProxy::setStatRequestsBandwidth(float statRequestsBandwidth){
      statProperties.set(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH, statRequestsBandwidth);
    }

    float RouterServerProxy::getStatRequestsBandwidth(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH);
    }

    void RouterServerProxy::setStatResponsesBandwidth(float statResponsesBandwidth){
      statProperties.set(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH, statResponsesBandwidth);
    }

    float RouterServerProxy::getStatResponsesBandwidth(void){
      return statProperties.get<float>(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH);
    }

    void RouterServerProxy::setPstatsAt(int64_t pstatsAt){
      handlerProperties.set(HCE::handlers::properties::PSTATS_AT, pstatsAt);
    }

    int64_t RouterServerProxy::getPstatsAt(void){
      return handlerProperties.get<int64_t>(HCE::handlers::properties::PSTATS_AT);
    }

    void RouterServerProxy::checkInit(void){
      //!<Check all stat variables
      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS))
        setStatRequests(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES))
        setStatResponses(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_SIZE))
        setStatRequestsSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_SIZE))
        setStatResponsesSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_PROCESSED))
        setStatRequestsProcessed(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_PROCESSED))
        setStatResponsesProcessed(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_PROCESSED_SIZE))
        setStatRequestsProcessedSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_PROCESSED_SIZE))
        setStatResponsesProcessedSize(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_QPS))
        setStatQPS(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RPS))
        setStatRPS(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_REQUESTS_BANDWIDTH))
        setStatRequestsBandwidth(0);

      if (!statProperties.isExist(HCE::handlers::properties::STAT_RESPONSES_BANDWIDTH))
        setStatResponsesBandwidth(0);
    }

    void RouterServerProxy::handleInternalMessage(void){
      //!<Fetch message from inproc_reducer_out
      zmsg msg(*_inprocReducerOutSock);
      logger.log(NODE_LOG_MODE_TRACE) << " MSG FROM inproc_reducer_out" << endl << msg.dump() << flush;

      unsigned int statResponses = getStatResponses();
      setStatResponses(++statResponses);

      //!<Parse message, create identity, messageId and body of dataclient's message
      std::string messageId((const char*)msg.pop_front().c_str());
      std::string messageBody((const char*)msg.pop_front().c_str());
      std::string messageRoute((msg.parts())?(const char*)msg.pop_front().c_str():"");

      setStatResponsesSize(getStatResponsesSize() + messageBody.size());
      logger.log(NODE_LOG_MODE_DEBUG) << " RECEIVED RESPONSE from inproc_reducer_out message [" << messageId << "] body " << messageBody.size() << " bytes, n=" << getStatResponses() << flush;
      //!<Generate unique client_Id and request Id from message_Id and rebuild message to send response to the client via ZMQ router
      std::vector<std::string> parts;
      //!<Split message_Id on two parts
      std::istringstream partsStream(messageId);
      std::string tmpString;

      while(std::getline(partsStream, tmpString, NODE_ROUTER_MSG_DELIMITER)){
        parts.push_back(tmpString);
      }
      if(parts.size() > 1){
        //!<Fill message with router`s client_Id, message_Id and messageBody
        msg.append(parts[0].c_str());
        msg.append(parts[1].c_str());
        msg.append(messageBody.c_str());

        logger.log(NODE_LOG_MODE_TRACE) << "MSG TO routerserver" << endl << msg.dump() << flush;

        //!<Send response message to client router`s client
        msg.send(*_routerServerSock);

        logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE SENT to routerserver socket from proxy node, n=" << getStatResponses() << flush;
      }else{
        //!<Wrong message_Id format
        logger.log(NODE_LOG_MODE_ERROR) << "ERROR : Wrong message_Id format [" << messageId << "] received from client" << flush;
      }
    }

    void RouterServerProxy::handleExternalMessage(void){
      //!<Fetch message from dataclient
      zmsg clientMsg(*_routerServerSock);

      //!<Data message
      if(clientMsg.parts() > NODE_MSG_DATA_MIN_FIELDS){
        logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM routerserver\n" << clientMsg.dump() << flush;

        unsigned int statRequest = getStatRequests();
        setStatRequests(++statRequest);

        //!<Parse message, create identity, messageId and body of dataclient's message
        std::string clientIdentity((const char*)clientMsg.pop_front().c_str());
        std::string messageId((const char*)clientMsg.pop_front().c_str());
        std::string messageBody((const char*)clientMsg.pop_front().c_str());
        std::string messageRoute((clientMsg.parts())?(const char*)clientMsg.pop_front().c_str():"");

        logger.log(NODE_LOG_MODE_DEBUG) << "RouterServerProxy::handleExternalMessage  Recieved clientIdentity: \"" << clientIdentity << "\"" << flush;
        logger.log(NODE_LOG_MODE_DEBUG) << "RouterServerProxy::handleExternalMessage  Recieved messageId: \"" << messageId << "\"" << flush;
        logger.log(NODE_LOG_MODE_DEBUG) << "RouterServerProxy::handleExternalMessage  Recieved body: \"" << messageBody << "\"" << flush;
        logger.log(NODE_LOG_MODE_DEBUG) << "RouterServerProxy::handleExternalMessage  Recieved route: \"" << messageRoute << "\"" << flush;

        setStatRequestsSize(getStatRequestsSize() + messageBody.size());
        logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from router`s client [" << getClientIdentity() << "], message [" << messageId << "] body "
            << messageBody.size() << " bytes, n=" << getStatRequests() << flush;

        logger.log(NODE_LOG_MODE_DEBUG) << "RouterServerProxy::handleExternalMessage  Recieved route: \"" << messageRoute << "\"" << flush;

        //!<Generate unique request Id from client_Id amd message_Id and rebuild message for farther processing
        clientMsg.clear();
        clientMsg.append(clientIdentity.c_str());
        clientMsg.append((clientIdentity + NODE_ROUTER_MSG_DELIMITER + messageId).c_str());
        clientMsg.append(messageBody.c_str());
        clientMsg.append(messageRoute.c_str());

        if(getLogPriority() >= NODE_LOG_MODE_DEBUG){
          logger.log(NODE_LOG_MODE_DEBUG) << " MSG TO inproc_proxy" << endl << clientMsg.dump() << flush;
        }

        //!<Send message to inproc_proxy socket
        clientMsg.send(*_inprocProxySock);

        logger.log(NODE_LOG_MODE_DEBUG) << " REQUEST SENT to inproc_proxy socket, n=" << getStatRequests() << flush;
      }else{
        logger.log(NODE_LOG_MODE_ERROR) << "ERROR MESSAGE RECEIVED from routerserver, parts=" << clientMsg.parts() << ", n=" << getStatRequests()
            << endl << clientMsg.dump() << flush;
      }
    }

    void RouterServerProxy::handleRouteMessage(void){

      //!<Fetch message from inproc_proxy
      zmsg msg(*_inprocProxySock);
      logger.log(NODE_LOG_MODE_DEBUG) << " MSG FROM inproc_proxy" << endl << msg.dump() << flush;

      std::string addr((msg.address())?msg.address():"");
      if (addr.compare(0, NODE_MSG_ROUTES.length(), NODE_MSG_ROUTES)==0)
      {
        std::vector<ClientWorkerItem> clients;
        RouteMessage routeMessage(clients);
        routeMessage.extract(NODE_MSG_ROUTES, addr);
        if (routeMessage.isError())
          logger.log(NODE_LOG_MODE_ERROR) << "Operation extract ROUTE MESSAGE has error: " << routeMessage.getErrorMsg() << flush;
        else
        {
          std::vector<ClientWorkerItem> clientsQueue;
          RouteMessage routesMsg(clientsQueue);
          ClientsQueueManager clientsManager(&clientsQueue, 0);
          clientsManager.refresh(getClientIdentity(), clients, "Here must be resources information...");
          std::string route = routesMsg.build();
          if (routesMsg.isError())
            logger.log(NODE_LOG_MODE_ERROR) << "Operation build ROUTE MESSAGE has error: " << routesMsg.getErrorMsg() << flush;
          else
          {
            logger.log(NODE_LOG_MODE_TRACE) << "RouterServerProxy::handleRouteMessage  ROUTE_TABLE: " << route << flush;
            s_send(*_routerServerSock, route);
          }
        }
      }
    }

    std::string RouterServerProxy::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << getStatRequests() << ",responses=" << getStatResponses() << ",requestsSize=" << getStatRequestsSize()
            << ",response_size=" << getStatResponsesSize();
        ret = result.str();
      }else if(command == NODE_MSG_rebuildServerConnection){
        //!<Rebuild server connection
        rebuildServerConnection(_routerServerSock, getConnectionString(), ZMQ_ROUTER);
      }else if(command == NODE_MSG_DISCONNECT_SERVER_CONNECTION){
        //!<Disconnect server connection
        _routerServerSock->close();
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    void RouterServerProxy::logPeriodicStatistics(void){
      if(s_clock() > getPstatsAt()){
        setStatQPS((getStatRequests() - getStatRequestsProcessed()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatRPS((getStatResponses() - getStatResponsesProcessed()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatRequestsBandwidth((getStatRequestsSize() - getStatRequestsProcessedSize()) / (float)((STAT_INTERVAL) / 1000.0));
        setStatResponsesBandwidth((getStatResponsesSize() - getStatResponsesProcessedSize()) / (float)((STAT_INTERVAL) / 1000.0));

        setStatRequestsProcessed(getStatRequests());
        setStatResponsesProcessed(getStatResponses());
        setStatRequestsProcessedSize(getStatRequestsSize());
        setStatResponsesProcessedSize(getStatResponsesSize());

        logger.log(NODE_LOG_MODE_DEBUG) << "requests=" << getStatRequests() << ", responses=" << getStatResponses() << ", qps=" << getStatQPS() << ", rps=" << getStatRPS()
            << ", requestsSize=" << getStatRequestsSize() << ", responsesSize=" << getStatResponsesSize() << ", requestsBandwidth=" << getStatRequestsBandwidth()
            << ", responsesBandwidth=" << getStatResponsesBandwidth() << flush;

        setPstatsAt(s_clock() + LOG_STATISTIC_PERIOD);
      }
    }

    //!<Get property information
    std::string RouterServerProxy::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "requests=" << getStatRequests() << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << getStatResponses()
          << NODE_ADMIN_COMMAND_DELIMITER << "requestsSize=" << getStatRequestsSize() << NODE_ADMIN_COMMAND_DELIMITER << "responseSize=" << getStatResponsesSize()
          << NODE_ADMIN_COMMAND_DELIMITER << "clientIdentity=" << getClientIdentity() << NODE_ADMIN_COMMAND_DELIMITER //<< "heartbeatAt" << static_cast<uint64_t>(_heartbeatAt)
          << NODE_ADMIN_COMMAND_DELIMITER << "pstatsAt=" << static_cast<uint64_t>(getPstatsAt()) << NODE_ADMIN_COMMAND_DELIMITER << "statQPS=" << getStatQPS()
          << NODE_ADMIN_COMMAND_DELIMITER << "statRPS=" << getStatRPS() << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidth=" << getStatRequestsBandwidth()
          << NODE_ADMIN_COMMAND_DELIMITER << "statResponsesBandwidth=" << getStatResponsesBandwidth();
      return result.str();
    }

    void* RouterServerProxy::process(void){
      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      while(getInProgress()){
        try {
          //!<Init polling pool sockets array
          std::vector<zmq::pollitem_t> items;
          if(_routerServerSock->connected()){
            items ={{*_routerServerSock, 0, ZMQ_POLLIN, 0},{*_inprocReducerOutSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0}};
          } else{
            items ={{*_inprocReducerOutSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0}};
          }

          //!<Poll sockets
          zmq::poll(&items[0], items.size(), getPollTimeout());

          //!<Handle routerserver's socket message if connected
          if(_routerServerSock->connected()){
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_reducer_out's socket message
          if(items[items.size() - 3].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleAdminMessage();
          }

          //!<Handle inproc_proxy's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleRouteMessage();
          }

          //!<Out periodical stats
          logPeriodicStatistics();

          //!<Send periodical property values
          sendPeriodicProperties();

          //!<Dump run-time variables at regular intervals
          dumpRunTimeVariables();

        }catch(zmq::error_t& error){
          if (EINTR!=error.num()){
            std::cout << "RouterServerProxy E: " << error.what() << std::endl;
          }
        }catch(std::exception& e){
          std::cout << "RouterServerProxy E: " << e.what() << std::endl;
        }catch(...){
          std::cout << "RouterServerProxy E: Unknown exception" << std::endl;
        }
      }

      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }

  }
}
