/**
 * @file RouterServerProxy.hpp
 *
 * HCE project RouterServerProxy handler.
 * Implements functionality for router server handler.
 * Contains client requests router server and cluster nodes (proxy or data) server.
 * This is root of cluster hierarchy and data streams root point for client-side applications connections and upper connection point for
 * data connections of all kind of nodes in cluster. Get request from client application, send request to all connected proxy or data nodes,
 * collects upper data flows in one response and send response back to client-side application.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_ROUTERSERVER_PROXY_INCLUDED__
#define __HCE_ROUTERSERVER_PROXY_INCLUDED__

#include <sstream>
#include <map>

#include "Handler.hpp"

namespace HCE {
namespace handlers {

class RouterServerProxy: public HCE::handlers::Handler {
public:
  RouterServerProxy(std::string name, zmq::context_t& ctx,
      const std::string& clientIdentity, std::string connectionString,
      int64_t startedAt, unsigned char logPriority, const std::string& iniFile);
  virtual ~RouterServerProxy();

protected:
  int initialize(void);
  void deinitialize(void);
  void* process(void);

  void handleInternalMessage(void);
  void handleExternalMessage(void);
  void handleRouteMessage(void);

  std::string processAdminCommands(std::string& command, std::string& parameters);

  void logPeriodicStatistics(void);

  //!<Get property information
  std::string getPropertyInformation(void);

  //!<Check initialization and set default value if necessary
  void checkInit(void);

  //!<ZMQ Sockets
  zmq::socket_t* _inprocProxySock;
  zmq::socket_t* _inprocReducerOutSock;
  zmq::socket_t* _routerServerSock;

  //!<Messages data accumulator map
  std::map<std::string, DefaultReducerResponseItem*> _responses;

  //!<Setters and getters for stat variables
  void setStatRequests(unsigned int statRequests);
  unsigned int getStatRequests(void);

  void setStatResponses(unsigned int statResponses);
  unsigned int getStatResponses(void);

  void setStatRequestsSize(unsigned int statRequestsSize);
  unsigned int getStatRequestsSize(void);

  void setStatResponsesSize(unsigned int statResponsesSize);
  unsigned int getStatResponsesSize(void);

  void setStatRequestsProcessed(unsigned int statRequestsProcessed);
  unsigned int getStatRequestsProcessed(void);

  void setStatResponsesProcessed(unsigned int statResponsesProcessed);
  unsigned int getStatResponsesProcessed(void);

  void setStatRequestsProcessedSize(unsigned int statRequestsProcessedSize);
  unsigned int getStatRequestsProcessedSize(void);

  void setStatResponsesProcessedSize(unsigned int statResponsesProcessedSize);
  unsigned int getStatResponsesProcessedSize(void);

  void setStatQPS(float statQPS);
  float getStatQPS(void);

  void setStatRPS(float statRPS);
  float getStatRPS(void);

  void setStatRequestsBandwidth(float statRequestsBandwidth);
  float getStatRequestsBandwidth(void);

  void setStatResponsesBandwidth(float statResponsesBandwidth);
  float getStatResponsesBandwidth(void);

  //!<Send out periodical stats at regular intervals
  void setPstatsAt(int64_t pstatsAt);
  int64_t getPstatsAt(void);
};

}
}

#endif /* __HCE_ROUTERSERVER_PROXY_INCLUDED__ */
